package rwd.abstractsyntax.eulerdiagram;
import java.util.*;
import java.io.Serializable;
/*
 * *
 * @author RWD, rnb
 * 
 * @version 11.8.05   12:26
 * 
 * This class models abstract contours. Contours each have a label, which must
 * be non-empty.  Derived contours have labels starting with the string "derived-".
 * The flyweight pattern is in use, so the constructor is private, 
 * but Contour objects can be obtained using the public static getContour method.
 * This pattern means that very few contour objects need to be built, 
 * and that contour objects can be compared by direct object equality.
 * Collections of contours can be compared simply using containsAll methods,
 * instead of having to extract each contour in turn and seek an equivalent 
 * one in the other collection.
 * 
 * The Serialisable and Comparable interfaces are left-overs from old code
 * and a good tidying up of the code may render these unnecessary.
 * In particular, Comparable was used so that collections of Contours could be
 * TreeSets (in the rwd.views.abstracttoconcrete.converting package), 
 * and if other collections are substituted, we may not meed to implement Comparable.
 * 
 * In principle, the toString() method should be alterable, but it would 
 * probably affect other code like 
 * (i) how textual diagram descriptions look
 * (ii) how diagrams are flattened to text files
 * (iii) a number of other places - alter toString with care!!
 * 
 * added method  isInContourLibrary  for debugging purposes RNB
 */
public class Contour 
	implements Comparable, Serializable 
{
	private static long id = 0;
	
	// constructor
	// -----------
	
	private Contour(String label) {
		setLabel(label);
	}

	
	// attributes
	// ----------
	
	public boolean equals(Object o) {
		if (o instanceof Contour) {
			return label.equals(((Contour)o).getLabel());
		}
		return false;
	}

	public Contour getClone() {
		return this;
	}

	public static Contour getContour(String label) {
		Contour result = getFromLibrary(label);
		if (result != null) {
			return result;
		} else {
			result = new Contour(label);
			contourLibrary.put(result.getLabel(), result);
			return result;
		}
	}

	public static Contour getFromLibrary(String requiredLabel) {
		return (Contour)contourLibrary.get(requiredLabel);
	}
	
	public String getLabel() {
		return label;
	}
	
	public boolean isDerived() {
		return label.startsWith("derived-");
	}
	
	/*
	 * pre:  c = contour object to be searched for in library
	 * post: returns true iff c is identically present in library
	 * 
	 */
	/*
	 * test:
	 *   c is in library
	 *   c is not in library
	 */
	public static boolean isInContourLibrary(Contour c){
		Iterator clit = contourLibrary.values().iterator();
		while (clit.hasNext()){
			Contour clc = (Contour) clit.next();
			if (clc==c) return true;
		}
		return false;		
	}

	public String toString() {
		return label;
	}


	// operations
	// ----------
	
	public int compareTo(Object o) {
		if (o instanceof Contour) {
			return label.compareTo(((Contour)o).label);
		} else {
			return -1;
		}
	}

	/**
	 * Post: Label has been set as required.
	 * @param s new value for label of this contour.
	 *        An empty string indicates a derived contour. 
	 */
	/*
	 * test:
	 *   s is null
	 *   s = ""
	 *       "]"
	 *       "a "
	 */
	public void setLabel(String s) {
		if(s == null || s.equals("") || s.equals("]")) { 
			label = "derived-"+id++; 
		}
		else
		    label = s.trim();
	}

	
	// Local properties
	// ----------------
	
	/* Contour label */
	private String label;
	
	/* Contour library */
	private static Map contourLibrary = new LinkedHashMap();
	
    /*
    // test program
    // ------------

    private static void pr(String s) {
      System.out.println(s);
    }

    private static void hd(String s) {
      System.out.println("\n" + s + "\n");
    }

    private static void pg() {
      try {
        int i;
        System.out.print("next page? (CR)");
        do i = System.in.read(); while (i != '\n');
        System.out.println();
      }
      catch (Exception e) {System.out.println("raised exception : " + e);}
    }

    public static void main(String args[]) throws Exception {
    	{
			hd("Tests of isInContourLibrary");
			pr("a1 = Contour.getContour(\"a\")");
			Contour a1 = Contour.getContour("a");
			pr("a1=" + a1 + "  a1.hashCode()="+a1.hashCode());
			pr("Contour.isInContourLibrary(a1)="+
				    Contour.isInContourLibrary(a1));
			pr("b = Contour.getContour(\"b\")");
			Contour b = Contour.getContour("b");
			pr("b=" + b + "  b.hashCode()="+b.hashCode());
			pr("Contour.isInContourLibrary(b)="+
				    Contour.isInContourLibrary(b));
			pr("a2 = Contour.getContour(\"a\")");
			Contour a2 = Contour.getContour("a");
			pr("a2=" + a2 + "  a2.hashCode()="+a2.hashCode());
			
			pr("Contour.isInContourLibrary(a2)="+
			    Contour.isInContourLibrary(a2));
			pg();
			pr("a3 = new Contour(\"a\")");
			Contour a3 = new Contour("a");
			pr("a2=" + a3 + "  a3.hashCode()="+a3.hashCode());
			pr("Contour.isInContourLibrary(a3)="+
			    Contour.isInContourLibrary(a3));
			pg();
    	}
    	
    	{
    		hd("Test of setLabel");
    		Contour c = new Contour("a");
    		pr("c="+c);
    		pg();
    		
    		pr("c.setLabel(null)");
    		c.setLabel(null);
    		pr("c="+c);
    		pg();
    		
    		pr("c.setLabel(\"\")");
    		c.setLabel("");
    		pr("c="+c);
    		pg();
    		
    		pr("c.setLabel(\"]\")");
    		c.setLabel("]");
    		pr("c="+c);
    		pg();
    		
    		pr("c.setLabel(\"b\")");
    		c.setLabel("b");
    		pr("c="+c);
    		pg();
    		
    	}
    
    }
    */
}
