package rwd.abstractsyntax.eulerdiagram;

import java.util.*;

import java.io.Serializable;

import rwd.abstractsyntax.spiderdiagram.UnitaryDiagram;
/*
 * *
 * @author RWD, rnb
 * 
 * @version 28.2.05   12:44
 * 
 * This class models abstract zones. Abstract zones hold information
 * about which abstract contours contain that zone and which exclude it.
 * 
 * Zones can belong to more than one diagram - this is important for
 * several reasons
 * (i) the purpose of the abstract syntax is to be close to the diagram
 * semantics - the same abstract zone maps to the same set
 * (ii) zones, even if they are in different diagrams, can be compared
 * using object equality instead of an equals method.  This is fast and
 * allows for simple comparison of zone sets (using containsAll).
 * 
 * Zones can be obtained using getZone(..), either providing two
 * Vectors of Contour obects, or providing two Strings, assuming contour
 * names are single letters.
 * 
 * The Serialisable and Comparable interfaces are left-overs from old code
 * and a good tidying up of the code may render these unnecessary.
 * In particular, Comparable was used so that collections of Zones could be
 * TreeSets (in the rwd.views.abstracttoconcrete.converting package), 
 * and if other collections are substituted, we may not meed to implement Comparable.
 * 
 * In principle, the toString() method should be alterable, but it would 
 * probably affect other code like 
 * (i) how textual diagram descriptions look
 * (ii) how diagrams are flattened to text files
 * (iii) a number of other places - alter toString with care!!
 * 
 * Made constructor private to avoid unintentional(?) cloning 25.2.05  07:38 
 * Made it public again because a constructor is needed in 
 * rwd.ocl.constraintdiagram.DependenceGraph to build a zone iteratively. RNB.
 * 
 * Overrode clone() to ensure that if cloning is ever used it will output a
 * console message, and return the original object.RNB.
 */

public class Zone implements Serializable, Comparable {
	
	// constructor
	// -----------
	
	/**
	 * Returns a new Zone object. 
	 * @param inside containing contours of this zone
	 * @param outside = excluding contours of this zone
	 */
	public Zone(Vector inside, Vector outside) {
		containingContours = inside;
		excludingContours = outside;
		// Order contours
		containingContours = sort(containingContours);
		excludingContours = sort(excludingContours);
//		orderedCContours = new TreeSet(containingContours);
//		orderedEContours = new TreeSet(excludingContours);
	}

	/**
	 * Method getContainingContours.
	 * @return Vector - because Zone objects are shared between Diagrams
	 * I would like to prevent clients from altering the contents of
	 * these attributes - could clone() but it'd be slow - not sure how to
	 * address this point.
	 * Any change needs to be made with great care - some work may indeed
	 * deliberately change the contents - and that work would need to be changed 
	 * before any change is made here.
	 */
	public final Vector getContainingContours() {
		return containingContours;
	}

	public final Vector getExcludingContours() {
		return excludingContours;
	}

	public Vector getAllContours() {
		Vector result = new Vector();
		result.addAll(containingContours);
		result.addAll(excludingContours);
		return result;
	}

	/**
	 * Method isShaded.
	 * @param concext
	 * @return boolean
	 * 
	 * This is the only meaningful way that Zone can know if it's shaded -
	 * it needs to be given a UnitaryDiagram as a context for the question.
	 * This is the only time that UnitaryDiagram is referenced in this package -
	 * maybe we could ask the diagram whether the Zone is shaded instead of 
	 * asking the Zone?
	 */
	public boolean isShaded(UnitaryDiagram context) {
		return context.containsShaded(this);
	}

	protected Vector sort(Vector l) {
		boolean sorted = false;
		while (!sorted) {
			sorted = true;
			for(int i=0; i<l.size()-1; i++) {
				Contour c1 = (Contour)l.get(i);
				Contour c2 = (Contour)l.get(i+1);
				if (c1.compareTo(c2)>0) {
					l.set(i, c2);
					l.set(i+1, c1);
					sorted = false;
				}
			}
		}
		return l;
	}

	public boolean isEmpty() {
		return this.containingContours.size() == 0 && this.excludingContours.size() == 0;
	}

	public boolean equals(Object z2) {
		if (z2 instanceof Zone) {
			int i = compareTo(z2);
			return i == 0;
		}
		return false;
	}

	public int compareTo(Object o) {
		if (!(o instanceof Zone)) return -1;
		Zone z = (Zone) o;
		int inSize1 = getContainingContours().size();
		int outSize1 = getExcludingContours().size();
		int inSize2 = z.getContainingContours().size();
		int outSize2 = z.getExcludingContours().size();
		if (inSize1 > inSize2) {
			return 1;
		} else if (inSize1 < inSize2) {
			return -1;
		} else {
//			int res = compareContours(this.orderedCContours, z.orderedCContours);
			int res = compareContours(this.containingContours, z.containingContours);
			if (res != 0)
				return res;
		}
		if (outSize1 > outSize2) {
			return 1;
		} else if (outSize1 < outSize2) {
			return -1;
		} else {
//			int res = compareContours(orderedEContours, z.orderedEContours);
			int res = compareContours(this.excludingContours, z.excludingContours);
			if (res != 0)
				return res;
		}
		return 0;
	}
	
	protected int compareContours(Collection c1, Collection c2) {
		if (c1 != c2) {
			Iterator thisIt = c1.iterator();
			Iterator thatIt = c2.iterator();
			while (thisIt.hasNext() && thatIt.hasNext()) {
				Contour thisc = (Contour) (thisIt.next());
				Contour thatc = (Contour) (thatIt.next());
				if (thisc.compareTo(thatc) != 0) {
					return thisc.compareTo(thatc);
				}
			}
		}
		return 0;
	}

	public String toString() {
		String result = "z( ";
		for (int i = 0; i < containingContours.size(); i++) {
			result += ("" + containingContours.get(i) + "");
		}
		result += " , ";
		for (int i = 0; i < excludingContours.size(); i++) {
			result += ("" + excludingContours.get(i) + " ");
		}
		result += " )";
		if (result.equals("")) {
			return "outside";
		} else {
			return result;
		}
	}

	/**
	 * Method getZoneFor.
	 * @param name - this method builds a containing contour set whose contours 
	 * are each labelled with single characters from this String
	 * @param allLabels this method builds a complete contour set whose contours 
	 * are each labelled with single characters from this String, - this is 
	 * expected to be a superset of the contours built from @param name. The
	 * excluding contours are found as the set difference.
	 * @return Zone
	 */
	public static Zone getZoneFor(String name, String allLabels) {
		// treat each character as contour label
		// there may be repetition in allLabels
		Vector contours = new Vector();
		// go through allLabels
		for (int i = 0; i < allLabels.length(); i++) {
			String letter = allLabels.substring(i, i + 1);
			Contour c = Contour.getContour(letter);
			if (!contours.contains(c)) {
				contours.add(c);
			}
		}
		// go through contours extracting those appearing in "name"
		Vector contoursContainingZone = new Vector();
		for (int i = 0; i < name.length(); i++) {
			String letter = name.substring(i, i + 1);
			Contour c = Contour.getContour(letter);
			if (!contoursContainingZone.contains(c)) {
				contours.remove(c);
				contoursContainingZone.add(c);
			}
		}
		return Zone.getZone(contoursContainingZone, contours);
	}
	/**
	 * Method getZone.
	 * @param inside is an Vector of Contour objects, representing those contours
	 * which contain the zone
	 * @param outside is an Vector of Contour objects, representing those contours
	 * which exclude the zone
	 * @return Zone - noting that this Zone is shared between diagrams.
	 */
	public static Zone getZone(Vector inside, Vector outside) {
		// scan through the library comparing zones there with requirements
		for (int i = 0; i < zoneLibrary.size(); i++) {
			Zone z = (Zone) (zoneLibrary.get(i));
			Vector in = z.containingContours;
			Vector out = z.excludingContours;
			if (in.containsAll(inside)
				&& inside.containsAll(in)
				&& out.containsAll(outside)
				&& outside.containsAll(out)) {
				return z;
			}
		}
		// zone wasn't in library so build it and add to library
		Zone z = new Zone(inside, outside);
		zoneLibrary.add(z);
		//System.out.println("just built a new zone:"+z.toString());
		return z;
	}
	
	/**
	 * Method getZoneForStrings.
	 * @param contours - an Vector of all the contours for a UnitaryDiagram 
	 * - this method only really works if the contours are labelled with
	 * single characters
	 * @param zone - a String whose characters give the contours which contain the
	 * zone
	 * @return Zone
	 */
	public static Zone getZoneForStrings(Vector contours, String zone) {
		// build contourSet and the
		// pick off one character at a time from zoneLabel
		//System.out.println("asked for a new Zone "+zone
		//                      +", from contours "+contours);
		Vector containingContours = new Vector();
		for (int i = 0; i < zone.length(); i++) {
			Contour c = Contour.getContour(String.valueOf(zone.charAt(i)));
			if (!containingContours.contains(c)) {
				containingContours.add(c);
			}
		}
		//System.out.println("contours is "+contours+": remove "+containingContours);
		for (int i = 0; i < containingContours.size(); i++) {
			Contour c = (Contour) (containingContours.get(i));
			//System.out.println("contour "+c+" is in contours?"+contours.contains(c));
			contours.remove(c);
		}
		//System.out.println("contours is "+contours+": after removing "+containingContours);
		return Zone.getZone(containingContours, contours);
	}
	
	protected Object clone() {
		System.out.println("Zone.clone: start");
		return this;
	}

	//
	// Local properties
	//
	public static Vector zoneLibrary = new Vector();
	private Vector containingContours;
	private Vector excludingContours;
//	private TreeSet orderedCContours;
//	private TreeSet orderedEContours;


	/*
	 * pre:  z = zone object to be searched for in library
	 * post: returns true iff z is identically present in library
	 * 
	 */
	/*
	 * test:
	 *   z is in library
	 *   z is not in library
	 */
	public static boolean isInZoneLibrary(Zone z){
		Iterator zlit = zoneLibrary.iterator();
		while (zlit.hasNext()){
			Zone zlz = (Zone) zlit.next();
			if (zlz==z) return true;
		}
		return false;		
	}

	/*
    // test program
    // ------------

    private static void pr(String s) {
      System.out.println(s);
    }

    private static void hd(String s) {
      System.out.println("\n" + s + "\n");
    }

    private static void pg() {
      try {
        int i;
        System.out.print("next page? (CR)");
        do i = System.in.read(); while (i != '\n');
        System.out.println();
      }
      catch (Exception e) {System.out.println("raised exception : " + e);}
    }

    public static void main(String args[]) throws Exception {
    	
  	  hd("Tests of isInZoneLibrary");

  	  Vector va = new Vector();
      va.add(Contour.getContour("a"));
  	  pr("va="+va);
  	  
      Vector vb = new Vector();
      vb.add(Contour.getContour("b"));
	  pr("vb="+vb);
	  
	  Zone zab = Zone.getZone(va,vb);
      pr("zab=" + zab + "  zab.hashCode()="+zab.hashCode());
      Zone zba = Zone.getZone(vb,va);
      pr("zba=" + zba + "  zba.hashCode()="+zba.hashCode());
      
      pr("Zone.isInZoneLibrary(zab)="+
          Zone.isInZoneLibrary(zab));
      pg();
      
      Zone zab2 = new Zone(va,vb);
      pr("zab2=" + zab2 + "  zab2.hashCode()="+zab2.hashCode());
      pr("Zone.isInZoneLibrary(zab2)="+
          Zone.isInZoneLibrary(zab2));
      pg();
    }
    */
   


}





