package rwd.concretesyntax.constraintdiagram;

import java.awt.BasicStroke;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GraphicsEnvironment;
import java.awt.Polygon;
import java.awt.Shape;

import javax.swing.JComponent;
import javax.swing.JLabel;

import rwd.concretesyntax.DiagramItem;
import rwd.concretesyntax.eulerdiagram.Polygonal;
import rwd.gui.editor.DiagramCanvas;

/**
 * @author Octavian Patrascoiu, rnb
 * @version 6.7.05   16:28
 *
 */
public class ConcreteTextBox 
	extends DiagramItem 
	implements Polygonal, java.io.Serializable 
{
	/**
	 * post: new ConcreteTextBox has been constructed with required properties
	 *  
	 * @param type type of box (NORMAL, BORDER)
	 * @param x x-coord of NW corner of box
	 * @param y y-coord of NW corner of box
	 * @param text text of box
	 */
	public ConcreteTextBox(int type, int x, int y, String text) {
		super.setLabel("");
		this.type = type;
		this.text = text;
		this.setLocation(x, y);
		this.setThickness(DiagramCanvas.THIN.getLineWidth());
		_centerX = median(this.xs, 4);
		_centerY = median(this.ys, 4);
	}
	
	public void drawString(Graphics g) {
		if (_font == null) _font = g.getFont();
		// Draw text
		Font backup = g.getFont();
		g.setFont(_font);
		String s = DiagramItem.replaceUnicodeSequences(text);

		// Get measures needed to set the location of the message
		FontMetrics fm = g.getFontMetrics();
		int msgWidth = computeWidth(s, fm);
		int msgHeight = computeHeight(s, fm);
		int lineHeight = computeHeight(fm);
		int ascent = computeAscent(fm);
		int descent = computeDescent(fm);
		
		// Use the string width to find the starting point
		int msgX = xs[0];
		int msgY = ys[0]+ascent;

		// Split text into lines
		String lines[] = s.split("\n");
		// Print each line
		for(int i=0; i<lines.length; i++) {
			g.drawString(lines[i], msgX, msgY);
			msgY += lineHeight; 
		}
		g.setFont(backup);
	}

	public Polygon getPolygon() {
		String s = DiagramItem.replaceUnicodeSequences(text);
		// Get measures needed to set the location of the message
		JComponent c = new JLabel();
		FontMetrics fm = c.getFontMetrics(_font);
		int msgWidth = computeWidth(s, fm);
		int msgHeight = computeHeight(s, fm);
		int ascent = computeAscent(fm);
		int descent = computeDescent(fm);
		int tempX[] = {
			xs[0],
			xs[0]+msgWidth,
			xs[0]+msgWidth,
			xs[0]
		};
		int tempY[]	= {
			ys[0],
			ys[0],
			ys[0]+msgHeight,
			ys[0]+msgHeight
		};			
		Polygon poly = new Polygon(tempX, tempY, 4);
		return poly;
	} 

	private int computeWidth(String text, FontMetrics fm) {
		if (fm == null) return 0;
		// Split text into lines
		String lines[] = text.split("\n");
		// Compute maximum width
		int msgWidth = 0;
		for(int i=0; i<lines.length; i++) {
			int width = fm.stringWidth(lines[i]);
			if (width > msgWidth) {
				msgWidth = width;
			}
		}
		return msgWidth;
	}

	private int computeHeight(FontMetrics fm) {
		if (fm == null) return 0;
		int msgHeight = fm.getHeight();
		return msgHeight;
	}

	private int computeLeading(FontMetrics fm) {
		if (fm == null) return 0;
		return fm.getLeading();
	}

	private int computeHeight(String text, FontMetrics fm) {
		if (fm == null) return 0;
		// Split text into lines
		String lines[] = text.split("\n");
		int msgHeight = fm.getHeight()*lines.length;
		return msgHeight;
	}

	private int computeAscent(FontMetrics fm) {
		if (fm == null) return 0;
		int ascent = fm.getMaxAscent();
		return ascent;
	}

	private int computeDescent(FontMetrics fm) {
		if (fm == null) return 0;
		int descent= fm.getMaxDescent();
		return descent;
	}

	public Shape getShape(boolean smooth) {
		return getPolygon();
	}
	public Shape getShape() {
		return getShape(false);
	}

	public void draw(Graphics2D g, boolean smoothed) {
		// Draw border
		if (getBoxType() == BORDER) {
			Polygon polygon = getPolygon();
			g.setColor(getForeColor());
			BasicStroke stroke = new BasicStroke(
					getThickness(),
					BasicStroke.CAP_BUTT,
					BasicStroke.JOIN_ROUND);
			g.setStroke(stroke);
			g.draw(polygon);
		}
		// Draw text
		String text = getText();
		drawString(g);
	}

	public String getText() {
		return text;
	}
	public void setText(String text) {
		this.text = text;
	}

	public String getType() {
		return "Text";
	}
	
	public int getBoxType() {
		return type;
	}
	public void setBoxType(int type) {
		this.type = type;
	}

	public java.util.Set getHandles() {
		// Set handles
		java.util.Set _handles = new java.util.LinkedHashSet();
		return _handles;
	}

	public void setLocation(int x, int y) {
		this.xs[0] = x;
		this.ys[0] = y;
	}
	
	public int getX() {
		return xs[0];
	}
	public int getY() {
		return ys[0];
	}

	public void translateTo(float dx, float dy) {
		for(int i=0; i<4; i++) {
			xs[i] += dx;
			ys[i] += dy;
		}
	}
	
    /**
     * Note: unlike clone(), this returns a DiagramItem, and does not throw
     * an exception.
     * @return a deep clone of this diagram item
     */
    /*
     * test: once thru
     */
	public DiagramItem getClone() {
		ConcreteTextBox res = new ConcreteTextBox(type, getX(), getY(), text);
		res.setAttributes(this);
		return res;
	}
	
	
	/**
	 * 
	 * @return a user-friendly description of this item
	 */
	public String id() {
		return "text box \"" + getText() + "\"";
	}
	

	
	// Global properties
	
	/** normal box type */
	public static final int NORMAL = 1;
	/** border box type */
	public static final int BORDER = 2;
	
	/** margin of text in box */
	public static final int MARGIN = 70;
	
	/** gap between lines in box */
	public static final int LINE_GAP = 30;
	
	
	// Local properties
	private String text;
	private int type;
	private int xs[] = new int[4];
	private int ys[] = new int[4];

	// Font name lister
	// ----------------
	
	public static void main(String args[]) {
	   GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
	   String fontNames[] = ge.getAvailableFontFamilyNames();
	   int j = fontNames.length;
	   for (int i=0; i<j; i++) {
		 	System.out.println(fontNames[i]);
	   }
	}
	
	
	/*
	// test program
    // ------------

    private static void pr(String s) {
      System.out.println(s);
    }

    private static void hd(String s) {
      System.out.println("\n" + s + "\n");
    }

    private static void pg() {
      try {
        int i;
        System.out.print("next page? (CR)");
        do i = System.in.read(); while (i != '\n');
        System.out.println();
      }
      catch (Exception e) {System.out.println("raised exception : " + e);}
    }

    public static void main(String args[]) throws Exception {
    	hd("Test of toString, hashCode, id");
    	ConcreteTextBox ctb = new ConcreteTextBox(NORMAL,1,2,"fred");
    	pr("ctb="+ctb);
    	pr("ctb.hashCode()="+ctb.hashCode());
    	pr("ctb.id()="+ctb.id());
    	pg();
    	
    	ConcreteTextBox ctbc = (ConcreteTextBox) ctb.getClone();
    	pr("ctbc="+ctbc);
    	pr("ctbc.hashCode()="+ctbc.hashCode());
    	pr("ctbc.id()="+ctbc.id());
    	pr("ctb.equals(ctbc)="+ctb.equals(ctbc));
    	pr("ctbc.equals(ctb)="+ctbc.equals(ctb));
    	pr("ctb.equals(null)="+ctb.equals(null));
    	pr("ctb.equals(\"jim\")="+ctb.equals("jim"));
    	pg();
    	
    	
    }
    */
}
