package rwd.concretesyntax.eulerdiagram;

import java.awt.*;
import java.awt.geom.*;

import rwd.abstractsyntax.eulerdiagram.Contour;
import rwd.concretesyntax.DiagramItem;
import rwd.gui.editor.ContourHandle;
import rwd.gui.editor.EditorFrame;
import rwd.metrics.AreaPolygonAdapter;

/**
 * This class represents a contour as shown on a concrete diagram.
 * Different concrete diagrams will hold unique ConcreteContour objects
 * as the shape of the contour will be different in different diagrams.
 * 
 * It implements DiagramItem and Serializable so that diagrams can 
 * be shown and saved from the Editor
 * 
 * A ConcreteContour has an associated polygon and label.
 * The polygon must have at least one point.
 * The label must be non-empty.
 * Derived contours have labels starting with the string "derived-".
 * 
 * A ConcreteContour has an associated abstract contour which is specified by the 
 * label.
 * 
 * @author RWD,Richard Bosworth
 * 
 * @version 11.8.05   12:05
 */

public class ConcreteContour 
	extends DiagramItem 
	implements Comparable, java.io.Serializable, Polygonal 
{
	
	// constructor
	// -----------
	
	/**
	 * post: concrete contour has been instantiated.
	 *       If its corresponding abstract contour is not derived,
	 *         its label has been positioned.
	 * @param contour abstract contour of this concrete contour
	 * @param poly polygon of this concrete contour (must have at least one point)
	 */
	/*
	 * test:
	 *   derived contour
	 *   non-derived contour
	 */
 	public ConcreteContour(Contour contour, Polygon poly) {
		this._numPoints = poly.npoints; 
		this._border = makeBorder(poly);
		this._polygon = makePolygon(_border, _numPoints);
        _centerX = median(poly.xpoints, poly.npoints);
        _centerY = median(poly.ypoints, poly.npoints);
        _label = contour.getLabel();
        if (!contour.isDerived())
        	autoPlaceLabel();
 	}
 	
    /**
     * Post: label of this item has been set to required value.
     *       If old label was derived, new label has been autopositioned.
     * @param label required value of label (may be empty)
     */
 	/*
 	 * pre: _label = current value of label
 	 */
    /*
     * test:
     *   old label empty
     *     new label empty
     *     new label non-empty
     *   old label non-empty
     *     new label empty
     *     new label non-empty
     */
    public void setLabel(String label) {
    	if (this.isDerived() && !label.equals(""))
    		autoPlaceLabel();
    	if (!(this.isDerived() && label.equals(""))) {
        	Contour cnew = Contour.getContour(label);
	        _label = cnew.getLabel();
    	}
    }
    
 	/**
 	 * @return the polygon associated with this contour
 	 */
	public Polygon getPolygon() {
		return _polygon;
	}

	/**
	 * post: polygon p has been associated with this contour.
	 * @param p polygon to be associated with this contour.
	 * 
	 */
	public void setPolygon(Polygon p) {
		_border = makeBorder(p);
		_numPoints = p.npoints;
		_polygon = p;
	}

	/**
	 * @return the abstract contour associated with this concrete contour
	 */
	/*
	 * test: once thru
	 */
	public Contour getContour() {
		return Contour.getContour(getLabel());
	}

	/**
	 * post: newc is associated with this concrete contour.
	 * @param newc abstract contour to be associated with this concrete contour
	 */
	/*
	 * test: once thru
	 */
	public void setContour(Contour newc) {
		setLabel(newc.getLabel());
	}

	/**
	 * @return string giving type of this DiagramItem
	 */
	public String getType() {
		return "contour";
	}
	
	/**
	 * @return x-coord of the center of this concrete contour
	 */
	public float getX() {
		return _centerX;
	}

	/**
	 * @return y-coord of the center of this concrete contour
	 */
	public float getY() {
		return _centerY;
	}
	
	/**
	 * @return unsmoothed shape of this concrete contour
	 */
	public Shape getShape(){
		return getShape(false);
	}
    
	/**
	 * @param smooth true iff the smoothed shape is to be returned
	 * @return shape of this concrete contour
	 */
	public Shape getShape(boolean smooth) {
		if (EditorFrame.options.isAsEllipse()) {
			return makeEllipse(_border);
		} else if (smooth) {
			return AreaPolygonAdapter.getSmoothedShape(
				getPolygon(),
				AreaPolygonAdapter.BEZIER_CONTROL_DISTANCE);
		} else {
			return getPolygon();
		}
	}

	/**
	 * post: this concrete contour has been drawn on g.
	 * @param g graphics object to be drawn on
	 * @param smoothed true iff smoothewd version of contour is to be drawn
	 */
	public void draw(Graphics2D g, boolean smoothed) {
		Shape shape = getShape(smoothed);
		BasicStroke stroke = new BasicStroke(
				getThickness(),
				BasicStroke.CAP_BUTT,
				BasicStroke.JOIN_ROUND);
		g.setStroke(stroke);
		g.setColor(getForeColor());
		g.draw(shape);
		// Draw label
		drawLabel(g);
	}

	/**
	 * @return set of handles for this concrete contour
	 */
    public java.util.Set getHandles() {
    	// Add corners
		java.util.Set res = new java.util.LinkedHashSet();
		for(int i=0; i<_border.npoints; i++) {
			ContourHandle h = new ContourHandle(_border.xpoints[i], _border.ypoints[i], i, this);
			h.setIsBorder(true);
			res.add(h);
		}
		// Add the rest
		for (int i = 0; i < _polygon.npoints; i++) {
			ContourHandle handle =
				new ContourHandle(_polygon.xpoints[i], _polygon.ypoints[i], i, this);
			handle.setIsBorder(false);
			res.add(handle);
		}
		return res;
    }
	
    /**
     * post: concrete contour and its associated label have been translated as 
     *       specified.
     * @param dx distance in the x-direction by which contour is to be translated
     * @param dy distance in the y-direction by which contour is to be translated
     */
	public void translateTo(float dx, float dy) {
		_border.invalidate();
		_polygon.invalidate();
		_border.translate((int)dx, (int) dy);
		_polygon.translate((int)dx, (int) dy);
		super.translateTo(dx,dy);
	}

	/**
	 * post: contour has been resized as specified.
	 *       Label has same offset from rightmost point of contour.
	 * @param point code for position of resize handle (top left, top right,
	 *          bottom right, bottom left)
	 * @param _previousX x-position of resize handle before resize
	 * @param _currentX x-position of resize handle after resize
	 * @param _previousY y-position of resize handle before resize
	 * @param _currentY y-position of resize handle after resize
	 */
	/*
	 * test:
	 *   label is not in original position
	 */
	public void resize(int point, float _previousX, float _currentX, float _previousY, float _currentY) {
		// find label offsets
		Point2D lpos = getLabelPosition();
		int[] xy = AreaPolygonAdapter.getRightmostPoint(_polygon);
		int xoff = (int) lpos.getX() - xy[0];
		int yoff = (int) lpos.getY() - xy[1];

		// Update the border
		updateBorder(_border, point, _previousX, _currentX, _previousY, _currentY);

		// Make a new polygon
		_numPoints = _polygon.npoints;
		Polygon newPoly = makePolygon(_border, _numPoints);
		_polygon.invalidate();
		_polygon = newPoly;

		// Trigger the AWT thread
		translateTo(0,0);

		// Place label
		autoPlaceLabel();
	}
	
	/**
	 * 
	 * @param point polygon point id (0,1,..)
	 * @param _previousX x-position of point before move
	 * @param _currentX x-position of point after move
	 * @param _previousY y-position of point before move
	 * @param _currentY y-position of point after move
	 */
	public void movePoint(int point, float _previousX, float _currentX, float _previousY, float _currentY) {
		// Set coordinates for point
		float dx = _currentX - _previousX;
		float dy = _currentY - _previousY;
		_polygon.xpoints[point] += dx;
		_polygon.ypoints[point] += dy;
	}

	/**
	 * post: border has been updated as specified.
	 * @param -border Polygon for the border
	 * @param point code for position of resize handle (top left, top right,
	 *          bottom right, bottom left)
	 * @param _previousX x-position of resize handle before resize
	 * @param _currentX x-position of resize handle after resize
	 * @param _previousY y-position of resize handle before resize
	 * @param _currentY y-position of resize handle after resize
	 */
	protected void updateBorder(Polygon _border, int point, float _previousX, float _currentX, float _previousY, float _currentY) {
		// Set coordinates for point
		float dx = _currentX - _previousX;
		float dy = _currentY - _previousY;
		_border.xpoints[point] += dx;
		_border.ypoints[point] += dy;
		// Top left
		if (point == getTopLeft()) {
			_border.xpoints[getBottomLeft()] = _border.xpoints[point];
			_border.ypoints[getTopRight()] = _border.ypoints[point];
		// Top right
		} else if (point == getTopRight()) {
			_border.xpoints[getBottomRight()] = _border.xpoints[point];
			_border.ypoints[getTopLeft()] = _border.ypoints[point];
		// Bottom right
		} else if (point == getBottomRight()) {
			_border.xpoints[getTopRight()] = _border.xpoints[point];
			_border.ypoints[getBottomLeft()] = _border.ypoints[point];
		// Bottom left
		} else if (point == getBottomLeft()) {
			_border.xpoints[getTopLeft()] = _border.xpoints[point];
			_border.ypoints[getBottomRight()] = _border.ypoints[point];
		}
	}

	/**
	 * post: label has been translated as specified.
	 * @param _previousX x-position of label before translation
	 * @param _currentX x-position of label after translation
	 * @param _previousY y-position of label before translation
	 * @param _currentY y-position of label after translation
	 */
	protected void translateLabel(float _previousX, float _currentX, float _previousY, float _currentY) {
		float dx = _currentX - _previousX;
		float dy = _currentY - _previousY;
		_labelX += dx;
		_labelY += dy;		
	}
	
	
	private static final int LABEL_Y_OFFSET = -60;
    /**
     * post: label has been placed for this concrete contour.
     */
	public void autoPlaceLabel() {
		Point p = AreaPolygonAdapter.getTopPoint(_polygon);
		p.y += LABEL_Y_OFFSET;
		setLabelPosition(p);
	}

	/**
	 * Compares this object with the specified object for order.
	 * @param o object to be compared with this object 
	 * @return a negative integer, zero, or a positive integer as this object 
	 *         is less than, equal to, or greater than the specified object.
	 */
	public int compareTo(Object o) {
		if (o instanceof ConcreteContour) {
			return _label.compareTo(((ConcreteContour)o)._label);
		} else {
			return -1;
		}
	}

	/**
	 * @return true iff this concrete contour's abstract contour is derived.
	 */
	/*
	 * test:
	 *   abstract contour is derived
	 *   abstract contour is not derived
	 */
	public boolean isDerived() {
		return Contour.getContour(getLabel()).isDerived();
	}
    
	/**
	 * @return a string representation of the object
	 */
	/*
	 * test: once thru
	 */
	public String toString() { 
		return ""+getLabel()+"-"+this.hashCode();
	}

	/**
	 * @return a deep clone of this concrete contour
	 */
	/*
	 * test: once thru
	 */
	public DiagramItem getClone() {
		Polygon newP = new Polygon(_polygon.xpoints, _polygon.ypoints, _polygon.npoints);
		ConcreteContour res = new ConcreteContour(getContour(),newP);
		// Set attributes
		res.setAttributes(this);
		return res;
	}
	
	
	/**
	 * 
	 * @return a user-friendly description of this item
	 */
	/*
	 * test: once thru
	 */
	public String id() {
		return "contour " + getLabel();
	}
	

	// protected methods
	// -----------------
	
	protected Polygon makeBorder(Polygon poly) {
		Rectangle2D rect = poly.getBounds2D();
		int minX = (int)rect.getMinX();
		int maxX = (int)rect.getMaxX();
		int minY = (int)rect.getMinY();
		int maxY = (int)rect.getMaxY();
		int xs[] = {
			minX, maxX, maxX, minX
		};
		int ys[] = {
			minY, minY, maxY, maxY 
		};
		return new Polygon(xs, ys, 4);
	}

	protected Polygon makePolygon(Polygon border, int numPoints) {
		Rectangle2D rect = border.getBounds2D();
		double minX = rect.getMinX();
		double maxX = rect.getMaxX();
		double minY = rect.getMinY();
		double maxY = rect.getMaxY();
		double x0 = (minX+maxX)/2;
		double y0 = (minY+maxY)/2;
		double a = (maxX-minX)/2;
		double b = (maxY-minY)/2;
		Polygon newPoly = makeEllipse(x0, y0, a, b, numPoints);
		return newPoly;
	}
	
	protected Polygon makeEllipse(double x0, double y0, double a, double b, int n) {
		int xs[] = new int[n];
		int ys[] = new int[n];
		double teta = 0;
		double step = 2*Math.PI/n;
		for(int i=0; i<n; i++) {
			double x = x0 + a*Math.cos(teta);
			double y = y0 + b*Math.sin(teta);
			xs[i] = (int)x;
			ys[i] = (int)y; 
			teta += step;
		}
		Polygon newPoly = new Polygon(xs, ys, n);
		return newPoly;
	}
	
	protected Shape makeEllipse(Polygon border) {
		Rectangle2D rect = border.getBounds2D();
		double minX = rect.getMinX();
		double maxX = rect.getMaxX();
		double minY = rect.getMinY();
		double maxY = rect.getMaxY();
		double x0 = (minX+maxX)/2;
		double y0 = (minY+maxY)/2;
		double a = (maxX-minX)/2;
		double b = (maxY-minY)/2;
		return new Ellipse2D.Double(x0-a, y0-b, a*2, b*2);		
	}

	protected int getTopLeft() {
		return 0;
	}
	protected int getTopRight() {
		return 1;
	}
	protected int getBottomRight() {
		return 2;
	}
	protected int getBottomLeft() {
		return 3;
	}
	
	//
	// 	Local properties
	//  

	public Polygon _border;
	protected int _numPoints;
	protected Polygon _polygon;
	
	/*
	// test program
    // ------------

    private static void pr(String s) {
      System.out.println(s);
    }

    private static void hd(String s) {
      System.out.println("\n" + s + "\n");
    }

    private static void pg() {
      try {
        int i;
        System.out.print("next page? (CR)");
        do i = System.in.read(); while (i != '\n');
        System.out.println();
      }
      catch (Exception e) {System.out.println("raised exception : " + e);}
    }

    public static void main(String args[]) throws Exception {
    	hd("Create new concrete contour");
    	Polygon p = new Polygon();
    	p.addPoint(1,2);
    	ConcreteContour cc = 
    		new ConcreteContour(Contour.getContour("a"),p);
    	pr("cc="+cc);
    	pr("cc._label="+cc._label);
    	pr("cc._labelX="+cc._labelX);
    	pr("cc._labelY="+cc._labelY);    	
    	pg();
    	
    	ConcreteContour ccd = 
    		new ConcreteContour(Contour.getContour(""),p);
    	pr("ccd="+ccd);
    	pr("ccd._label="+ccd._label);
    	pr("ccd._labelX="+ccd._labelX);
    	pr("ccd._labelY="+ccd._labelY);    	
    	pg();
    	
    	{
    		hd("Label tests");
			pr("cc.getContour()="+cc.getContour());
		    pr("cc.isDerived()="+cc.isDerived());
			pg();
	 
			pr("cc.setContour(Contour.getContour(\"b\"))");
		    cc.setContour(Contour.getContour("b"));
		    pr("cc="+cc);
		    pr("cc.isDerived()="+cc.isDerived());
		    pg();
		    
		
		    pr("cc.setLabel(\"\")");
		    cc.setLabel("");
		    pr("cc="+cc);
		    pr("cc.isDerived()="+cc.isDerived());
		    pg();
	
		    pr("cc.setLabel(\"\")");
		    cc.setLabel("");
		    pr("cc="+cc);
		    pr("cc.isDerived()="+cc.isDerived());
		    pg();
	
		    pr("cc.setLabel(\"c\")");
		    cc.setLabel("c");
		    pr("cc="+cc);
		    pr("cc.isDerived()="+cc.isDerived());
		    pg();
	
		    pr("cc.setLabel(\"d\")");
		    cc.setLabel("d");
		    pr("cc="+cc);
		    pr("cc.isDerived()="+cc.isDerived());
		    pg();
	
		    pr("cc.setLabel(\"derived-e\")");
		    cc.setLabel("derived-e");
		    pr("cc="+cc);
		    pr("cc.isDerived()="+cc.isDerived());
		    pg();
    	}
    	
    	{
		    hd("Test of getClone");
		    pr("cc="+cc); 
		    pr("cc.getLabel()="+cc.getLabel()); 
		    pr("cc.getPolygon()="+cc.getPolygon());
		    Polygon pol = cc.getPolygon();
		    for (int i=0;i<pol.npoints;i++){
		    	pr("  x="+pol.xpoints[i]+",y="+pol.ypoints[i]);
		    }
		    pg();
		    
		    ConcreteContour cc2 = (ConcreteContour) cc.getClone();
		    pr("cc2="+cc2); 
		    pr("cc2.getLabel()="+cc2.getLabel()); 
		    pr("cc2.getPolygon()="+cc2.getPolygon());
		    pol = cc.getPolygon();
		    for (int i=0;i<pol.npoints;i++){
		    	pr("  x="+pol.xpoints[i]+",y="+pol.ypoints[i]);
		    }
		    pg();
    	}
	   
    	
    	{
    		hd("Test of equals");
        	
        	Object o1 = new Object();
        	pr("o1="+o1);
        	pr("cc.equals(o1)="+cc.equals(o1));
        	pg();
        	
        	pr("cc.getClone()="+cc.getClone());
        	pr("cc.equals(cc.getClone())="+cc.equals(cc.getClone()));
        	pg();
        	
     	}
    	
    	{
    		hd("Test of id");
    		pr("cc.id="+cc.id());
    		pg();
    	}
    	
    }
    */
}
