package rwd.concretesyntax.eulerdiagram;

import java.awt.Color;
//import java.awt.Polygon;
import java.awt.Shape;
import java.util.Iterator;
import java.util.Vector;
import java.util.LinkedHashSet;
import java.util.Set;

import rwd.abstractsyntax.eulerdiagram.Contour;
import rwd.abstractsyntax.eulerdiagram.Zone;
import rwd.concretesyntax.DiagramItem;
import rwd.metrics.AreaPolygonAdapter;
import rwd.utilities.AreaRWD;

/**
 * This class represents a zone as shown on a concrete diagram.
 * Different concrete diagrams will hold unique ConcreteZone objects
 * so these can hold information about e.g. shading
 * 
 * This class implements DiagramItem and Serializable so that diagrams can 
 * be shown and saved from the Editor
 * 
 * A ConcreteZone has an Area attribute, and two collections of 
 * Polygonal objects (at the moment, ConcreteContour and DecorationBox)
 * 
 * Because flyweight objects are not successfully Serializable, the 
 * link to the corresponding abstract zone is transient.
 * 
 * @author jaf28,rnb
 * 
 * @version 8.8.05   17:05
 */
public class ConcreteZone 
	extends DiagramItem 
	implements java.io.Serializable 
{
    private transient AreaRWD area = null;
    private Vector containingPolygonals;
    private Vector excludingPolygonals;
    protected transient Zone abstractZone = null;

    /**
     * Constructor for ConcreteZone.
     */
    public ConcreteZone(
        	Vector containingPolygonals,
        	Vector excludingPolygonals) 
    {
        setAttributes(containingPolygonals, excludingPolygonals);
        refreshArea(false);
    }

    // attributes
    // ----------
    
    public Vector getContainingPolygonals() {
        return containingPolygonals;
    }

    public Vector getExcludingPolygonals() {
        return excludingPolygonals;
    }
    
	public Vector getAllPolygonals() {
		Set result = new LinkedHashSet();
		result.addAll(getContainingPolygonals());
		result.addAll(getExcludingPolygonals());
		return new Vector(result);
	}

	public boolean isOuter() {
		if (containingPolygonals.size() == 1) {
			Polygonal p = (Polygonal) containingPolygonals.get(0);
			if (p instanceof DecorationBox)
				return true;
		}
		return false;
	}

	public Shape getShape(boolean smooth) {
		refreshArea(smooth);
		return area;
	}
    public Shape getShape() {
        return getShape(false);
    }

	public AreaRWD getArea() {
		if (area == null)
			refreshArea(false);
		return area;
	}
    public AreaRWD getArea(boolean smooth) {
    	refreshArea(smooth);
        return area;
    }

	public Set getHandles() {
		// Set handles
		java.util.Set _handles = new java.util.LinkedHashSet();
		return _handles;
	}

    public Zone getAbstractZone() {
        setAttributes(this.containingPolygonals, this.excludingPolygonals);
        return abstractZone;
    }

    public boolean isShaded() {
        Color col = getBackColor();
        return !col.equals(Color.white);
    }

    public String toString() {
        String result = "inside:";
        for (int i = 0; i < containingPolygonals.size(); i++) {
            result =
                result
                    + ((Polygonal) containingPolygonals.get(i)).toString()
                    + "_";
        }
        result = result + "outside:";
        for (int i = 0; i < excludingPolygonals.size(); i++) {
            result =
                result
                    + ((Polygonal) excludingPolygonals.get(i)).toString()
                    + "_";
        }
        result = result + "colour:" + this.getBackColor();
        return result;
    }

	/**
	 * 
	 * @return a user-friendly description of this item
	 */
    /*
     * test: 
     *   zone with no polygonals
     *   zone with two containing polygonals and two excluding polygonals
     */
	public String id() {
		String s =  "zone inside ";
		if (this.getContainingPolygonals().isEmpty()) s += "nothing";
		else {
			Iterator cit = this.getContainingPolygonals().iterator();
			while (cit.hasNext()) {
				s += ((Polygonal) cit.next()).getLabel();
				if (cit.hasNext()) s+= ",";
			}
		}
		s += " and outside ";
		if (this.getExcludingPolygonals().isEmpty()) s += "nothing";
		else {
			Iterator eit = this.getExcludingPolygonals().iterator();
			while (eit.hasNext()) {
				s += ((Polygonal) eit.next()).getLabel();
				if (eit.hasNext()) s+= ",";
			}
		}
		return s;
	}
	
    public boolean isSubZone(ConcreteZone other) {
        // Same set of contours
        Vector all = containingPolygonals;
        all.addAll(excludingPolygonals);
        Vector otherAll = other.containingPolygonals;
        all.addAll(other.excludingPolygonals);
        // all contains otherAll
        if (all.containsAll(otherAll)) {
            // otherAll contains all
            if (otherAll.containsAll(all)) {
                // Same set of contours
                return containingPolygonals.containsAll(
                    other.containingPolygonals);
                // 
            } else {
                return false;
            }
            // otherAll includes otherAll 
        } else if (all.containsAll(otherAll)) {
            return false;
        } else {
            return false;
        }
    }

    
    // operations
    // ----------
    
    public void refreshArea(boolean smoothed) {
        //System.out.println("calculating area for "+this);
        if (containingPolygonals.size() == 0) {
            //    throw new IllegalArgumentException();
            //System.out.println("not bounded - quit");
            return;
        }
        area =
            new AreaRWD(
                smoothed
                    ? AreaPolygonAdapter.getSmoothedShape(
                        	((Polygonal) (containingPolygonals.get(0)))
                            .getPolygon(),
        					AreaPolygonAdapter.BEZIER_CONTROL_DISTANCE)
                    : ((Polygonal) (containingPolygonals.get(0))).getPolygon());
        for (int i = 1; i < containingPolygonals.size(); i++) {
            area.intersect(
                new AreaRWD(
                    smoothed
                        ? AreaPolygonAdapter.getSmoothedShape(
                            ((Polygonal) (containingPolygonals.get(i)))
                                .getPolygon(),
                            AreaPolygonAdapter.BEZIER_CONTROL_DISTANCE)
                        : ((Polygonal) (containingPolygonals.get(i)))
                            .getPolygon()));
        }
        for (int i = 0; i < excludingPolygonals.size(); i++) {
            Object o = excludingPolygonals.get(i);
            //System.out.println("o is "+o.getClass());
            area.subtract(
                new AreaRWD(
                    smoothed
                        ? AreaPolygonAdapter.getSmoothedShape(
                            ((Polygonal) o).getPolygon(),
                            AreaPolygonAdapter.BEZIER_CONTROL_DISTANCE)
                        : ((Polygonal) o).getPolygon()));
        }
    }
    
	public void translateTo(float x, float y) {
		setLocation(x, y);
	}

    public DiagramItem getClone() {
    	Vector newCP = makeClone(containingPolygonals);
    	Vector newEP = makeClone(excludingPolygonals);
    	ConcreteZone res = new ConcreteZone(newCP, newEP);
    	res.setAttributes(this);
    	return res;
    }
    protected Vector makeClone(Vector list) {
    	Vector newList = new Vector();
    	for(int i=0; i<list.size(); i++) {
    		Object obj = list.get(i);
    		if (obj instanceof ConcreteContour) {
    			newList.add(((ConcreteContour)obj).getClone());
    		} else if (obj instanceof DecorationBox) {
				newList.add(((DecorationBox)obj).getClone());
    		}
    	}
    	return newList;
    }
    

    // private methods
    // ---------------
    
    private void setAttributes(
            Vector containingPolygonals,
            Vector excludingPolygonals) {
            this.containingPolygonals = containingPolygonals;
            this.excludingPolygonals = excludingPolygonals;

            // work out abstract zone
            Vector containingContours = new Vector();
            for (int i = 0; i < containingPolygonals.size(); i++) {
                Polygonal p = (Polygonal) containingPolygonals.get(i);
                //System.out.println("containing polygonal is "+p);
                if (p instanceof ConcreteContour) {
                    Contour c = ((ConcreteContour) p).getContour();
                    if (!containingContours.contains(c)) {
                        //System.out.println("add containing polygonal is "+p);
                        containingContours.add(c);
                    } else {
                        abstractZone = null;
                        return;
                    }
                }
            }
            Vector excludingContours = new Vector();
            for (int i = 0; i < excludingPolygonals.size(); i++) {
                Polygonal p = (Polygonal) excludingPolygonals.get(i);
                //System.out.println("excluding polygonal is "+p);
                if (p instanceof ConcreteContour) {
                    //System.out.println("add containing polygonal is "+p);
                    Contour c = ((ConcreteContour) p).getContour();
                    if (!excludingContours.contains(c)) {
                        //System.out.println("add containing polygonal is "+p);
                        excludingContours.add(c);
                    } else {
                        abstractZone = null;
                        return;
                    }
                }
            }
            abstractZone = Zone.getZone(containingContours, excludingContours);
            //System.out.println("abstract zone is "+abstractZone);

        }

        
    /*
    // test program
    // ------------

    private static void pr(String s) {
      System.out.println(s);
    }

    private static void hd(String s) {
      System.out.println("\n" + s + "\n");
    }

    private static void pg() {
      try {
        int i;
        System.out.print("next page? (CR)");
        do i = System.in.read(); while (i != '\n');
        System.out.println();
      }
      catch (Exception e) {System.out.println("raised exception : " + e);}
    }

    public static void main(String args[]) throws Exception {
    	
    	{
    		hd("Tests of hashCode, equals, id");
    		Vector cont0 = new Vector();
    		Vector excl0 = new Vector();
    		ConcreteZone cz1 = new ConcreteZone(cont0,excl0);
    		ConcreteZone cz2 = new ConcreteZone(cont0,excl0);
    		pr("cz1="+cz1);
    		pr("cz1.hashCode()="+cz1.hashCode());
    		pr("cz1.id()="+cz1.id());
    		pr("cz2="+cz2);
    		pr("cz2.hashCode()="+cz2.hashCode());
    		pr("cz2.id()="+cz2.id());
    		pr("cz1.equals(cz2)="+cz1.equals(cz2));
    		pg();
    		
    		Polygon p1 = new Polygon();
    		p1.addPoint(1,1);
    		Contour a = Contour.getContour("a");
    		ConcreteContour c1 = new ConcreteContour(a,p1);
    		Contour b = Contour.getContour("b");
    		ConcreteContour c2 = new ConcreteContour(b,p1);
    		Contour c = Contour.getContour("c");
    		ConcreteContour c3 = new ConcreteContour(c,p1);
    		Contour d = Contour.getContour("d");
    		ConcreteContour c4 = new ConcreteContour(d,p1);
    		
    		Vector cont1 = new Vector();
    		Vector excl1 = new Vector();
    		cont1.add(c1);
    		cont1.add(c2);
    		excl1.add(c3);
    		excl1.add(c4);
    		ConcreteZone cz3 = new ConcreteZone(cont1,excl1);
    		pr("cz3="+cz3);
    		pr("cz3.hashCode()="+cz3.hashCode());
    		pr("cz3.id()="+cz3.id());
    		pg();
    		
    		pr("cz1.equals(null)="+cz1.equals(null));
    		pg();
    		
    		pr("cz1.equals(cont0)="+cz1.equals(cont0));
    		pg();
    		
    		pr("cz1="+cz1);
    		pr("cz3="+cz3);
    		pr("cz1.equals(cz3)="+cz1.equals(cz3));
    		pg();    		
    		
    	}
    	
    }
    */
}
