package rwd.gui.editor;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.util.*;
import java.util.List;

import javax.swing.JComponent;

import org.jibble.epsgraphics.EpsGraphics2D;

//import rwd.abstractsyntax.eulerdiagram.Contour;
import rwd.abstractsyntax.eulerdiagram.Zone;
import rwd.concretesyntax.*;
import rwd.concretesyntax.eulerdiagram.*;
import rwd.concretesyntax.spiderdiagram.*;
import rwd.concretesyntax.constraintdiagram.*;
import rwd.gui.editor.listeners.*;
import rwd.gui.editor.listeners.diagramcontrollers.ConstraintDiagramController;
import rwd.gui.editor.listeners.diagramcontrollers.DiagramControllerState;
import rwd.gui.editor.viewers.*;
import rwd.utilities.AreaRWD;

/**
 * A component to display the diagram.  A separate image is stored
 * and painted to make animation smoother.  The user may scroll around
 * the diagram using the mouse and zoom in by using the mousewheel.
 * 
 * The diagram is translated and scaled before being displayed.
 * Canvas coordinates refer to the x,y position relative to the NW corner 
 * of the canvas.
 * Diagram coordinates refer to the x,y position relative to the NW corner 
 * of the diagram.
 * Mouse-coordinates from a mouse-event are canvas coordinates,
 * and must be transformed to diagram coordinates before being applied 
 * to the diagram.
 * Diagram item coordinates are diagram coordinates, and must be 
 * transformed to canvas coordinates before the item is displayed. 
 * 
 * @author  op,rnb
 * @version 16.8.05   16:15
 */
public class DiagramCanvas
	extends JComponent
	implements MouseMotionListener, MouseListener, MouseWheelListener, KeyListener,
	           DiagramControllerState
{

	//
	// Constructors
	//
	public DiagramCanvas(DiagramViewerInterface viewer,
		                 boolean justForViewing) {
		_viewer = viewer;

		_popup = new EditPopupMenu(_viewer);
		setBackground(Color.white);
		setOpaque(true);

		// Add component listener
		this.addComponentListener(new ComponentAdapter() {
			public void componentResized(ComponentEvent ce) {
				scrapImage();
				drawImageNow();
			}
		});

		// Add mouse and key listeners
		if (!justForViewing) {
			ConstraintDiagramController cdc = new ConstraintDiagramController(this);
			cdc.setState(this);
			addMouseMotionListener(cdc);
			addMouseListener(cdc);
			addKeyListener(cdc);
			//addMouseMotionListener(this);
			//addMouseListener(this);
			addMouseWheelListener(this);
			//addKeyListener(this);
		}
	}

	public DiagramCanvas(DiagramViewerInterface viewer) {
		this(viewer, false);
	}

	
	//
	// Drawing methods
	//
	public synchronized void scrapImage() {
		_image = null;
	}

	public synchronized void initImage() {
		if (_image == null) {
			try {
				_image = this.createImage(getWidth(), getHeight());
			} catch (IllegalArgumentException e) {
				System.out.println("image not ready");
				e.printStackTrace();
			}
			if (_image != null) {
				//System.out.println("initing image"+hashCode());;
				_g = (Graphics2D) _image.getGraphics();
				_g.setRenderingHint(
					RenderingHints.KEY_ANTIALIASING,
					RenderingHints.VALUE_ANTIALIAS_ON);
				_g.setColor(Color.white);
				_g.fillRect(0, 0, getWidth(), getHeight());
				_g.setStroke(STROKE);
				rescaleTransform();
				repaint();
			}
		}
	}

	// Paint only the affected area by painting the image over that area.
	protected synchronized void paintComponent(Graphics g) {
		initImage();
		Rectangle rect = g.getClipBounds();
		int x = (int) rect.getX();
		int y = (int) rect.getY();
		int width = (int) rect.getWidth();
		int height = (int) rect.getHeight();
		g.drawImage(
			_image,
			x,
			y,
			x + width,
			y + height,
			x,
			y,
			x + width,
			y + height,
			null);
	}

	// Paint all of the shapes we have onto the image.
	protected synchronized void repaintImage() {
		try {
			//System.out.println("repainting image"+hashCode()+" about to init:");
			initImage();
			//System.out.println("repainting init done "+hashCode());
			if (_g != null) {
				_g.setTransform(_transform);
				_g.setColor(Color.white);
				_g.fillRect(
					(int) - _xTranslate,
					(int) - _yTranslate,
					(int) (getWidth() / _scale),
					(int) (getHeight() / _scale));
				_g.setColor(Color.black);
				_g.setStroke(STROKE);
				ConcreteDiagram cd = _viewer.getConcreteDiagram();
				if (cd != null) {
					cd.refreshZones(_smoothed);
					synchronized (cd) {
						// Draw shaded zones
						Set zoneAreas = cd.getVisibleConcreteZones();
						for (Iterator i = zoneAreas.iterator(); i.hasNext();) {
							ConcreteZone cz = (ConcreteZone) i.next();
							// Draw area
							_g.setColor(cz.getBackColor());
							Shape s = cz.getArea(_smoothed);
							if (s != null) {
								_g.fill(s);
							}
						}

						// Highlight the zones with contradictions in tableau
						Set contZones = _viewer.getContradictionZones();
						for (Iterator i = contZones.iterator(); i.hasNext();) {
							ConcreteZone za = (ConcreteZone) i.next();
							Zone az = za.getAbstractZone();
							if (contZones.contains(az)) {
								BasicStroke stroke = new BasicStroke(
										za.getThickness() + 50,
										BasicStroke.CAP_BUTT,
										BasicStroke.JOIN_ROUND);
								_g.setStroke(stroke);
								Shape shape = za.getShape(_smoothed);
								_g.setColor(Color.red);
								_g.draw(shape);
								_g.setColor(za.getBackColor());
								Shape s = za.getArea();
								_g.fill(s);
							}
						}

						// Draw contours
						Set contours = cd.getConcreteContours();
						Iterator ii = contours.iterator();
						while (ii.hasNext()) {
							ConcreteContour cp = (ConcreteContour) ii.next();
							synchronized(cp) {
								// Draw shape
								cp.draw(_g, _smoothed);
							}
						}

						// Draw spider legs
						Vector spiderLegs = cd.getFeetAndLegsSet().getLegs();
						for (int i = 0; i < spiderLegs.size(); i++) {
							ConcreteSpiderLeg leg =	(ConcreteSpiderLeg) spiderLegs.get(i);
							leg.draw(_g, _smoothed);
						}

						// Draw spider feet
						java.util.List spiderFeet =	cd.getFeetAndLegsSet().getFeet();
						for (int i = 0; i < spiderFeet.size(); i++) {
							ConcreteSpiderFoot foot = (ConcreteSpiderFoot) spiderFeet.get(i);
							foot.draw(_g, _smoothed);
						}

						// Draw boxes
						Vector boxes = cd.getDecorationSet().getBoxes();
						for (int i = 0; i < boxes.size(); i++) {
							DecorationBox box = (DecorationBox) boxes.get(i);
							box.draw(_g, _smoothed);
						}

						// Draw arcs
						Vector arcs = cd.getArcSet().getArcs();
						for (int i = 0; i < arcs.size(); i++) {
							ConcreteArc arc = (ConcreteArc) arcs.get(i);
							arc.draw(_g, _smoothed);
						}

						// Draw texts
						Vector texts = cd.getTextSet().getTexts();
						for (int i = 0; i < texts.size(); i++) {
							ConcreteTextBox textBox = (ConcreteTextBox) texts.get(i);
							textBox.draw(_g, _smoothed);
						}

						// Draw selected items
						if (_selectedItems.size() != 0) {
							// Draw selected items
							Iterator i = _selectedItems.iterator();
							while (i.hasNext()) {
								DiagramItem selectedItem =
									(DiagramItem) i.next();
								// Set stroke
								BasicStroke stroke = new BasicStroke(
										selectedItem.getThickness()+8,
										BasicStroke.CAP_BUTT,
										BasicStroke.JOIN_ROUND);
								_g.setStroke(stroke);
								// Draw shape
								Shape shape = selectedItem.getShape(_smoothed);
								_g.setColor(Color.magenta);
								_g.draw(shape);
								// Draw label
								if (selectedItem instanceof ConcreteContour) {
									Point2D labelPosition =
										selectedItem.getLabelPosition();
									if (selectedItem.getLabel() != null) {
										_g.drawString(
											selectedItem.getLabel(),
											(float) labelPosition.getX(),
											(float) labelPosition.getY());
									}
								}
							}

							// Draw handles
							_g.setStroke(STROKE);
							Set handles = computeHandles(_selectedItems);
							Iterator j = handles.iterator();
							while (j.hasNext()) {
								Object obj = j.next();
								Handle handle = (Handle) obj;
								handle.draw(_g, _smoothed);
							}
						}

						// Draw resizing contours
						if (_newShape != null) {
							_g.draw(_newShape);
						}

						// Draw new edges.
						if (_leg != null) {
							_g.setColor(new Color(255, 0, 0, 153));
							_g.setStroke(DOT);
							_g.draw(_leg.getShape());
						}

					}
				}
			}
		} catch (Exception e) {
			System.out.println("Exception in canvas's repaint");
			e.printStackTrace();
		}
	}

	// Draws the contents of the image onto the component immediately,
	// without waiting for the AWT thread to catch up.
	public synchronized void drawImageNow() {
        final Graphics g = this.getGraphics();
		try {
			paint(g);
		} catch (NullPointerException x) {
            System.out.println("graphics not ready to paint");
			_waitingToBeDrawn = true;
            // Question: when does this get a graphics object?
			// that's the time to call this method.
		}
	}

    public synchronized void paint(Graphics g) {
		if (_waitingToBeDrawn) {
			System.out.println("paint(g) called");
			_waitingToBeDrawn = false;
		}
		repaintImage();
		super.paint(g);
	}

	public synchronized void drawString(String label, float x, float y) {
		_g.setColor(Color.black);
		_g.drawString(label, x, y);
	}

	
	//
	// Mouse events
	//
	// The mouse has entered the canvas
	public synchronized void mouseEntered(MouseEvent e) {
		e.consume();
		Cursor c = new Cursor(Cursor.CROSSHAIR_CURSOR);
		setCursor(c);
	}

	// Press mouse
	/**
	 * @param e mouse pressed event
	 */
	/*
	 * pre:  _currentX,_currentY = mouse position in float coordinates 
	 *                               before this event
	 *       _translateX,_translateY = translation x and y before this event
	 * post: _currentX,_currentY =  mouse position in float coordinates
	 *       _previousX,_previousY = mouse position in float coordinates 
	 *       if shift is down,
	 *         if spider foot is selected,
	 *           spider leg has been added
	 *       if shift is not down,
	 *         _translating = true
	 */
	public synchronized void mousePressed(MouseEvent e) {
		requestFocus();
		e.consume();

		// Compute mouse coordinates
		_currentX = computeCurrentX(e);
		_currentY = computeCurrentY(e);

		// Store (X, Y) for future dragging
		_previousX = _currentX;
		_previousY = _currentY;

		// Prepare for future events - dragging
		if (e.isShiftDown()) {
			// Add a leg between two feet if a spider foot is selected
			if (_selectedItems.size() == 1) {
				// Compute the source foot
				DiagramItem selectedItem =
					(DiagramItem) _selectedItems.iterator().next();
				if (selectedItem instanceof ConcreteSpiderFoot) {
					_sourceFoot = (ConcreteSpiderFoot) selectedItem;
				} else {
					//					// Resize something
					//					if (selectedItem instanceof ConcreteContour
					//						|| selectedItem instanceof DecorationBox
					//						|| selectedItem instanceof ConcreteTextBox) {
					//						// Store parameters for future resizing
					//						_originalSize = selectedItem.getShape().getBounds2D();
					//						_originalLabelPosition = selectedItem.getLabelPosition();
					//					}
				}
			}
		// Dragging the elements
		} else {
			if (isInItems(_selectedItems,_currentX,_currentY))
			{
				_translating = true;
			}
		}
	}

	// Release mouse
	/**
	 * @param e = mouseReleased event
	 */
	/*
	 * pre:  _translating = true, iff items have been translated
	 * 	     _previousX,previousY  = mouse position before this event
	 * post: _translating = false
	 */
	public synchronized void mouseReleased(MouseEvent e) {
		e.consume();

		// Compute mouse coordinates
		_currentX = computeCurrentX(e);
		_currentY = computeCurrentY(e);

		// If right click => launch the context menu 
		if (e.isPopupTrigger()) {
			_popup.show(e.getComponent(), e.getX(), e.getY());
			_popup.validateItems();
			return;
		}

		// If add leg => add it
		if (_leg != null) {
			_viewer.add(_leg);
			_sourceFoot = null;
			_targetFoot = null;
			_leg = null;
			return;
		}

		// If translating => end of translation
		if (_translating) {
			_translating = false;
		}

		// Drag handle => 
		if (_handleMoved) {
			_handleMoved = false;
			DiagramItem item = _handle.getItem();
			// Spider foot
			if (item instanceof ConcreteSpiderFoot) {
				// Update location of corresponding arc
				if (item.isAttached()) {
					_viewer.updateArcsLocation();
				}
				// Notify observers
				item.notifyObservers(rwd.concretesyntax.Event.CHANGE_ITEM);
			// Arc end
			} else if (item instanceof ConcreteArcEnd) {
				// Attach arc end if not middle
				ConcreteArcEnd end = (ConcreteArcEnd) item;
				if (!end.isMiddle()) {
					_viewer.attachArcEnd((ConcreteArcEnd) item);
					// Notify observers
					item.notifyObservers(rwd.concretesyntax.Event.CHANGE_ITEM);
				}
			// Resize a box
			} else if (item instanceof DecorationBox) {
				DecorationBox box = (DecorationBox) item;
				box.resize(((ContourHandle) _handle).getPoint(),_previousX,_currentX,_previousY,_currentY);
				// Notify observers
				item.notifyObservers(rwd.concretesyntax.Event.CHANGE_ITEM);
			// Resize a contour
			} else if (item instanceof ConcreteContour) {
				ConcreteContour c = (ConcreteContour) item;
				ContourHandle ch = (ContourHandle)_handle; 
				int point = ch.getPoint();
				Polygon poly = c.getPolygon();
				// Resize contour
				if (ch.isBorder()) {
					c.resize(point,_previousX,_currentX,_previousY,_currentY);
				// Move point
				} else {
					c.movePoint(point, _previousX, _currentX, _previousY, _currentY);
				}
				// Notify observers
				item.notifyObservers(rwd.concretesyntax.Event.CHANGE_ITEM);
			// else - Contour handle ?
			} else {
				// Refresh zones
				ConcreteDiagram cd = _viewer.getConcreteDiagram();
				cd.refreshZones(_viewer.isBezier());
				cd.notifyObservers();
			}
			_viewer.store();
			return;
		}

		// Invalidate old dragged objects
		invalidateDraggedObjects();

		repaintImage();
		repaint();
	}

	// The mouse has been clicked (pressed and released) in canvas
	public synchronized void mouseClicked(MouseEvent e) {
		e.consume();
		
		// If right click => launch the context menu 
		if (e.isPopupTrigger()) {
			_popup.show(e.getComponent(), e.getX(), e.getY());
			_popup.validateItems();
			return;
		}

		// Compute mouse coordinates		
		_currentX = computeCurrentX(e);
		_currentY = computeCurrentY(e);

		// Get command
		String cmd = _viewer.getControlPanel().getCmd();
		// Select
		if (cmd == ControlPanel.SELECT) {
			if (e.getButton() == MouseEvent.BUTTON1)
				select(_currentX, _currentY, e.isControlDown());
		// New box
		} else if (cmd == ControlPanel.BOX) {
			(new DrawNewBoxListener(_viewer)).addNewBox(
				(int) _currentX,
				(int) _currentY);
		// New contour
		} else if (cmd == ControlPanel.CONTOUR) {
			(new DrawNewContourListener(_viewer)).addNewContour(
				(int) _currentX,
				(int) _currentY);
		// New foot
		} else if (cmd == ControlPanel.FOOT) {
			(new DrawNewFootListener(
					_viewer,
					ConcreteSpiderFoot.NORMAL)).addFoot(
				ConcreteSpiderFoot.NORMAL,
				(int) _currentX,
				(int) _currentY);
		// New universal spider
		} else if (cmd == ControlPanel.USPIDER) {
			(new DrawNewFootListener(
					_viewer,
					ConcreteSpiderFoot.UNIVERSAL)).addFoot(
				ConcreteSpiderFoot.UNIVERSAL,
				(int) _currentX,
				(int) _currentY);
		// Schrodinger
		} else if (cmd == ControlPanel.SSPIDER) {
			(
				new DrawNewFootListener(
					_viewer,
					ConcreteSpiderFoot.SCHROEDINGER)).addFoot(
				ConcreteSpiderFoot.SCHROEDINGER,
				(int) _currentX,
				(int) _currentY);
		// New leg
		} else if (cmd == ControlPanel.LEG) {
			(new DrawNewLegListener(_viewer, ConcreteSpiderLeg.NORMAL)).addLeg(
				ConcreteSpiderLeg.NORMAL,
				(int) _currentX,
				(int) _currentY);
		// New tie
		} else if (cmd == ControlPanel.TIE) {
			(new DrawNewLegListener(_viewer, ConcreteSpiderLeg.TIE)).addLeg(
				ConcreteSpiderLeg.TIE,
				(int) _currentX,
				(int) _currentY);
		// New strand
		} else if (cmd == ControlPanel.STRAND) {
			(new DrawNewLegListener(_viewer, ConcreteSpiderLeg.STRAND)).addLeg(
				ConcreteSpiderLeg.STRAND,
				(int) _currentX,
				(int) _currentY);
		// New arc
		} else if (cmd == ControlPanel.ARC) {
			(new DrawNewArcListener(_viewer, ConcreteArc.QUAD)).addNewArc(
				ConcreteArc.QUAD,
				(int) _currentX,
				(int) _currentY);
		// New line
		} else if (cmd == ControlPanel.LINE) {
			(new DrawNewArcListener(_viewer, ConcreteArc.LINE)).addNewArc(
				ConcreteArc.LINE,
				(int) _currentX,
				(int) _currentY);
		// New comment
		} else if (cmd == ControlPanel.TEXT) {
			(new DrawNewTextListener(_viewer,ConcreteTextBox.NORMAL)).addNewText(
				ConcreteTextBox.NORMAL,
				(int) _currentX,
				(int) _currentY);
		}

		// Reset default button
		_viewer.getControlPanel().setDefaultButton();

		// Repaint image
		repaintImage();
		repaint();
	}

	// The mouse has exited the canvas
	public void mouseExited(MouseEvent e) {
		e.consume();
		Cursor def = new Cursor(Cursor.DEFAULT_CURSOR);
		setCursor(def);
	}

	// Move mouse
	public synchronized void mouseMoved(MouseEvent e) {
		e.consume();
	}

	/**
	 * @param e mouseDragged event
	 */
	/*
	 * pre:  _previousX,_previousY = 
	 *          mouse position in diagram before this event
	 *       _selectedItems = currently selected items in canvas
	 * post: _currentX,_currentY = current mouse position in diagram
	 *       _previousX,_previousY = 
	 *          current mouse position in diagram
	 *       selected items have been translated to new mouse position
	 *         from previous mouse position
	 *       image has been redrawn
	 */
	public synchronized void mouseDragged(MouseEvent e) 
	{
		e.consume();

		// Compute mouse coordinates		
		_currentX = computeCurrentX(e);
		_currentY = computeCurrentY(e);

		// Making a leg
		if (_sourceFoot != null) {
			if (_sourceFoot.getType() == ConcreteSpiderFoot.NORMAL) {
				_targetFoot =
					_viewer
						.getConcreteDiagram()
						.getFeetAndLegsSet()
						.getNearestFoot(
						(int) _currentX,
						(int) _currentY);
				if (_targetFoot != null && _sourceFoot != _targetFoot) {
					int legType = getLegType();
					_leg =
						new ConcreteSpiderLeg(
							"",
							_sourceFoot,
							_targetFoot,
							legType);
				}
			}
		// Moving items or handles
		} else {
			// Try to move a label
			if (e.isControlDown()) {
				if (_selectedItems.size() == 1) {
					// Get selected element and move label
					DiagramItem selectedItem =
						(DiagramItem) _selectedItems.iterator().next();
					if (selectedItem.getLabelPosition() != null) {
						Point2D point = selectedItem.getLabelPosition();
						float dx = _currentX - _previousX;
						float dy = _currentY - _previousY;
						point.setLocation(point.getX() + dx, point.getY() + dy);
						selectedItem.setLabelPosition(point);
					}
				} else {
				}
			// Try to move a reading forest
			} else if (e.isShiftDown()) {
				if (_selectedItems.size() == 1) {
					// Get selected element and move label
					DiagramItem selectedItem =
						(DiagramItem) _selectedItems.iterator().next();
					if (selectedItem instanceof DecorationBox) {
						DecorationBox box = (DecorationBox)selectedItem;
						float dx = _currentX - _previousX;
						float dy = _currentY - _previousY;
						Point2D point = box.getReadingForestPosition();
						if (point != null) {
							point.setLocation(point.getX() + dx, point.getY() + dy);
							box.setReadingForestPosition(point);
						}
					}
				} else {
				}
			// Try to move a handle
			} else if (selectedHandle()) {
				if (_translating)
				{
					_handle = (Handle) _selectedItems.iterator().next();
					DiagramItem item = _handle.getItem();
					// Set new location
					float dx = _currentX - _previousX;
					float dy = _currentY - _previousY;
					_handle.setLocation(
						(int) (_handle.getX() + dx),
						(int) (_handle.getY() + dy));
					_handleMoved = true;
					// Handle belongs to a foot
					if (item instanceof ConcreteSpiderFoot) {
						item.setLocation(
							(int) (_handle.getX() + dx),
							(int) (_handle.getY() + dy));
						// Update location of corresponding arc
						if (item.isAttached()) {
							_viewer.updateArcsLocation();
						}
						// Notify observers
						item.notifyObservers(rwd.concretesyntax.Event.CHANGE_ITEM);
					// Handle belongs to an arc end 
					} else if (item instanceof ConcreteArcEnd) {
						item.setLocation((int) (_handle.getX() + dx), (int) (_handle.getY() + dy));
						// Detach arc end 
						((ConcreteArcEnd) item).detach();
						// Notify observers
						item.notifyObservers(rwd.concretesyntax.Event.CHANGE_ITEM);
					// Resize a box
					} else if (item instanceof DecorationBox) {
						DecorationBox box = (DecorationBox) item;
						box.resize(((ContourHandle) _handle).getPoint(),_previousX,_currentX,_previousY,_currentY);
						// Notify observers
						item.notifyObservers(rwd.concretesyntax.Event.CHANGE_ITEM);
					// Resize a contour or move contour points
					} else if (item instanceof ConcreteContour) {
						ConcreteContour c = (ConcreteContour) item;
						ContourHandle ch = (ContourHandle)_handle; 
						int point = ch.getPoint();
						Polygon poly = c.getPolygon();
						// Resize contour
						if (ch.isBorder()) {
							c.resize(point,_previousX,_currentX,_previousY,_currentY);
						// Move point
						} else {
							c.movePoint(point, _previousX, _currentX, _previousY, _currentY);
						}
						// Notify observers
						item.notifyObservers(rwd.concretesyntax.Event.CHANGE_ITEM);
					// else - Contour handle ?
					} else {
						// Refresh zones
						ConcreteDiagram cd = _viewer.getConcreteDiagram();
						cd.refreshZones(_viewer.isBezier());
						cd.notifyObservers();
					}
				}
			// Try to move selected items
			} else {
				if (_translating){
					float dx = _currentX - _previousX;
					float dy = _currentY - _previousY;
					Iterator i = _selectedItems.iterator();
					while (i.hasNext()) {
						DiagramItem item = (DiagramItem) i.next();
						if (! (item instanceof ConcreteZone))
							item.translateTo(dx, dy);
					}
				}
			}
		}

		// Repaint image
		//		  repaintImage();
		drawImageNow();

		_previousX = _currentX;
		_previousY = _currentY;
	}
	
	/*
	protected String toString(Polygon poly) {
		String res = "";
		for(int i=0; i<poly.npoints; i++) {
			res += "("+poly.xpoints[i]+","+poly.ypoints[i]+")";
		}
		return res;
	}
	*/

	// Move mouse wheel
	public synchronized void mouseWheelMoved(MouseWheelEvent e) {
		int clicks = e.getWheelRotation();
		updateZoom(clicks);
	}

	
	// 
	// Keyboard events
	// 
	public synchronized void keyPressed(KeyEvent e) {
		e.consume();
		int code = e.getKeyCode();
		// DELETE is pressed
		if (code == KeyEvent.VK_DELETE) {
			_viewer.deleteSelectedItems();
			// T is pressed
		} else if (code == KeyEvent.VK_T) {
			_viewer.changeType();
		}
	}

	public synchronized void keyTyped(KeyEvent e) {
	}

	public synchronized void keyReleased(KeyEvent e) {
	}

	//
	// Selection methods
	//
	
	/**
	 * Post: if mouse click hits an item, that item is selected (with provisos).
	 *       If addSelection, 
	 *         the selected item is added to the items already selected.
	 *       if addSelection,
	 *          no handle remains selected (including those selected 
	 *            before execution).
	 *       if not addSelection,
	 *         all items selected before execution are deselected (even if no 
	 *           new item is selected).
	 * @param x,y coords of mouse click
	 * @param addSelection true, iff user wishes to accumulate selected items 
	 *        (CNTRL key is pressed)
	 */
	/*
	 * pre:  _selectedItems = set of currently selected items 
	 * post: _selectedItems = set of currently selected items 
	 */
	/*
	 * tests:
	 *   not addSelection
	 *     select handle
	 *            foot
	 *            leg
	 *            arc
	 *            contour
	 *            box
	 *            text
	 *            zone
	 *   addSelection with a handle (or foot) already selected
	 *     select handle
	 *            foot
	 *            leg
	 *            arc
	 *            contour
	 *            box
	 *            text
	 *            zone
	 *            all of these
	 */
	public synchronized void select(float x, float y, boolean addSelection) {

		// Start by working out whether the user selected an item...
		int rectSize = SELECTION_SIZE;
		Rectangle2D rect =
			new Rectangle2D.Float(
				x - rectSize,
				y - rectSize,
				rectSize * 2,
				rectSize * 2);

		// find handles of current items
		java.util.Set handles = computeHandles(_selectedItems);

		// if not adding, remove current selections and check handles
		if (!addSelection) {
			deselect();
			Iterator ih = handles.iterator();
			while (ih.hasNext()) {
				Handle h = (Handle) ih.next();
				Shape shape = h.getShape();
				if (shape.intersects(rect)) {
					selectItem(h);
					return;
				}
			}
		}
		else {  // addSelection
			
			// remove any handle already there
			Iterator sit = _selectedItems.iterator();
			while (sit.hasNext()) {
				DiagramItem di = (DiagramItem) sit.next();
				if (di instanceof Handle)
					sit.remove();
			}
		}
			

		// Try feet
		java.util.List feet =
			_viewer.getConcreteDiagram().getFeetAndLegsSet().getFeet();
		for (int i = 0; i < feet.size(); i++) {
			ConcreteSpiderFoot foot = (ConcreteSpiderFoot) feet.get(i);
			Shape shape = foot.getShape();
			if (shape.intersects(rect)) {
				selectItem(foot);
				return;
			}
		}

		// Try legs
		Vector legs =
			_viewer.getConcreteDiagram().getFeetAndLegsSet().getLegs();
		for (int i = 0; i < legs.size(); i++) {
			ConcreteSpiderLeg leg = (ConcreteSpiderLeg) legs.get(i);
			Shape shape = leg.getShape();
			if (shape.intersects(rect)) {
				selectItem(leg);
				return;
			}
		}

		// Try arcs
		Vector arcs = _viewer.getConcreteDiagram().getArcSet().getArcs();
		for (int i = 0; i < arcs.size(); i++) {
			ConcreteArc arc = (ConcreteArc) arcs.get(i);
			//            ConcreteArcEnd start = arc.getStart();
			//            Shape shape = start.getShape();
			//            if (shape.intersects(rect)) {
			//                selectItem(start);
			//            }
			//            ConcreteArcEnd middle = arc.getMiddle();
			//            if (middle.getShape().intersects(rect)) {
			//                selectItem(middle);
			//            }
			//            ConcreteArcEnd end = arc.getEnd();
			//            if (end.getShape().intersects(rect)) {
			//                selectItem(end);
			//            }
			Shape shape = arc.getShape();
			if (shape.intersects(rect)) {
				selectItem(arc);
				return;
			}
		}

		// Try contours
		Set contours = _viewer.getConcreteDiagram().getConcreteContours();
		Iterator ii = contours.iterator();
		while (ii.hasNext()) {
			ConcreteContour cp = (ConcreteContour) ii.next();
			Polygon p = cp.getPolygon();
			p.invalidate();
			if (!p.contains(rect) && p.intersects(rect)) {
				selectItem(cp);
				return;
			}
		}

		// Try boxes
		Vector boxes =
			_viewer.getConcreteDiagram().getDecorationSet().getBoxes();
		for (int i = 0; i < boxes.size(); i++) {
			DecorationBox box = (DecorationBox) boxes.get(i);
			Polygon polygon = box.getPolygon();
			polygon.invalidate();
			if (!polygon.contains(rect) && polygon.intersects(rect)) {
				selectItem(box);
				return;
			}
		}

		// Try text
		Vector texts = _viewer.getConcreteDiagram().getTextSet().getTexts();
		for (int i = 0; i < texts.size(); i++) {
			ConcreteTextBox textBox = (ConcreteTextBox) texts.get(i);
			Polygon polygon = textBox.getPolygon();
			polygon.invalidate();
			if (polygon.contains(rect) || polygon.intersects(rect)) {
				selectItem(textBox);
				return;
			}
		}

		// Try zones
		Set zoneAreas = _viewer.getConcreteDiagram().getVisibleConcreteZones();
		//System.out.println("got mouse click");
		for (Iterator i = zoneAreas.iterator(); i.hasNext();) {
			ConcreteZone za = (ConcreteZone) i.next();
			//System.out.println("checking for mous click in "+za);
			AreaRWD a = za.getArea();
			//a.invalidate();
			if (a != null && a.contains(x, y)) {
				selectItem(za);
				return;
			}
		}
	}
	
	/**
	 * Post: item has been added to the set of selected items (with provisos).
	 * <UL>
	 *   <LI>if item was the handle of a spider foot, the foot has been selected.
	 *   <LI>if item is an arc end or a handle of an arc end, 
	 *         and the arc has already been selected,
	 *         the item is not added.
	 *   <LI>if item is an arc, 
	 *         and an associated arc end or arc end handle has already been selected,
	 *         the arc end (handle) item is removed.
	 *   <LI>if item is a spider foot or the handle of a spider foot, 
	 *         and a leg containing this foot has been selected,
	 *         the item is not added.
	 *   <LI>if item is a spider leg, 
	 *         and an associated spider foot or spider foot handle has been selected,
	 *         the spider foot (handle) is removed.
	 * </UL>
	 *       The toolbar has been updated to correspond to the selected items.
	 *       The canvas has been repainted to correspond to the selected items.
	 * @param selectedItem item to be selected.
	 */
	/*
	 * pre:  _selectedItems = currently selected items
	 * post: _selectedItems = currently selected items
	 *       _oldShape = polygon associated with contour, 
	 *                     if contour handle was selected
	 */
	/*
	 * test: 
	 *   item is handle
	 *     handle of contour
	 *     handle of spider foot
	 *       associated leg is not selected
	 *       associated leg is selected
	 *     handle of arc end
	 *       associated arc is not selected
	 *       associated arc is selected
	 *     handle of decoration box
	 *   item is spider foot
	 *     associated leg is not selected
	 *     associated leg is selected
	 *   item is arc end
	 *     associated arc is not selected
	 *     associated arc is selected
	 *   item is spider leg
	 *     associated foot (handle) is not selected
	 *     associated foot (handle) is selected
	 *   item is arc
	 *     associated arc end (handle) is not selected
	 *     associated arc end (handle) is selected
	 *   item is not handle or arc or arc end or foot 
	 */
	public synchronized void selectItem(DiagramItem selectedItem) {
		
		boolean addItem = true;
		
		// convert foot handle to foot
		if (selectedItem instanceof Handle) {
			DiagramItem item = ((Handle)selectedItem).getItem();
			if (item instanceof ConcreteSpiderFoot) {
				selectedItem = item;
			}			
		}
		
		if (selectedItem instanceof Handle) {
			// Store original size of a contour
			DiagramItem item = ((Handle)selectedItem).getItem();
			if (item instanceof ConcreteContour) {
				ConcreteContour c = (ConcreteContour)item;
				//				_originalSize = c.getPolygon().getBounds2D();
				_oldShape = c.getPolygon();
			}
		}
		
		if (selectedItem instanceof ConcreteArcEnd ||
				selectedItem instanceof Handle) {

			// check if associated arc already there
			Iterator sit = _selectedItems.iterator();
			while (sit.hasNext()) {
				DiagramItem di = (DiagramItem) sit.next();
				if (di instanceof ConcreteArc) {
					ConcreteArc ca = (ConcreteArc) di;
					if (isAssociated(ca,selectedItem)) {
						addItem = false; // if so, don't add the item
						break;
					}
				}
			}
		}
		
		if (selectedItem instanceof ConcreteArc) {
			ConcreteArc ca = (ConcreteArc) selectedItem;
			
			// remove any associated arc ends or handles
			Iterator sit = _selectedItems.iterator();
			while (sit.hasNext()) {
				DiagramItem di = (DiagramItem) sit.next();
				if (isAssociated(ca,di)) {
					sit.remove();
				}
			}
		}
			
		if (selectedItem instanceof ConcreteSpiderFoot ||
				selectedItem instanceof Handle) {

			// check if associated leg already there
			Iterator sit = _selectedItems.iterator();
			while (sit.hasNext()) {
				DiagramItem di = (DiagramItem) sit.next();
				if (di instanceof ConcreteSpiderLeg) {
					ConcreteSpiderLeg csl = (ConcreteSpiderLeg) di;
					if (isAssociated(csl,selectedItem)) {
						addItem = false; // if so, don't add the item
						break;
					}
				}
			}
		}
		
		if (selectedItem instanceof ConcreteSpiderLeg) {
			ConcreteSpiderLeg csl = (ConcreteSpiderLeg) selectedItem;
			
			// remove any associated feet or handles
			Iterator sit = _selectedItems.iterator();
			while (sit.hasNext()) {
				DiagramItem di = (DiagramItem) sit.next();
				if (isAssociated(csl,di)) {
					sit.remove();
				}
			}
		}
			
		// Add selected element if necessary
		if (addItem && !_selectedItems.contains(selectedItem))
			_selectedItems.add(selectedItem);

		// Update toolbar
		_viewer.getControlPanel().enableButtons(_selectedItems);

		// Update canvas
		repaint();
	}
	
	protected boolean selectedHandle() {
		if (_selectedItems.size() != 1)
			return false;
		DiagramItem item = (DiagramItem) _selectedItems.iterator().next();
		return item instanceof Handle;
	}
	
	protected synchronized java.util.Set computeHandles(
		Collection selectedItems) {
		java.util.Set handles = new LinkedHashSet();
		Iterator i = selectedItems.iterator();
		while (i.hasNext()) {
			DiagramItem item = (DiagramItem) i.next();
			handles.addAll(item.getHandles());
		}
		return handles;
	}
	
	public synchronized void deselect() {
		_selectedItems.clear();
		repaint();
	}
	
	//
	// Auxiliary methods
	//

	/*
	 * pre:  list = list of DiagramItems
	 *       x,y = mouse coordinates in the diagram
	 * post: returns true iff mouse coordinates hit one or more of the items
	 */
	/*
	 * test:
	 *   empty list
	 *   list of two or more items
	 *     last item hit
	 *     no item hit
	 */
	private boolean isInItems(List items, float x, float y)
	{
		Rectangle2D rect = getRect(x,y);
		Iterator it = items.iterator();
		while (it.hasNext())
		{
			DiagramItem di = (DiagramItem) it.next();
			if (di.getShape().intersects(rect))
				return true;
		}
		return false;
	}
	/*
	 * pre:  x,y coords of centre of required rectangle
	 * post: returns a rectangle around x,y whose sides are 
	 *         SELECTION_SIZE from the point x,y
	 */
	/*
	 * test: once thru
	 */
	private Rectangle2D getRect(float x, float y)
	{
		int rectSize = SELECTION_SIZE;
		Rectangle2D rect =
			new Rectangle2D.Float(
				x - rectSize,
				y - rectSize,
				rectSize * 2,
				rectSize * 2);
		return rect;
	}


	/*
	 * pre:  csl = concrete spider leg
	 *       di = diagram item to be checked against csl
	 * post: returns true iff di is associated with csl (i.e. is an associated
	 *         concrete spider foot or concrete spider foot handle) 
	 */
	/*
	 * test: 
	 *   di is not handle or spider foot
	 *   di is a contour handle 
	 *   di is a spider handle
	 *     di is a foot of cs 
	 *     di is not associated with csl
	 *   di is a spider foot
	 *     di is a foot of cs 
	 *     di is not associated with csl
	 */
	private boolean isAssociated(ConcreteSpiderLeg csl,DiagramItem di) {
		
		if (di instanceof Handle) {
			di = ((Handle) di).getItem();
		}
		
		if (!(di instanceof ConcreteSpiderFoot))
		  return false;
		
		ConcreteSpiderFoot csf = (ConcreteSpiderFoot) di;
		return (csf == csl.getSource() || csf == csl.getTarget());
	}

	/*
	 * pre:  ca = concrete arc
	 *       di = diagram item to be checked against ca
	 * post: returns true iff di is associated with ca (i.e. is an associated
	 *         concrete arc end or concrete arc end handle) 
	 */
	/*
	 * test: 
	 *   di is not handle or arc end
	 *   di is a contour handle 
	 *   di is an arc end handle
	 *     di is start of ca 
	 *     di is middle of ca 
	 *     di is end of ca 
	 *     di is not associated with ca 
	 *   di is an arc end
	 *     di is start of ca 
	 *     di is middle of ca 
	 *     di is end of ca 
	 *     di is not associated with ca 
	 */
	private boolean isAssociated(ConcreteArc ca,DiagramItem di) {
		
		if (di instanceof Handle) {
			di = ((Handle) di).getItem();
		}
		
		if (!(di instanceof ConcreteArcEnd))
		  return false;
		
		ConcreteArcEnd cae = (ConcreteArcEnd) di;
		return (ca.getStart() == cae ||
				ca.getMiddle() == cae ||
				ca.getEnd() == cae);
	}

	protected void invalidateDraggedObjects() {
		Iterator i = _selectedItems.iterator();
		while (i.hasNext()) {
			DiagramItem selectedItem = (DiagramItem) i.next();
			if (selectedItem instanceof ConcreteContour
				|| selectedItem instanceof DecorationBox
				|| selectedItem instanceof ConcreteTextBox
				|| selectedItem instanceof Handle) {
				Shape s = selectedItem.getShape();
				if (s instanceof Polygon) {
					Polygon p = (Polygon) s;
					p.invalidate();
				}
			}
		}
	}
	protected ConcreteSpiderFoot findNearestFoot(float x, float y) {
		ConcreteSpiderFeetAndLegsSet fls =
			_viewer.getConcreteDiagram().getFeetAndLegsSet();
		return fls.getNearestFoot((int) x, (int) y);

	}

	public synchronized int getLegType() {
		ControlPanel control = _viewer.getControlPanel();
		if (control.cmd == "Tie")
			return ConcreteSpiderLeg.TIE;
		if (control.cmd == "Strand")
			return ConcreteSpiderLeg.STRAND;
		return ConcreteSpiderLeg.NORMAL;
	}
	public synchronized void updateZoom(int clicks) {
		int newClicks = _scaleClicks + clicks;
		if (newClicks > _minClicks && newClicks <= _maxClicks) {
			_scaleClicks = newClicks;
			float oldScale = _scale;
			_scale = 1 / (float) _scaleClicks;
			_xTranslate = _xTranslate * (oldScale / _scale);
			_yTranslate = _yTranslate * (oldScale / _scale);
			rescaleTransform();
			//            repaintImage();
			drawImageNow();
		}
	}
	public synchronized void zoomScale(int scale) {
		float oldScale = _scale;
		_scale = (float) scale / 1000;
		int oldClicks = _scaleClicks;
		_scaleClicks = (int) (1.0 / _scale);
		_xTranslate = _xTranslate * (oldScale / _scale);
		_yTranslate = _yTranslate * (oldScale / _scale);
		rescaleTransform();
		//        repaintImage();
		drawImageNow();
	}

	public synchronized void rescaleTransform() {
		_transform.setToIdentity();
		_transform.scale(_scale, _scale);
		_transform.translate(_xTranslate, _yTranslate);
		if (_g != null) {
			_g.setTransform(_transform);
		}
	}

	public synchronized String getEPS() {
		EpsGraphics2D epsg = new EpsGraphics2D();
		epsg.setFont(_g.getFont());
		synchronized (_g) {
			Graphics2D oldG = _g;
			_g = epsg;
			repaintImage();
			_g = oldG;
		}
		epsg.setTitle("Constraint Diagram");
		return epsg.toString();
	}
	
	//
	// Getters and setters
	//
	public java.util.List getSelectedItems() {
		return _selectedItems;
	}

	public synchronized BufferedImage getBufferedImage(Rectangle rect) {
		BufferedImage image =
			new BufferedImage(
				_image.getWidth(null),
				_image.getHeight(null),
				BufferedImage.TYPE_INT_RGB);
		synchronized (_g) {
			Graphics2D oldG = _g;
			_g = (Graphics2D) image.getGraphics();
			_g.setFont(oldG.getFont());
			_g.setRenderingHint(
				RenderingHints.KEY_ANTIALIASING,
				RenderingHints.VALUE_ANTIALIAS_ON);
			repaintImage();
			_g = oldG;
		}
		if (rect != null) {
			//			image = image.getSubimage(rect.x, 500, rect.width, rect.height);
		}
		return image;
	}

	/**
	 * Post: offsets for x and y translation from diagram to canvas
	 *       have been set up.
	 * @param x translation x-offset for diagram relative to canvas 
	 * @param y translation y-offset for diagram relative to canvas 
	 */
	public synchronized void setTranslation(float x, float y) {
		_xTranslate = x;
		_yTranslate = y;
		rescaleTransform();
	}

	/**
	 * Post: scale factor from diagram to can vas has been set up.
	 * @param scale scale factor from diagram to canvas
	 */
	public synchronized void setScale(float scale) {
		_scale = scale;
		rescaleTransform();
	}

	public synchronized boolean getSmoothed() {
		return _smoothed;
	}

	public synchronized void setSmoothed(boolean smoothed) {
		_smoothed = smoothed;
	}

	/**
	 * @param e mouse event
	 * @return x coordinate of e in diagram coordinates
	 */
	public float computeCurrentX(MouseEvent e) {
		float x = e.getX() / _scale - _xTranslate;
		return x;
	}

	/**
	 * @param e mouse event
	 * @return y coordinate of e in diagram coordinates
	 */
	public float computeCurrentY(MouseEvent e) {
		float y = e.getY() / _scale - _yTranslate;
		return y;
	}

	public synchronized float getCurrentX() {
		return _currentX;
	}
	public synchronized void setCurrentX(float f) {
		_currentX = f;
	}

	public synchronized float getCurrentY() {
		return _currentY;
	}
	public synchronized void setCurrentY(float f) {
		_currentY = f;
	}

	//
	// Local properties
	//
	protected DiagramViewerInterface _viewer;

	protected Image _image;
	protected Graphics2D _g;

	protected boolean _smoothed = false;

	protected float _currentX; // current x-coord in diagram coords
	protected float _currentY; // current y-coord in diagram coords

	protected float _previousX; // previous x-coord in diagram coords
	protected float _previousY; // previous y-coord in diagram coords

	protected boolean _waitingToBeDrawn = false;
	protected boolean _handleMoved = false;
	protected boolean _translating = false;

	protected int _minClicks = 4;
	protected int _maxClicks = 20;
	protected float _minScale = 0.01f;
	protected float _maxScale = 0.2f;
	protected int _initClicks = 10;
	protected float _initScale = 1 / (float) _initClicks;
	protected int _scaleClicks = _initClicks;
	protected float _scale = _initScale;
	protected float _oldScale = 0.1f;
	protected float _xTranslate = 2500.0f;
	protected float _yTranslate = 2500.0f;
	protected AffineTransform _transform = new AffineTransform();
	// transform from diagram to canvas

	private java.util.List _selectedItems = new java.util.Vector();

	//    protected Rectangle2D _originalSize = null;
	//    protected Rectangle2D _newSize = null;
	//    protected Point2D _originalLabelPosition = null;
	//    protected Point2D _newLabelPosition = null;
	//    protected Shape _newShape = null;

	// Used to resize contours by dragging a corner handle
	protected Polygon _oldShape = null;
	protected Polygon _newShape = null;

	// Used to draw a leg by shift+dragging a spider foot
	protected ConcreteSpiderFoot _sourceFoot = null;
	protected ConcreteSpiderFoot _targetFoot = null;
	protected ConcreteSpiderLeg _leg = null;
	protected Handle _handle = null;

	protected EditPopupMenu _popup;

	// Release/debug version
	protected boolean _debug = true;

	//
	// Static members
	//
	// Make this bigger to increase the selection area
	public static final int SELECTION_SIZE = 45;

	public static BasicStroke DEFAULT =
		new BasicStroke(30.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_ROUND);
	public static BasicStroke THIN =
		new BasicStroke(8.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_ROUND);
	public static BasicStroke THICK =
		new BasicStroke(150.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_ROUND);
	public static BasicStroke DOT =
		new BasicStroke(
			16.0f,
			BasicStroke.CAP_SQUARE,
			BasicStroke.JOIN_MITER,
			1.0f,
			new float[] { 20.0f, 40.0f },
			40.0f);

	public static BasicStroke STROKE = DEFAULT;

	/*
	// test program
    // ------------

    private static void pr(String s) {
      System.out.println(s);
    }

    private static void hd(String s) {
      System.out.println("\n" + s + "\n");
    }

    private static void pg() {
      try {
        int i;
        System.out.print("next page? (CR)");
        do i = System.in.read(); while (i != '\n');
        System.out.println();
      }
      catch (Exception e) {System.out.println("raised exception : " + e);}
    }

    public static void main(String args[]) throws Exception {
    	    	
    	DiagramCanvas dc = new DiagramCanvas(null);
    	/*
    	{
	    	hd("test of isAssociated for spider legs");
	    	
	    	ConcreteSpiderFoot f1 = 
	    		new ConcreteSpiderFoot("f1",ConcreteSpiderFoot.NORMAL);
		    	pr("f1="+f1.id());
		    	ConcreteSpiderFoot f2 = 
		    		new ConcreteSpiderFoot("f2",ConcreteSpiderFoot.NORMAL);
		    	pr("f2="+f2.id());
		    	ConcreteSpiderFoot f3 = 
		    		new ConcreteSpiderFoot("f3",ConcreteSpiderFoot.NORMAL);
		    	pr("f3="+f3.id());
	    	pg();
	    	
	    	ConcreteSpiderLeg csl = 
	    		new ConcreteSpiderLeg("fred",f1,f2,ConcreteSpiderLeg.NORMAL);
	    	pr("csl="+csl);
	    	pg();
	    	
	    	pr("dc.isAssociated(csl,csl)="+dc.isAssociated(csl,csl));
	    	pg();
	    	
	    	Polygon p = new Polygon();
	    	p.addPoint(100,100);
	    	ConcreteContour cc = new ConcreteContour(Contour.getContour("cc"),p);
	    	pr("cc="+cc);
	    	Handle ch = new Handle(100,100,cc);
	    	pr("ch="+ch);
	    	pr("dc.isAssociated(cs,ch)="+dc.isAssociated(csl,ch));
	    	pg();
	    	
	    	Handle chf1 = new Handle(100,100,f1);
	    	pr("chf1="+chf1);
	    	pr("dc.isAssociated(csl,chf1)="+dc.isAssociated(csl,chf1));
	    	
	    	Handle chf2 = new Handle(100,100,f2);
	    	pr("chf2="+chf2);
	    	pr("dc.isAssociated(csl,chf2)="+dc.isAssociated(csl,chf2));
	    	
	    	Handle chf3 = new Handle(100,100,f3);
	    	pr("che="+chf3);
	    	pr("dc.isAssociated(csl,chf3)="+dc.isAssociated(csl,chf3));
	    	pg();
	    	
	    	pr("dc.isAssociated(csl,f1)="+dc.isAssociated(csl,f1));
	    	pr("dc.isAssociated(csl,f2)="+dc.isAssociated(csl,f2));
	    	pr("dc.isAssociated(csl,f3)="+dc.isAssociated(csl,f3));
	    	pg();
    	}
    	*/
    	/*
    	{
	    	hd("test of isAssociated for arcs");
	    	ConcreteArcEnd start = new ConcreteArcEnd(1,2);
	    	pr("start="+start.id());
	    	ConcreteArcEnd middle = new ConcreteArcEnd(3,4); 
	    	pr("middle="+middle.id());
	    	ConcreteArcEnd end = new ConcreteArcEnd(5,6);
	    	pr("end="+end.id());
	    	ConcreteArcEnd other = new ConcreteArcEnd(7,8);
	    	pr("other="+other.id());
	    	pg();
	    	
	    	ConcreteArc ca = new ConcreteArc();
	    	pr("ca="+ca);
	    	ca.setStart(start);
	    	ca.setMiddle(middle);
	    	ca.setEnd(end);
	    	ca.setLabel("ca");
	    	pr("ca="+ca);
	    	pg();
	    	
	    	pr("dc.isAssociated(ca,ca)="+dc.isAssociated(ca,ca));
	    	pg();
	    	
	    	
	    	
	    	Polygon p = new Polygon();
	    	p.addPoint(100,100);
	    	ConcreteContour cc = new ConcreteContour(Contour.getContour("cc"),p);
	    	pr("cc="+cc);
	    	Handle ch = new Handle(100,100,cc);
	    	pr("ch="+ch);
	    	pr("dc.isAssociated(ca,ch)="+dc.isAssociated(ca,ch));
	    	pg();
	    	
	    	Handle chs = new Handle(100,100,start);
	    	pr("chs="+chs);
	    	pr("dc.isAssociated(ca,chs)="+dc.isAssociated(ca,chs));
	    	
	    	Handle chm = new Handle(100,100,middle);
	    	pr("chm="+chm);
	    	pr("dc.isAssociated(ca,chm)="+dc.isAssociated(ca,chm));
	    	
	    	Handle che = new Handle(100,100,end);
	    	pr("che="+che);
	    	pr("dc.isAssociated(ca,che)="+dc.isAssociated(ca,che));
	    	
	    	Handle cho = new Handle(100,100,other);
	    	pr("che="+cho);
	    	pr("dc.isAssociated(ca,cho)="+dc.isAssociated(ca,cho));
	    	pg();
	    	
	    	pr("dc.isAssociated(ca,start)="+dc.isAssociated(ca,start));
	    	pr("dc.isAssociated(ca,middle)="+dc.isAssociated(ca,middle));
	    	pr("dc.isAssociated(ca,end)="+dc.isAssociated(ca,end));
	    	pr("dc.isAssociated(ca,other)="+dc.isAssociated(ca,other));
	    	pg();
    	}
    	
    }
    */
}
