package rwd.ocl.constraintdiagram;

import java.awt.Polygon;
import java.util.ArrayList;
import java.util.Vector;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import rwd.abstractsyntax.constraintdiagram.Arrow;
import rwd.abstractsyntax.eulerdiagram.Contour;
import rwd.abstractsyntax.eulerdiagram.Zone;
import rwd.abstractsyntax.spiderdiagram.Spider;
import rwd.abstractsyntax.spiderdiagram.UnitaryDiagram;
import rwd.concretesyntax.ConcreteUnitaryDiagram;
import rwd.concretesyntax.DiagramItem;
import rwd.concretesyntax.constraintdiagram.ConcreteArc;
import rwd.concretesyntax.eulerdiagram.ConcreteContour;
import rwd.concretesyntax.eulerdiagram.ConcreteZone;
import rwd.concretesyntax.eulerdiagram.DecorationBox;
import rwd.concretesyntax.eulerdiagram.Polygonal;
import rwd.concretesyntax.spiderdiagram.ConcreteSpider;
import rwd.concretesyntax.spiderdiagram.ConcreteSpiderFoot;

/**
 * @author Octavian Patrascoiu, rnb
 * @version 25.8.05   09:36
 *
 */
public class DependenceGraph {
	public DependenceGraph(ConcreteUnitaryDiagram cud) {
		// Initialize input members
		this.cud = cud;
		if (cud == null) return;
		cud.refreshZones(false);
		this.graph = new Graph(new Vector(cud.getConcreteSpiders()));
		// Add edges 
		addPairsRule1(cud, graph);
		addPairsRule2(cud, graph);
		addPairsRule3(cud, graph);
		addPairsRule4(cud, graph);
		addPairsRule5(cud, graph);
	}

	public DependenceGraph(UnitaryDiagram ud) {
		// Initialize input members
		this.ud = ud;
		if (ud == null) return;
		this.graph = new Graph(new Vector(ud.getSpiders()));
		// Add edges 
		addPairsRule1(ud, graph);
		addPairsRule2(ud, graph);
		addPairsRule3(ud, graph);
		addPairsRule4(ud, graph);
		addPairsRule5(ud, graph);
	}

	public List getNodes() {
		List nodes = new Vector();
		if (graph == null) return nodes;
		return graph.getNodes();
	}
	
	public List getEdges() {
		List edges = new Vector();
		if (graph == null) return edges;
		return graph.getEdges();		
	}
	
	//
	// Add pairs according to rules
	//
	// (s, t) <=> "Their habitat have non-empty intersection"
	//
	protected void addPairsRule1(ConcreteUnitaryDiagram cud, Graph graph) {
		List spiders = graph.nodes;
		for(int i=0; i<spiders.size(); i++) {
			ConcreteSpider s = (ConcreteSpider)spiders.get(i);
			Spider as1 = s.getAbstractSpider(cud);
			List habitat1 = as1.getHabitat();
			for(int j=i+1; j<spiders.size(); j++) {
				ConcreteSpider t = (ConcreteSpider)spiders.get(j);
				Spider as2 = t.getAbstractSpider(cud);
				List habitat2 = as2.getHabitat();
				habitat2.retainAll(habitat1);
				if (habitat2.size() != 0) {
					graph.addPair(s, t);
					graph.addPair(t, s);
				}
			}
		}
	}
	protected void addPairsRule1(UnitaryDiagram cud, Graph graph) {
		List spiders = graph.nodes;
		for(int i=0; i<spiders.size(); i++) {
			Spider s = (Spider)spiders.get(i);
			List habitat1 = s.getHabitat();
			for(int j=i+1; j<spiders.size(); j++) {
				Spider t = (Spider)spiders.get(j);
				List habitat2 = t.getHabitat();
				habitat2.retainAll(habitat1);
				if (habitat2.size() != 0) {
					graph.addPair(s, t);
					graph.addPair(t, s);
				}
			}
		}
	}

	//
	// There is a derived arrow sequence with source s and target t 
	//
	protected void addPairsRule2(ConcreteUnitaryDiagram cud, Graph graph) {
		// (s, t) <=> if there is an arrow from s to t
		Graph rel = new Graph(graph.nodes);
		List arrows = cud.getArcSet().getArcs();
		for(int i=0; i<arrows.size(); i++) {
			ConcreteArc arc = (ConcreteArc)arrows.get(i);
			DiagramItem source = arc.getSource();
			DiagramItem target = arc.getTarget();
			if (source instanceof ConcreteSpiderFoot && target instanceof ConcreteSpiderFoot) {
				ConcreteSpider s = cud.getConcreteSpider((ConcreteSpiderFoot)source);
				ConcreteSpider t = cud.getConcreteSpider((ConcreteSpiderFoot)target);
				if (s != t) {
					rel.addPair(s, t);
				}
			}
		}
		rel = rel.closure();
		// Add pairs (s, t)
		List nodes = rel.getNodes();
		for(int i=0; i<nodes.size(); i++) {
			Object s = nodes.get(i);
			List ts = rel.getNodes(s);
			for(int j=0; j<ts.size(); j++) {
				Object t = ts.get(j);
				graph.addPair(s, t);
			}
		}		
	}
	protected void addPairsRule2(UnitaryDiagram cud, Graph graph) {
		// (s, t) <=> if there is an arrow from s to t
		Graph rel = new Graph(graph.nodes);
		List arrows = ud.getArrows();
		for(int i=0; i<arrows.size(); i++) {
			Arrow arc = (Arrow)arrows.get(i);
			Object source = arc.getSource();
			Object target = arc.getTarget();
			if (source instanceof Spider && target instanceof Spider) {
				Spider s = (Spider)source;
				Spider t = (Spider)target;
				if (s != t) {
					rel.addPair(s, t);
				}
			}
		}
		rel = rel.closure();
		// Add pairs (s, t)
		List nodes = rel.getNodes();
		for(int i=0; i<nodes.size(); i++) {
			Object s = nodes.get(i);
			List ts = rel.getNodes(s);
			for(int j=0; j<ts.size(); j++) {
				Object t = ts.get(j);
				graph.addPair(s, t);
			}
		}		
	}

	//
	// The habitat of s has a minimal description which involves a derived contour target(a)
	// where a is a potential arrow consequence of t
	//
	// or the same condition with s and t exchanged
	/*
	Note: this method has some problems, I think.
	First of all, the derived contour is not extracted from the minimal habitat,
	so the contour c is never found, because only the zones of the minimal habitat
	are extracted, and none of these is a contour.
	Secondly, the potential arrow consequences, tArrows, always seems to be empty, 
	even for spiders which have arrows to derived contours. 
	*/
	protected void addPairsRule3(ConcreteUnitaryDiagram cud, Graph graph) {
		System.out.println("DependenceGraph.addPairsRule3:start");
		// For each spider s
		List spiders = new Vector(cud.getConcreteSpiders());
		for(int i=0; i<spiders.size(); i++) {
			ConcreteSpider s = (ConcreteSpider)spiders.get(i);
			System.out.println("DependenceGraph.addPairsRule3:s="+s);
			List min = minimalHabitat(s, cud);
			System.out.println("DependenceGraph.addPairsRule3:min="+min);
			Collection dcs = derivedContoursOf(min);
			System.out.println("DependenceGraph.addPairsRule3:dcs="+dcs);
			// For each spider t
			for(int j=0; j<spiders.size(); j++) {
				if (i != j) {
					ConcreteSpider t = (ConcreteSpider)spiders.get(j);
					System.out.println("DependenceGraph.addPairsRule3:t="+t);
					List tArrows = new Vector(potentialArrowConsequences(t, cud));
					System.out.println("DependenceGraph.addPairsRule3:tArrows="+tArrows);
					Iterator dcit = dcs.iterator();
					while (dcit.hasNext()) {
						Object obj = dcit.next();
						System.out.println("DependenceGraph.addPairsRule3:obj="+obj);
						if (obj instanceof ConcreteContour) {
							ConcreteContour c = (ConcreteContour)obj;
							System.out.println("DependenceGraph.addPairsRule3:c="+c);
							if (c.isDerived()) {
								for(int l=0; l<tArrows.size(); l++) {
									ConcreteArc arrow = (ConcreteArc)tArrows.get(l);
									System.out.println("DependenceGraph.addPairsRule3:arrow="+arrow);
									DiagramItem target = arrow.getTarget();
									System.out.println("DependenceGraph.addPairsRule3:target="+target);
									if (target == c) {
										graph.addPair(s, t);
									}
								}
							}
						}
					}
				}
			}
		}
	}
	protected void addPairsRule3(UnitaryDiagram cud, Graph graph) {
		// For each spider s
		List spiders = new Vector(ud.getSpiders());
		for(int i=0; i<spiders.size(); i++) {
			Spider s = (Spider)spiders.get(i);
			List min = minimalHabitat(s, ud);
			// For each spider t
			for(int j=0; j<spiders.size(); j++) {
				if (i != j) {
					Spider t = (Spider)spiders.get(j);
					List tArrows = new Vector(potentialArrowConsequences(t, ud));
					for(int k=0; k<min.size(); k++) {
						Object obj = min.get(k);
						if (obj instanceof Contour) {
							Contour c = (Contour)obj;
							if (c.isDerived()) {
								for(int l=0; l<tArrows.size(); l++) {
									Arrow arrow = (Arrow)tArrows.get(l);
									Object target = arrow.getTarget();
									if (target == c) {
										graph.addPair(s, t);
									}
								}
							}
						}
					}
				}
			}
		}
	}
	
	//
	// There are derived contours target(af) and target(ag) where af and ag are potential
	// consequences of the spiders s and t, and where 
	// target(af) is in DC(Rsmall(target(ag))) or target(af) in DC(Rbig(target(ag))))
	//
	// or the same condition with s and t exchanged
	//
	protected void addPairsRule4(ConcreteUnitaryDiagram cud, Graph graph) {
		// For each spider s
		List spiders = new Vector(cud.getConcreteSpiders());
		for(int i=0; i<spiders.size(); i++) {
			ConcreteSpider s = (ConcreteSpider)spiders.get(i);
			List sArrows = new Vector(potentialArrowConsequences(s, cud));
			// For each spider t
			for(int j=0; j<spiders.size(); j++) {
				if (i != j) {
					ConcreteSpider t = (ConcreteSpider)spiders.get(j);
					List tArrows = new Vector(potentialArrowConsequences(t, cud));
					// For each pair dc1 and dc2 where dc1 target of an sArrow and dc2 target of an tArrow
					for(int si=0; si<sArrows.size(); si++) {
						ConcreteArc sArrow = (ConcreteArc)sArrows.get(si);
						DiagramItem sTarget = sArrow.getTarget();
						if (sTarget instanceof ConcreteContour && ((ConcreteContour)sTarget).isDerived()) {
							for(int ti=0; ti<tArrows.size(); ti++) {
								ConcreteArc tArrow = (ConcreteArc)tArrows.get(ti);
								DiagramItem tTarget = tArrow.getTarget();
								if (tTarget instanceof ConcreteContour && ((ConcreteContour)tTarget).isDerived()) {
									if (dc(rsmall((ConcreteContour)tTarget, cud)).contains(sTarget)) {
										graph.addPair(s, t);
									}
									if (dc(rbig((ConcreteContour)tTarget, cud)).contains(sTarget)) {
										graph.addPair(s, t);
									}
								}
							}
						}
						
					}
				}
			}
		}
	}
	protected void addPairsRule4(UnitaryDiagram cud, Graph graph) {
		// For each spider s
		List spiders = new Vector(ud.getSpiders());
		for(int i=0; i<spiders.size(); i++) {
			Spider s = (Spider)spiders.get(i);
			List sArrows = new Vector(potentialArrowConsequences(s, ud));
			// For each spider t
			for(int j=0; j<spiders.size(); j++) {
				if (i != j) {
					Spider t = (Spider)spiders.get(j);
					List tArrows = new Vector(potentialArrowConsequences(t, ud));
					// For each pair dc1 and dc2 where dc1 target of an sArrow and dc2 target of an tArrow
					for(int si=0; si<sArrows.size(); si++) {
						Arrow sArrow = (Arrow)sArrows.get(si);
						Object sTarget = sArrow.getTarget();
						if (sTarget instanceof Contour && ((Contour)sTarget).isDerived()) {
							for(int ti=0; ti<tArrows.size(); ti++) {
								Arrow tArrow = (Arrow)tArrows.get(ti);
								Object tTarget = tArrow.getTarget();
								if (tTarget instanceof Contour && ((Contour)tTarget).isDerived()) {
									if (dc(rsmall((Contour)tTarget, cud)).contains(sTarget)) {
										graph.addPair(s, t);
									}
									if (dc(rbig((Contour)tTarget, cud)).contains(sTarget)) {
										graph.addPair(s, t);
									}
								}
							}
						}
						
					}
				}
			}
		}
	}

	//
	// There is a shaded zone which is potential consequence of both s and t
	//
	protected void addPairsRule5(ConcreteUnitaryDiagram cud, Graph graph) {
		// For each spider s
		List spiders = new Vector(cud.getConcreteSpiders());
		for(int i=0; i<spiders.size(); i++) {
			ConcreteSpider s = (ConcreteSpider)spiders.get(i);
			List sZones = new Vector(potentialShadedZonesConsequences(s, cud));
			// For each spider t
			for(int j=i+1; j<spiders.size(); j++) {
				ConcreteSpider t = (ConcreteSpider)spiders.get(j);
				List tZones = new Vector(potentialShadedZonesConsequences(t, cud));
				// If tZones and sZones intersect
				tZones.retainAll(sZones);
				if (tZones.size() != 0) {
					graph.addPair(s, t);
					graph.addPair(t, s);
				}
			}
		}
	}
	protected void addPairsRule5(UnitaryDiagram cud, Graph graph) {
		// For each spider s
		List spiders = new Vector(ud.getSpiders());
		for(int i=0; i<spiders.size(); i++) {
			Spider s = (Spider)spiders.get(i);
			List sZones = new Vector(potentialShadedZonesConsequences(s, ud));
			// For each spider t
			for(int j=i+1; j<spiders.size(); j++) {
				Spider t = (Spider)spiders.get(j);
				List tZones = new Vector(potentialShadedZonesConsequences(t, ud));
				// If tZones and sZones intersect
				tZones.retainAll(sZones);
				if (tZones.size() != 0) {
					graph.addPair(s, t);
					graph.addPair(t, s);
				}
			}
		}
	}
	
	//
	// Computes the minimal habitat of s:
	//
	// the set of zones that uses the minimal number of contours	
	//
	protected List minimalHabitat(ConcreteSpider s, ConcreteUnitaryDiagram cud) {
		List res = new Vector();
		List feet = new Vector(s.getFeet());
		for(int i=0; i<feet.size(); i++) {
			ConcreteSpiderFoot foot = (ConcreteSpiderFoot)feet.get(i);
			ConcreteZone cz = cud.getConcreteZone(foot);
			res.add(cz);			
		}
		return res;
	}
	protected List minimalHabitat(Spider s, UnitaryDiagram ud) {
		List res = s.getHabitat();
		return res;
	}

	//
	// Computes rsmall(c):
	//
	// the smallest set of zones that contains c 
	//
	// Make zbig and remove zones for which zexc=(C+-{c},C-+{c}) is a zone
	//
	public static Set rsmall(ConcreteContour c, ConcreteUnitaryDiagram cud) {
		// Make rbig
		List rbig = new Vector(rbig(c, cud));
		Set res = new LinkedHashSet();
		List zones = new Vector(cud.getConcreteZones());
		for(int i=0; i<rbig.size(); i++) {
			ConcreteZone cz = (ConcreteZone)rbig.get(i);
			Vector newCPolys = new Vector();
			newCPolys.addAll(cz.getContainingPolygonals());
			Vector newEPolys = new Vector();
			newEPolys.addAll(cz.getExcludingPolygonals());
			ConcreteZone shallowCopy = new ConcreteZone(newCPolys, newEPolys); 
			ConcreteZone zexc = shallowCopy;
			zexc.getContainingPolygonals().remove(c);
			zexc.getExcludingPolygonals().add(c);
			if (!isZone(zexc, zones)) {
				res.add(zexc);
			}
		}
		return res;
	}
	public static Set rsmall(Contour c, UnitaryDiagram ud) {
		// Make rbig
		List rbig = new Vector(rbig(c, ud));
		Set res = new LinkedHashSet();
		List zones = ud.getZones();
		for(int i=0; i<rbig.size(); i++) {
			Zone az = (Zone)rbig.get(i);
			Vector newCPolys = new Vector();
			newCPolys.addAll(az.getContainingContours());
			Vector newEPolys = new Vector();
			newEPolys.addAll(az.getExcludingContours());
			Zone shallowCopy = new Zone(newCPolys, newEPolys); 
			Zone zexc = shallowCopy;
			zexc.getContainingContours().remove(c);
			zexc.getExcludingContours().add(c);
			if (!isZone(zexc, zones)) {
				res.add(zexc);
			}
		}
		return res;
	}
	protected static boolean isZone(ConcreteZone zexc, Collection zones) {
		List cPolys1 = zexc.getContainingPolygonals();
		List ePolys1 = zexc.getExcludingPolygonals();
		Iterator i = zones.iterator();
		while (i.hasNext()) {
			ConcreteZone z = (ConcreteZone)i.next();
			List cPolys2 = z.getContainingPolygonals();
			List ePolys2 = z.getExcludingPolygonals();
			boolean b1 = cPolys2.containsAll(cPolys1) && cPolys1.containsAll(cPolys2);
			boolean b2 = ePolys2.containsAll(ePolys1) && ePolys1.containsAll(ePolys2); 
			if (b1 && b2)
				return true;
		}
		return false;
	}
	protected static boolean isZone(Zone zexc, Collection zones) {
		List cPolys1 = zexc.getContainingContours();
		List ePolys1 = zexc.getExcludingContours();
		Iterator i = zones.iterator();
		while (i.hasNext()) {
			Zone z = (Zone)i.next();
			List cPolys2 = z.getContainingContours();
			List ePolys2 = z.getExcludingContours();
			boolean b1 = cPolys2.containsAll(cPolys1) && cPolys1.containsAll(cPolys2);
			boolean b2 = ePolys2.containsAll(ePolys1) && ePolys1.containsAll(ePolys2); 
			if (b1 && b2)
				return true;
		}
		return false;
	}

	//
	// Computes rbig(c):
	//
	// the largest set of zones that contains c
	// 
	// For each zone z=(C+, C-) make zbig=(C+-{c}, C-)
	// 
	public static Set rbig(ConcreteContour c, ConcreteUnitaryDiagram cud) {
		Set res = new LinkedHashSet();
		List zones = new Vector(cud.getConcreteZones());
		for(int i=0; i<zones.size(); i++) {
			ConcreteZone cz = (ConcreteZone)zones.get(i);
			List polys = cz.getContainingPolygonals();
			if (polys.contains(c)) {
				Vector newCPolys = new Vector();
				newCPolys.addAll(cz.getContainingPolygonals());
				Vector newEPolys = new Vector();
				newEPolys.addAll(cz.getExcludingPolygonals());
				ConcreteZone shallowClone = new ConcreteZone(newCPolys, newEPolys);
				ConcreteZone zbig = shallowClone;
				zbig.getContainingPolygonals().remove(c);
				res.add(zbig);
			}
		}
		return res;
	}
	public static Set rbig(Contour c, UnitaryDiagram ud) {
		Set res = new LinkedHashSet();
		List zones = new Vector(ud.getZones());
		for(int i=0; i<zones.size(); i++) {
			Zone az = (Zone)zones.get(i);
			List polys = az.getContainingContours();
			if (polys.contains(c)) {
				Vector newCPolys = new Vector();
				newCPolys.addAll(az.getContainingContours());
				Vector newEPolys = new Vector();
				newEPolys.addAll(az.getExcludingContours());
				Zone shallowClone = new Zone(newCPolys, newEPolys);
				Zone zbig = shallowClone;
				zbig.getContainingContours().remove(c);
				res.add(zbig);
			}
		}
		return res;
	}
	
	//
	// Computes the derived concrete contours contained into a region
	//
	protected Set dc(Collection c) {
		Set res = new LinkedHashSet();
		Iterator i = c.iterator();
		while (i.hasNext()) {
			Object obj = i.next();
			if (obj instanceof ConcreteZone) {
				List cs = ((ConcreteZone)obj).getAllPolygonals();
				for(int j=0; j<cs.size(); j++) {
					Object obj1 = cs.get(j);
					if (obj1 instanceof ConcreteContour	 && ((ConcreteContour)obj1).isDerived()) {
						res.add(obj);
					}
				}
			}
			if (obj instanceof Zone) {
				List cs = ((Zone)obj).getAllContours();
				for(int j=0; j<cs.size(); j++) {
					Object obj1 = cs.get(j);
					if (obj1 instanceof Contour	 && ((Contour)obj1).isDerived()) {
						res.add(obj);
					}
				}
			}
		}
		return res;		
	}

	//
	// Compute the potential arrow consequences of s. That is, the set of arrows
	// a such that:
	//   There is a derived arrow sequence including a with source s or
	//     s is target of a
	//
	public static Set potentialArrowConsequences(ConcreteSpider sp, ConcreteUnitaryDiagram cud) {
		pr("potentialArrowConsequences:start");
		// Initialize result and feet set
		Set res = new LinkedHashSet();
		Set items = new LinkedHashSet();
		addItem(items, sp);
		pr("  res="+res);
		pr("  items="+items);
		// Add items until no more items can be added
		int tempSize;
		List arrows = new Vector(cud.getConcreteArcs());
		pr("  arrows="+arrows);
		do {
			tempSize = res.size();
			// For each arrow with source in res
			for(int j=0; j<arrows.size(); j++) {
				ConcreteArc arc = (ConcreteArc)arrows.get(j);
				pr("      arc="+arc);
				pr("      arc.getSource()="+arc.getSource());
				pr("      arc.getTarget()="+arc.getTarget());
				DiagramItem source = arc.getSource();
				pr("      source="+source);
				if (items.contains(source)) {
					Object target = arc.getTarget();
					if (target instanceof ConcreteSpider)
						addItem(items, target);
					res.add(arc);
				}
				pr("    res="+res);
				pr("    items="+items);
			}
		} while (tempSize != res.size());
		// Add arrow for which s is a target
		for(int j=0; j<arrows.size(); j++) {
			ConcreteArc arc = (ConcreteArc)arrows.get(j);
			// If target is a spider
			Object target = arc.getTarget();
			if (sp.getFeet().contains(target)) {
				res.add(arc);
			}
		}
		pr("potentialArrowConsequences:end");
		return res;
	}
	public static Set potentialArrowConsequences(Spider sp, UnitaryDiagram ud) {
		// Initialize result and feet set
		Set res = new LinkedHashSet();
		Set items = new LinkedHashSet();
		addItem(items, sp);
		// Add items until now more items are added
		int tempSize;
		List arrows = new Vector(ud.getArrows());
		do {
			tempSize = res.size();
			// For each arrow with source in res
			for(int j=0; j<arrows.size(); j++) {
				Arrow arc = (Arrow)arrows.get(j);
				Object source = arc.getSource();
				if (items.contains(source)) {
					Object target = arc.getTarget();
					if (target instanceof Spider)
						addItem(items, target);
					res.add(arc);
				}
			}
		} while (tempSize != res.size());
		// Add arrow for which s is a target
		for(int j=0; j<arrows.size(); j++) {
			Arrow arc = (Arrow)arrows.get(j);
			// If target is a spider
			Object target = arc.getTarget();
			if (sp == target) {
				res.add(arc);
			}
		}		
		return res;
	}
	protected static Set addItem(Set s, Object item) {
		if (item instanceof ConcreteSpider)
			s.addAll(((ConcreteSpider)item).getFeet());
		else if (item instanceof Spider)
			s.addAll(((Spider)item).getHabitat());
		else
			s.add(item);
		return s;
	}
	
	//
	// Compute the potential shaded zones consequences of s:
	//
	// s touches z of 
	// one of the minimal descriptions of z involves a derived contour
	// which is the target of a potential arrow consequence of s
	//
	public static Set potentialShadedZonesConsequences(ConcreteSpider s, ConcreteUnitaryDiagram cud) {
		// Initialize result
		Set res = new LinkedHashSet();
		List feet = new Vector(s.getFeet());
		for(int i=0; i<feet.size(); i++) {
			ConcreteSpiderFoot foot = (ConcreteSpiderFoot)feet.get(i);
			ConcreteZone cz = cud.getConcreteZone(foot);
			if (cz.isShaded()) {
				res.add(cz);				
			}
		}
		// Compute potential arrow consequences
		List arrows = new Vector(potentialArrowConsequences(s, cud));
		// For each shaded zone
		List zones = new Vector(cud.getConcreteZones());
		for(int i=0; i<zones.size(); i++) {
			ConcreteZone cz = (ConcreteZone)zones.get(i);
			if (cz.isShaded()) {
				// For each derived contour defining cz
				List contours = cz.getAllPolygonals();
				for(int j=0; j<contours.size(); j++) {
					Object obj = contours.get(j);
					if (obj instanceof ConcreteContour) {
						ConcreteContour c = (ConcreteContour)obj;
						if (c.isDerived()) {
							for(int k=0; k<arrows.size(); k++) {
								ConcreteArc arrow = (ConcreteArc)arrows.get(k);
								DiagramItem target = arrow.getTarget();
								if (target == c) {
									res.add(cz);
								}
							}
						}
					}
				}
			}
		}
		return res;
	}
	public static Set potentialShadedZonesConsequences(Spider s, UnitaryDiagram ud) {
		// Initialize result
		Set res = new LinkedHashSet();
		List zones = new Vector(s.getHabitat());
		for(int i=0; i<zones.size(); i++) {
			Zone az = (Zone)zones.get(i);
			if (az.isShaded(ud)) {
				res.add(az);				
			}
		}
		// Compute potential arrow consequences
		List arrows = new Vector(potentialArrowConsequences(s, ud));
		// For each shaded zone
		for(int i=0; i<zones.size(); i++) {
			Zone az = (Zone)zones.get(i);
			if (az.isShaded(ud)) {
				// For each derived contour defining cz
				List contours = az.getAllContours();
				for(int j=0; j<contours.size(); j++) {
					Object obj = contours.get(j);
					if (obj instanceof Contour) {
						Contour c = (Contour)obj;
						if (c.isDerived()) {
							for(int k=0; k<arrows.size(); k++) {
								Arrow arrow = (Arrow)arrows.get(k);
								Object target = arrow.getTarget();
								if (target == c) {
									res.add(az);
								}
							}
						}
					}
				}
			}
		}
		return res;
	}

	public String toString() {
		String res = "";
		int n = graph.n;
		List nodes = graph.nodes;
		for(int i=0; i<n; i++) {
			ConcreteSpider sp1 = (ConcreteSpider)nodes.get(i);
			String label1 = sp1.getLabel();
			for(int j=0; j<n; j++) {
				if (graph.edges[i][j]) {
					ConcreteSpider sp2 = (ConcreteSpider)nodes.get(j);
					String label2 = sp2.getLabel();
					res += "("+label1+", "+label2+")\n";
				}
			}
		}
		return res;
	}
	
	/*
	 * pre:  zones = set of concrete zones to be scanned for derived contours
	 * post: returns set of concrete derived contours in zones
	 */
	/*
	 * tests:
	 *   zones is empty
	 *   zones contains two zones
	 *     each zone has at least one non-derived concrete contour 
	 *       and one derived concrete contour 
	 *   
	 */
	private Collection derivedContoursOf(Collection zones) {
		Collection result = new LinkedHashSet();
		Iterator zit = zones.iterator();
		while (zit.hasNext()) {
			ConcreteZone cz = (ConcreteZone) zit.next();
			Collection polys = cz.getAllPolygonals();
			Iterator pit = polys.iterator();
			while (pit.hasNext()) {
				Polygonal p = (Polygonal) pit.next();
				if (p instanceof ConcreteContour) {
					ConcreteContour cc = (ConcreteContour) p;
					if (cc.isDerived())
						result.add(cc);
				}
			}
		}
		return result;
	}

	/** Concrete diagram */
	protected ConcreteUnitaryDiagram cud;
	protected UnitaryDiagram ud;
	/** Dependence relation */
	protected Graph graph;
	
	// test program
    // ------------

    private static void pr(String s) {
      System.out.println(s);
    }

    private static void hd(String s) {
      System.out.println("\n" + s + "\n");
    }

    private static void pg() {
      try {
        int i;
        System.out.print("next page? (CR)");
        do i = System.in.read(); while (i != '\n');
        System.out.println();
      }
      catch (Exception e) {System.out.println("raised exception : " + e);}
    }

    public static void main(String args[]) throws Exception {
    	Polygon p = new Polygon();
    	DecorationBox db = new DecorationBox("db","",p);
		ConcreteUnitaryDiagram cud = new ConcreteUnitaryDiagram(db);
		DependenceGraph dg = new DependenceGraph(cud);
		pr("dg="+dg);
		pg();
		
    	{
    		hd("Tests of derivedContoursOf");
    		ArrayList zones = new ArrayList();
    		pr("zones="+zones);
    		pr("dg.derivedContoursOf(zones)="+dg.derivedContoursOf(zones));
    		pg();
    		
    		p.addPoint(100,100);
    		ConcreteContour a = 
    			new ConcreteContour(Contour.getContour("a"),p);
    		pr("a="+a);
    		ConcreteContour derived_b = 
    			new ConcreteContour(Contour.getContour("derived-b"),p);
    		pr("derived_b="+derived_b);
    		ConcreteContour derived_c = 
    			new ConcreteContour(Contour.getContour("derived-c"),p);
    		pr("derived_c="+derived_c);
    		ConcreteContour d = 
    			new ConcreteContour(Contour.getContour("d"),p);
    		pr("d="+d);
    		
    		Vector va = new Vector();
    		va.add(a);
    		Vector vderived_b = new Vector();
    		vderived_b.add(derived_b);
    		Vector vderived_c = new Vector();
    		vderived_c.add(derived_c);
    		Vector vd = new Vector();
    		vd.add(d);
    		
    		ConcreteZone cz1 = new ConcreteZone(va,vderived_b);
    		ConcreteZone cz2 = new ConcreteZone(vderived_c,vd);
    		zones.add(cz1);
    		zones.add(cz2);
    		pr("zones="+zones);
    		pr("dg.derivedContoursOf(zones)="+dg.derivedContoursOf(zones));
    		pg();
    	}
    	
    }

}
