package rwd.ocl.spiderdiagram;

import java.io.FileReader;
import java.util.Vector;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import rwd.abstractsyntax.eulerdiagram.Contour;
import rwd.abstractsyntax.eulerdiagram.Zone;
import rwd.abstractsyntax.spiderdiagram.Spider;
import rwd.abstractsyntax.spiderdiagram.UnitaryDiagram;
import rwd.ocl.eulerdiagram.Uml2EulerDiagram;
import rwd.ocl.spiderdiagram.parser.OclParser;
import rwd.ocl.spiderdiagram.parser.OclParserImpl;

import uk.ac.kent.cs.kmf.util.ILog;
import uk.ac.kent.cs.ocl20.syntax.ast.contexts.ConstraintAS;
import uk.ac.kent.cs.ocl20.syntax.ast.contexts.ContextDeclarationAS;
import uk.ac.kent.cs.ocl20.syntax.ast.contexts.VariableDeclarationAS;
import uk.ac.kent.cs.ocl20.syntax.ast.expressions.ArrowSelectionExpAS;
import uk.ac.kent.cs.ocl20.syntax.ast.expressions.DotSelectionExpAS;
import uk.ac.kent.cs.ocl20.syntax.ast.expressions.IntegerLiteralExpAS;
import uk.ac.kent.cs.ocl20.syntax.ast.expressions.IteratorExpAS;
import uk.ac.kent.cs.ocl20.syntax.ast.expressions.LetExpAS;
import uk.ac.kent.cs.ocl20.syntax.ast.expressions.LogicalExpAS;
import uk.ac.kent.cs.ocl20.syntax.ast.expressions.NotExpAS;
import uk.ac.kent.cs.ocl20.syntax.ast.expressions.OclExpressionAS;
import uk.ac.kent.cs.ocl20.syntax.ast.expressions.OperationCallExpAS;
import uk.ac.kent.cs.ocl20.syntax.ast.expressions.PathNameExpAS;
import uk.ac.kent.cs.ocl20.syntax.ast.expressions.SelectionExpAS;

/**
 * @author Octavian Patrascoiu
 *
 */
public class Ocl2SpiderDiagram
	extends Uml2EulerDiagram 
{
	/** Constructor */
	public Ocl2SpiderDiagram(ILog log) {
		super(log);
	}
	
	/** Generate the spider diagram */
	public UnitaryDiagram makeSD(String oclFileName) {
		UnitaryDiagram ud = null;
		try {			
			// Parse the input and compute the oclExpression
			FileReader oclReader = new FileReader(oclFileName);
			ContextDeclarationAS context = parser.parse(oclReader, getLog());
			OclExpressionAS body = ((ConstraintAS)context.getConstraints().get(0)).getBodyExpression();
			oclReader.close();
			
			// Make contours
			Vector contours = makeContours(body);
		
			// Make all zones and shaded zones
			Vector zones = makeZones(body, contours);
			Vector shadedZones = computeShadedZones(body, contours);
			
			// Make spiders
			List spiders = makeSpiders(body, zones); 
		
			// Make the unitary diagram
			ud = new UnitaryDiagram(new Vector(zones), "ud");
			ud.setShadedZones(new Vector(shadedZones));
			ud.getSpiders().addAll(spiders);

			// Print result
			getLog().reportMessage(ud.toDebugString());
		} catch (Exception e) {
			getLog().reportError("Input error "+e.toString());
		}
		return ud;
	}
	
	/** Make contours */
	protected Vector makeContours(OclExpressionAS letExp) {
		Vector contours = new Vector();
		// Scan every variable declaration
		List vars = ((LetExpAS)letExp).getVariables();
		for(int i=0; i<vars.size(); i++) {
			// Get init expression
			VariableDeclarationAS var = (VariableDeclarationAS)vars.get(i);
			OclExpressionAS init = var.getInitExp();
			// Scan all filters
			IteratorExpAS callExp = (IteratorExpAS)init;
			LogicalExpAS andExp = (LogicalExpAS)callExp.getBody();
			Stack filters = new Stack();
			filters.push(andExp);
			while (!filters.isEmpty()) {
				OclExpressionAS exp = (OclExpressionAS)filters.pop();
				// Filter expression: oclIsKindOf call 
				if (exp instanceof OperationCallExpAS){
					OperationCallExpAS callExp1 = (OperationCallExpAS)exp;
					PathNameExpAS pathExp = (PathNameExpAS)callExp1.getArguments().get(0);
					List pathName = pathExp.getPathName();
					String filter = makeString(pathName, ".");
					// Create a contour and add it
					Contour cont = Contour.getContour(filter);
					if (!contours.contains(cont)) contours.add(cont);					
				// NOT filter expression
				} else if (exp instanceof NotExpAS) {
					LogicalExpAS logExp = (LogicalExpAS)exp;
					filters.push(logExp.getLeftOperand());
				// AND expression 
				} else if (exp instanceof LogicalExpAS) {
					LogicalExpAS logExp = (LogicalExpAS)exp;
					filters.push(logExp.getLeftOperand());
					filters.push(logExp.getRightOperand());
				}
			}
		}
		return contours;
	}

	/** Make zones */
	protected Map varName2zone = new HashMap();
	protected Map zone2varName = new HashMap();
	protected Vector makeZones(OclExpressionAS letExp, Vector contours) {
		Vector zones = new Vector();
		// Scan every variable declaration and make a zone
		List vars = ((LetExpAS)letExp).getVariables();
		for(int i=0; i<vars.size(); i++) {
			// Initialize in and out contours
			Vector in = new Vector();
			Vector out = new Vector();
			// Get init expression
			VariableDeclarationAS var = (VariableDeclarationAS)vars.get(i);
			OclExpressionAS init = var.getInitExp();
			// Scan all filters
			IteratorExpAS callExp = (IteratorExpAS)init;
			LogicalExpAS andExp = (LogicalExpAS)callExp.getBody();
			Stack filters = new Stack();
			filters.push(andExp);
			while (!filters.isEmpty()) {
				OclExpressionAS exp = (OclExpressionAS)filters.pop();
				// Filter expression: oclIsKindOf call 
				if (exp instanceof OperationCallExpAS){
					OperationCallExpAS callExp1 = (OperationCallExpAS)exp;
					PathNameExpAS pathExp = (PathNameExpAS)callExp1.getArguments().get(0);
					List pathName = pathExp.getPathName();
					String filter = makeString(pathName, ".");
					// Create a contour and add it to in
					Contour cont = Contour.getContour(filter);
					if (!in.contains(cont)) in.add(cont);					
				// NOT filter expression
				} else if (exp instanceof NotExpAS) {
					LogicalExpAS logExp = (LogicalExpAS)exp;
					OperationCallExpAS callExp1 = (OperationCallExpAS)logExp.getLeftOperand();
					PathNameExpAS pathExp = (PathNameExpAS)callExp1.getArguments().get(0);
					List pathName = pathExp.getPathName();
					String filter = makeString(pathName, ".");
					// Create a contour and add it to out
					Contour cont = Contour.getContour(filter);
					if (!out.contains(cont)) out.add(cont);					
				// AND expression 
				} else if (exp instanceof LogicalExpAS) {
					LogicalExpAS logExp = (LogicalExpAS)exp;
					filters.push(logExp.getLeftOperand());
					filters.push(logExp.getRightOperand());
				}
			}
			// Make a zone, add it, and store the name-zone mapping
			// except the outer zone
			if (in.size() != 0) {
				Zone zone = Zone.getZone(in, out);
				if (!zones.contains(zone)) {
					zones.add(zone); 
					varName2zone.put(var.getName(), zone);
					zone2varName.put(zone, var.getName());
				}
			}
		}
		
		// Add outer zone
		Zone outer = Zone.getZone(new Vector(), new Vector(contours));
		if (!zones.contains(outer)) {
			zones.add(outer);
			varName2zone.put("setOut", outer);
			zone2varName.put(outer, "setOut");
		} 
		
		return zones;
	}
	
	/** Make shaded zones */
	protected Map varName2size = new HashMap();
	protected Vector computeShadedZones(OclExpressionAS letExp, Vector contours) {
		Vector shadedZones = new Vector();
		// Scan every relational expression from body
		OclExpressionAS andExp = ((LetExpAS)letExp).getIn();
		Stack sizes = new Stack();
		sizes.push(andExp);
		while (!sizes.isEmpty()) {
			OclExpressionAS exp = (OclExpressionAS)sizes.pop();
			// Relation expression
			if (exp instanceof OperationCallExpAS){
				// Compute the boundary
				OperationCallExpAS callExp1 = (OperationCallExpAS)exp;
				DotSelectionExpAS selExp1 = (DotSelectionExpAS)callExp1.getSource();
				IntegerLiteralExpAS intExp = (IntegerLiteralExpAS)callExp1.getArguments().get(0);
				int size = intExp.getValue().intValue();
				String operName = selExp1.getName();
				// Set mapping
				varName2size.put(getVarName(exp), new Integer(size));
				// Add shaded zones: size is zero or relation is <=
				if (size == 0 || operName.equals("<=")) {
					OperationCallExpAS callExp2 = (OperationCallExpAS)selExp1.getSource();
					ArrowSelectionExpAS selExp2 = (ArrowSelectionExpAS)callExp2.getSource();
					OclExpressionAS source = selExp2.getSource();
					if (source instanceof PathNameExpAS) {
						PathNameExpAS pathNameExp = (PathNameExpAS)source;
						String varName = makeString(pathNameExp.getPathName(), ".");
						Zone zone = (Zone)varName2zone.get(varName);
						shadedZones.add(zone);
					}
				}
			// AND expression 
			} else if (exp instanceof LogicalExpAS) {
				LogicalExpAS logExp = (LogicalExpAS)exp;
				sizes.push(logExp.getLeftOperand());
				sizes.push(logExp.getRightOperand());
			}
		}
		return shadedZones;
	}

	/** Get the name of the variable used in a relation expression */
	protected String getVarName(OclExpressionAS exp) {
		while (true) {
			if (exp instanceof OperationCallExpAS) {
				exp = ((OperationCallExpAS)exp).getSource();
			} else if (exp instanceof DotSelectionExpAS) {
				exp = ((DotSelectionExpAS)exp).getSource();
			} else if (exp instanceof ArrowSelectionExpAS) {
				exp = ((ArrowSelectionExpAS)exp).getSource();
			} else if (exp instanceof PathNameExpAS) {
				return (String)((PathNameExpAS)exp).getPathName().get(0);
			} else {
				return "";
			}
		}
	}
	
	/** Make spiders from a relational expression */
	protected Map zone2spiders = new HashMap();
	protected List makeSpiders(OclExpressionAS letExp, List zones) {
		List spiders = new Vector();
		//
		// Add all spiders with more than one foot
		//
		// Scan all relational expression which contain union expressions
		// and compute for each zone the number of spiders touched by
		OclExpressionAS andExp = ((LetExpAS)letExp).getIn();
		Stack sizes = new Stack();
		sizes.push(andExp);
		while (!sizes.isEmpty()) {
			OclExpressionAS exp = (OclExpressionAS)sizes.pop();
			// Relation expression
			if (exp instanceof OperationCallExpAS){
				// Skip relational level
				OperationCallExpAS callExp1 = (OperationCallExpAS)exp;
				DotSelectionExpAS selExp1 = (DotSelectionExpAS)callExp1.getSource();
				// Skip size level
				OperationCallExpAS callExp2 = (OperationCallExpAS)selExp1.getSource();
				ArrowSelectionExpAS selExp2 = (ArrowSelectionExpAS)callExp2.getSource();
				// Union or varName level
				OclExpressionAS source = selExp2.getSource();
				// Union level
				if (source instanceof OperationCallExpAS) {
					// Compute zones from union expression
					List habitat = new Vector(computeZonesFromUnion(source));
					// Create a spider and add it
					Spider spider = new Spider(new Vector(habitat));
					spiders.add(spider);
					// Set zone-spiders mapping
					for(int i=0; i<habitat.size(); i++) {
						Zone zone = (Zone)habitat.get(i);
						Set set = (Set)zone2spiders.get(zone);
						if (set == null) set = new LinkedHashSet();
						set.add(spider);
						zone2spiders.put(zone, set);
					}
				}
			// AND expression 
			} else if (exp instanceof LogicalExpAS) {
				LogicalExpAS logExp = (LogicalExpAS)exp;
				sizes.push(logExp.getLeftOperand());
				sizes.push(logExp.getRightOperand());
			}
		}
		//
		// Add spiders with only one foot
		//
		// Scan all zones andd add spiders with only one foot which touch the zone
		//
		for(int i=0; i<zones.size(); i++) {
			// Compute the number of elements from zone
			Zone zone = (Zone)zones.get(i);
			String varName = (String)zone2varName.get(zone);
			Integer size = (Integer)varName2size.get(varName);
			int noFeet = 0;
			if (size != null) noFeet = size.intValue();
			// Compute the number of spiders that inhabit zone
			Set set = (Set)zone2spiders.get(zone);
			int noSpider = 0;
			if (set != null) noSpider = set.size();
			// Create required one foot spiders 
			for(int j=noSpider; j<noFeet; j++) {
				Vector habitat = new Vector();
				habitat.add(zone);
				Spider spider = new Spider(habitat);
				spiders.add(spider);
			}
		}
		return spiders;
	}

	/** Compute all zones used in a union expression */
	protected Set computeZonesFromUnion(OclExpressionAS exp) {
		Set zones = new LinkedHashSet();
		while (exp instanceof OperationCallExpAS) {
			// Get the argument
			OperationCallExpAS call = (OperationCallExpAS)exp; 
			List args = call.getArguments();
			PathNameExpAS arg = (PathNameExpAS)args.get(0);
			String varName = (String)arg.getPathName().get(0);
			Zone zone = (Zone)varName2zone.get(varName);
			zones.add(zone);
			// Descend one level
			SelectionExpAS selExp = (SelectionExpAS)call.getSource();
			exp = selExp.getSource();
		}
		// Add last zone
		PathNameExpAS arg = (PathNameExpAS)exp;
		String varName = (String)arg.getPathName().get(0);
		Zone zone = (Zone)varName2zone.get(varName);
		zones.add(zone);		
		return zones;
	}
	
	/** Make a string from a sequence using a separator */
	protected String makeString(List items, String sep) {
		String res = "";
		for(int i=0; i<items.size(); i++) {
			if (i != 0) res += sep;
			res += items.get(i);
		}
		return res;
	}

	/** Input OCL file name */
	private String oclFileName;
	protected String getOclFileName() {
		return oclFileName;
	}
	protected void setOclFileName(String string) {
		oclFileName = string;
	}

	/** OCL Parser */
	private OclParser parser = new OclParserImpl();	
}
