package rwd.ocl.spiderdiagram;

import java.text.SimpleDateFormat;
import java.util.Vector;
import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.TreeSet;

import rwd.abstractsyntax.eulerdiagram.Contour;
import rwd.abstractsyntax.eulerdiagram.Zone;
import rwd.abstractsyntax.spiderdiagram.Diagram;
import rwd.abstractsyntax.spiderdiagram.Spider;
import rwd.abstractsyntax.spiderdiagram.UnitaryDiagram;
import rwd.abstractsyntax.spiderdiagram.Utilities;
import rwd.concretesyntax.ConcreteDiagram;
import rwd.concretesyntax.ConcreteUnitaryConverter;
import rwd.concretesyntax.ConcreteUnitaryDiagram;
import rwd.concretesyntax.eulerdiagram.ConcreteZone;
import rwd.concretesyntax.eulerdiagram.ConcreteContour;
import rwd.concretesyntax.eulerdiagram.DecorationBox;
import rwd.tableaux.diagram.SpiderSplitter;
import rwd.tableaux.diagram.SpiderSplitterImpl;
import rwd.tableaux.semantic.OptionsImpl;
import rwd.tableaux.syntax.ast.AndExp;
import rwd.tableaux.syntax.ast.AtomicExp;
import rwd.tableaux.syntax.ast.EquivalentExp;
import rwd.tableaux.syntax.ast.ImpliesExp;
import rwd.tableaux.syntax.ast.LogicalExp;
import rwd.tableaux.syntax.ast.NotExp;
import rwd.tableaux.syntax.ast.OrExp;
import rwd.tableaux.syntax.parser.Parser;
import rwd.tableaux.syntax.parser.ParserImpl;
import uk.ac.kent.cs.kmf.util.ILog;

/**
 * @author Octavian Patrascoiu
 *
 */
public class SpiderDiagram2UmlOcl {
	/** Constructor */
	public SpiderDiagram2UmlOcl(List diagrams, ILog log) {
		// Initialize input members
		this.cds = new LinkedHashSet(diagrams);
		this.log = log;
		// Compute cuds, cd, and ud
		this.cuds = new LinkedHashSet();
		Iterator i = cds.iterator();
		while (i.hasNext()) {
			ConcreteDiagram cd = (ConcreteDiagram)i.next();
			Set cuds = cd.getConcreteUnitaryDiagrams();
			this.cuds.addAll(cuds);
		}
	}

	/** Debug/release flag */
	public boolean isDebug() {
		return debug;
	}
	public void setDebug(boolean b) {
		debug = b;
	}

	/** Generate the equivalent OCL */
	public String makeOCL(String stringExp) {
		LogicalExp exp = parser.parse(stringExp, log);
		String oclStr = getOCL(exp);
		if (oclStr == null || oclStr.length() == 0) {
			oclStr = "true";
		} 
		return "context OclVoid inv: "+oclStr;
	}
	/** Compute the OCL string for a logical expression */
	protected String getOCL(LogicalExp exp) {
		if (exp instanceof AtomicExp) {
			String name = ((AtomicExp)exp).getName();
			// Simple cases
			if (name.equals("true")) return "true";
			else if (name.equals("false")) return "false";
			else return getOCL4UnitaryDiagram(name);
		} else if (exp instanceof NotExp) {
			return " not "+getOCL(exp.getLeft());
		} else {
			// Generate for left and right Add left and right expressions
			LogicalExp leftExp = exp.getLeft();
			LogicalExp rightExp = exp.getRight();
			String leftStr = getOCL(leftExp);
			String rightStr = getOCL(rightExp);
			// P and Q goes to P, Q
			if (exp instanceof AndExp) 
				return leftStr + " and " + rightStr;
			else if (exp instanceof OrExp) 
				return leftStr + " or " + rightStr;
			else if (exp instanceof ImpliesExp) 
				return leftStr + ".not() or " + rightStr;
			else if (exp instanceof EquivalentExp) 
				return "("+leftStr + ".not() or " + rightStr + ") and ("+leftStr + " or " + rightStr + ".not())";
			return null;
		}
	}
	/** Compute the OCL string for a unitary diagram with a given name */
	protected String getOCL4UnitaryDiagram(String name) {
		// Compute cd, cud, and all the concreteZones
		ConcreteDiagram cd = null;
		ConcreteUnitaryDiagram cud = null;
		// Compute cuds, cd, and ud
		Iterator i = cds.iterator();
		while (i.hasNext()) {
			ConcreteDiagram cdTemp = (ConcreteDiagram)i.next();
			Set cuds = cdTemp.getConcreteUnitaryDiagrams();
			this.cuds.addAll(cuds);
			Iterator j = cuds.iterator();
			ConcreteUnitaryDiagram cudTemp = null;
			while (j.hasNext()) {
				cudTemp = (ConcreteUnitaryDiagram)j.next();
				if (cudTemp.getName().equals(name)) {
					cd = cdTemp;
					cud = cudTemp;
					break;
				}
			}
		}
		// Check if the diagram exists
		if (cd == null || cud == null) {
			log.reportError("Cannot find diagram '"+name+"'");
			return null;
		}
		
		// Compute ocl
		List vars = new Vector();
		List exps = new Vector();
		// Compute all concrete zones
		Set allConcreteZones = makeAllConcreteZones(cud);
		// Compute variables
		vars = getFullVars(allConcreteZones);
		// Compute body
		exps = getFullBodyExp(cud);
		
		// Compute invariant
		String ocl = "let \n";
		ocl += "\t"+makeString(vars, ",\n\t")+"\n";
		ocl += "in\n";
		String body = makeString(exps, " or \n\t");
		if (body == null || body.length() == 0) body = "true"; 
		ocl += "\t"+body+"\n";
		return ocl;
	}

	/** Compute all the concrete zones within the concrete diagram */
	protected Set makeAllConcreteZones(ConcreteUnitaryDiagram cud) {
		Set result = new LinkedHashSet();
		// collect together contours and boxes
		Vector polygonals = new Vector();
		polygonals.addAll(cud.getConcreteContours());
		DecorationBox box = cud.getDecorationBox(); 
		// for each subset of polygons
		List powerSet = Utilities.powerSet(polygonals);
		for (int i = 0; i < powerSet.size(); i++) {
			Vector subset = (Vector) powerSet.get(i);
			Vector rest = (Vector) polygonals.clone();
			rest.removeAll(subset);
			// Compute the corresponding concrete zone
			if (subset.size() > 0) {
				ConcreteZone cz = new ConcreteZone(subset, rest);
				result.add(cz);
			}
		}
		return result;
	}
/*
	protected Set computeAllAbstractZones(ConcreteUnitaryDiagram cud) {
		Set result = new LinkedHashSet();
		Set zoneAreas = cud.getConcreteZones();
		Iterator i = zoneAreas.iterator();
		while (i.hasNext()) {
			ConcreteZone zoneArea = (ConcreteZone)i.next();
			result.add(zoneArea.getAbstractZone());
		}
		return result;		
	}
    */

//	/** Compute all shaded abstract zones */
//	protected Set computeAllAbstractShadedZones(ConcreteUnitaryDiagram cud) {
//		Set result = new LinkedHashSet();
//		Set zoneAreas = cud.getSZoneAreas();
//		Iterator i = zoneAreas.iterator();
//		while (i.hasNext()) {
//			ConcreteZone zoneArea = (ConcreteZone)i.next();
//			result.add(zoneArea.getAbstractZone());
//		}
//		return result;		
//	}

//	/** Compute all the variables */
//	protected List getVars(Set allZones, Set shadedZones) {
//		Set vars = new LinkedHashSet();
//		// Generate variable for each zone
//		Collection zones = allZones;
//		Iterator i = zones.iterator();
//		while (i.hasNext()) {
//			Zone zone = (Zone)i.next();
//			// Compute the corresponding set of elements
//			String varName = getVarName(zone);
//			String init = getInitValue(zone);
//			vars.add(varName+": Set(OclAny) = "+init);
//		}
//		return new Vector(vars);
//	}

	/** Compute all the variables */
	protected List getFullVars(Set allConcreteZones) {
		Set vars = new LinkedHashSet();
		// Generate variable for each zone
		Collection zones = allConcreteZones;
		Set contours = new TreeSet();
		Iterator i = zones.iterator();
		while (i.hasNext()) {
			ConcreteZone zone = (ConcreteZone)i.next();
			contours.addAll(zone.getAbstractZone().getAllContours());
			// Compute the corresponding set of elements
			String varName = getVarName(zone);
			String init = getInitValue(zone);
			vars.add(varName+": Set(OclAny) = "+init);
		}
		// Add out variable
		String initOut = getInitValue4Out(new Vector(contours));
		vars.add("out :Set(OclAny) = "+initOut);
		return new Vector(vars);
	}

	/** Compute the OCL variable name attached to a zone */
	protected String getVarName(ConcreteZone cz) {
		return "set"+getUMLType(cz.getAbstractZone());
	}

	/** Compute the OCL variable name attached to a zone */
	protected String getVarName(Zone az) {
		return "set"+getUMLType(az);
	}

	/** Compute the UML type associated with a zone */
	protected String getUMLType(Zone az) {
		// Get contours
		return getUMLType(az.getContainingContours(), az.getExcludingContours());
	}

	/** Sort a list of contours */
	protected List sort(List list) {
		// Sort them
		boolean ok = false;
		do {
			ok = true;
			for(int i=0; i<list.size()-1; i++) {
				Comparable c1 = (Comparable)list.get(i);
				Comparable c2 = (Comparable)list.get(i+1);
				if (c1.compareTo(c2) > 0) {
					list.set(i, c2);
					list.set(i+1, c1);
					ok = false;
				}
			}
			
		}
		while (!ok);
		return list;
	}

	/** Compute the initial value of the variable */
	protected String getUMLType(Contour c) {
		return modelName+"::"+prefix+c.getLabel();
	}
	protected String getInitValue(ConcreteZone cz) {
		// Compute initial collection
		String set = "OclAny.allInstances()";
		// Compute selection predicate
		List selPred = new Vector();
		Zone az = null;
        az = cz.getAbstractZone();
        
		List inCont = az.getContainingContours();
		for(int i=0; i<inCont.size(); i++) {
			Contour c1 = (Contour)inCont.get(i);
			String cName = getUMLType(c1);
			selPred.add("x.oclIsKindOf("+cName+")");
		}
		List exCont = az.getExcludingContours();
		for(int i=0; i<exCont.size(); i++) {
			Contour c1 = (Contour)exCont.get(i);
			String cName = getUMLType(c1);
			selPred.add(" not x.oclIsKindOf("+cName+")");
		}
		String selPredStr = "";
		for(int i=0; i<selPred.size(); i++) {
			if (i != 0) selPredStr += " and ";
			selPredStr += selPred.get(i);
		}
		// Else, compute selected set
		set += "->select(x : OclAny | "+selPredStr+")";
		return set;
	}
	protected String getInitValue4Out(Vector exCont) {
		// Compute initial collection
		String set = "OclAny.allInstances()";
		// Compute selection predicate
		List selPred = new Vector();
		for(int i=0; i<exCont.size(); i++) {
			Contour c1 = (Contour)exCont.get(i);
			String cName = getUMLType(c1);
			selPred.add(" not x.oclIsKindOf("+cName+")");
		}
		String selPredStr = "";
		for(int i=0; i<selPred.size(); i++) {
			if (i != 0) selPredStr += " and ";
			selPredStr += selPred.get(i);
		}
		// Else, compute selected set
		set += "->select(x : OclAny | "+selPredStr+")";
		return set;		
	}
	protected String getInitValue(Zone az) {
		// Compute initial collection
		String set = "OclAny.allInstances()";
		// Compute selection predicate
		List selPred = new Vector();
		List inCont = az.getContainingContours();
		for(int i=0; i<inCont.size(); i++) {
			Contour c1 = (Contour)inCont.get(i);
			String cName = getUMLType(c1);
			selPred.add("x.oclIsKindOf("+cName+")");
		}
		List exCont = az.getExcludingContours();
		for(int i=0; i<exCont.size(); i++) {
			Contour c1 = (Contour)exCont.get(i);
			String cName = getUMLType(c1);
			selPred.add(" not x.oclIsKindOf("+cName+")");
		}
		String selPredStr = "";
		for(int i=0; i<selPred.size(); i++) {
			if (i != 0) selPredStr += " and ";
			selPredStr += selPred.get(i);
		}
		// Else, compute selected set
		set += "->select(x : OclAny | "+selPredStr+")";
		return set;
	}

	/** Compute the body of the invariant for a concrete unitary diagram */
	protected List getFullBodyExp(ConcreteUnitaryDiagram cud) {
		// Split spiders
		SpiderSplitter splitter = new SpiderSplitterImpl(new OptionsImpl(), log);
		UnitaryDiagram aud = ConcreteUnitaryConverter.makeAUD(cud);
		Diagram d = splitter.splitSpiders(aud);
		// Get list of unitary diagrams
		List auds = d.getUnitaryParts();
		List exps = new Vector();
		// Make an or of ands
		for(int i=0; i<auds.size(); i++) {
			String andExp = getBodyExp((UnitaryDiagram)auds.get(i));
			if (andExp.length() != 0) 
				exps.add(andExp);
		}
		return exps;
	}
	protected String getBodyExp(UnitaryDiagram d) {
		List zones = d.getCloneOfZoneSet();
		List sZones = d.getCloneOfShadedZones();
		List spiders = d.getSpiders();
		List result = new Vector();
		for(int i=0; i<zones.size(); i++) {
			Zone az = (Zone)zones.get(i);
			// Compute number of feet and isShaded
			int feetNo = 0;
			for(int j=0; j<spiders.size(); j++) {
				Spider spider = (Spider)spiders.get(j);
				Collection habitat = spider.getHabitat();
				if (habitat.contains(az))
					feetNo++;
			}
			boolean isShaded = sZones.contains(az);
			// Compute expression
			String varName = getVarName(az);
			String exp = "(";
			if (isShaded) {
				if (feetNo == 0) exp += varName+"->size() = 0";
				else exp += varName+"->size() = "+feetNo;
			} else {
				if (feetNo == 0);
				else exp += varName+"->size() >= "+feetNo;
			}
			exp += ")";
			if (!exp.equals("()")) {
				result.add(exp);
			}
		}
		String str = makeString(result, " and ");
		return str;	
	}

	/** Generate the equivalent model - XMI */
	public String makeXMI(String stringExp) {
		// Compute all the diagram names
		Set diagramNames = new LinkedHashSet();
		StringTokenizer strTok = new StringTokenizer(stringExp, " <=>", false);
		while (strTok.hasMoreTokens()) {
			String token = strTok.nextToken();
			if (token.equals("true"));
			else if (token.equals("false"));
			else if (token.equals("not"));
			else if (token.equals("and"));
			else if (token.equals("or"));
			else diagramNames.add(token);
		}
		// Compute all the concrete unitary diagrams used in expression
		Set cuds = new LinkedHashSet();
		Iterator i = diagramNames.iterator();
		while (i.hasNext()) {
			// Compute the diagram
			String dn = (String)i.next();
			ConcreteUnitaryDiagram cud = getConcreteUnitaryDiagram(this.cuds, dn);
			if (cud == null) {
				log.reportError("Cannot find diagram '"+dn+"'");				
			} else {
				cuds.add(cud);
			}
		}
		// Compute all the UML types
		Set umlTypes = getAllUMLTypes(cuds);
		// Compute date
		Calendar cal = Calendar.getInstance();
		String DATE_FORMAT = "dd MMMMM yyyy HH:mm:ss";
		SimpleDateFormat dateFormatter = new java.text.SimpleDateFormat(DATE_FORMAT);
		Date date = cal.getTime();
		// Compute the xmi
		String xmi = "";
		xmi += "<?xml version = '1.0' encoding = 'UTF-8' ?>\n";
		xmi += "<XMI xmi.version = '1.2' xmlns:UML = 'org.omg.xmi.namespace.UML' timestamp = '"+dateFormatter.format(date)+"'>\n";
		xmi += getXMIHeader();
		xmi += getXMIContent(new Vector(umlTypes));
		xmi += "</XMI>\n";
		return xmi;
	}
	
	/** Search for a CUD */
	protected ConcreteUnitaryDiagram getConcreteUnitaryDiagram(Collection cuds, String name) {
		Iterator i = cuds.iterator();
		while (i.hasNext()) {
			ConcreteUnitaryDiagram cud = (ConcreteUnitaryDiagram)i.next();
			if (cud.getName().equals(name)) {
				return cud;
			}
		}
		return null;
	}

	/** Compute all the UML types for a set of CUDs */
	protected Set getAllUMLTypes(Set cuds) {
		Set types = new LinkedHashSet();
		// Collect together contour names
		Iterator i = cuds.iterator();
		Set contourNames = new LinkedHashSet(); 
		while (i.hasNext()) {
			ConcreteUnitaryDiagram cud = (ConcreteUnitaryDiagram)i.next();
			if (cud != null) {
				Iterator j = cud.getConcreteContours().iterator();
				while (j.hasNext()) {
					ConcreteContour cp = (ConcreteContour)j.next();
					contourNames.add(cp.getLabel());
				}
			}
		}
		// Generate all the in/out combinations - abstract zones
		List powerSet = Utilities.powerSet(new Vector(contourNames));
		for (int k = 0; k < powerSet.size(); k++) {
			Vector subset = (Vector) powerSet.get(k);
			Vector rest = new Vector(contourNames);
			rest.removeAll(subset);
			// Compute the corresponding UML type
			String umlType = getUMLType(subset, rest);
			types.add(umlType);
		}
		return types;
	}
	
	/** Compute the UML type associated with a zone */
	private String prefix = "";
	private String modelName = "RwD";
	protected String getUMLType(List in, List out) {
		// Get contours
		List inc = sort(in);
		// Make the name of the type
		String type = null;
		if (inc.isEmpty()) type = prefix+"Out";
		else type = prefix+makeString(inc, "_");
		return type;
	}
	
	/** Compute XMI header */
	protected String tab = "  ";
	protected String getXMIHeader() {
		// Compute header
		String header = "";
		header += "<XMI.header>\n";
		header += tab+"<XMI.documentation>\n";
		header += tab+tab+"<XMI.exporter>XMI Writer</XMI.exporter>\n";
		header += tab+tab+"<XMI.exporterVersion>1.0</XMI.exporterVersion>\n";
		header += tab+"</XMI.documentation>\n";
		header += "</XMI.header>\n";
		return header;
	}
	
	/** Compute content */
	protected String getXMIContent(List types) {
		String content = "";
		// Open content
		content += "<XMI.content>\n";
		// Open model
		content += tab+"<UML:Model xmi.id = '"+getNewId()+"' name = '"+modelName+"' isSpecification = 'false' isRoot = 'false' isLeaf = 'false' isAbstract = 'false'>\n";
		// Save an id for tagged value definition
		String tagDefRef = getNewId();
		// Add tagged value
		content += getXMI4TaggedValue(tab+tab, getNewId(), tagDefRef);
		// Open namespace
		content += tab+tab+"<UML:Namespace.ownedElement>\n";
		// Add all classes
		for(int i=0; i<types.size(); i++) {
			String type = (String)types.get(i);
			content += getXMI4Class(tab+tab+tab, type, getTypeId(type), tagDefRef);
		}
		// Add all inheritances
		for(int i=0; i<types.size(); i++) {
			String type = (String)types.get(i);
			content += getXMI4InheritanceRelations(tab+tab+tab, type, getTypeId(type), tagDefRef);
		}
		// Close Namespace
		content += tab+tab+"</UML:Namespace.ownedElement>\n";
		// Close model
		content += tab+"</UML:Model>\n";
		// Add tag value definition 
		content += getXMI4TagDefinition(tab, tagDefRef);
		// Close content		
		content += "</XMI.content>\n";
		return content;
	}
	
	/** Compute XMI for a taggged value */
	protected String getXMI4TaggedValue(String indent, String id, String idRef) {
		String res = "";
		res += indent+"<UML:ModelElement.taggedValue>\n";
		res += indent+tab+"<UML:TaggedValue xmi.id = '"+id+"' isSpecification = 'false' dataValue = '"+id+"'>\n";
		res += indent+tab+tab+"<UML:TaggedValue.type>\n";
		res += indent+tab+tab+tab+"<UML:TagDefinition xmi.idref = '"+idRef+"'/>\n";
		res += indent+tab+tab+"</UML:TaggedValue.type>\n";
		res += indent+tab+"</UML:TaggedValue>\n";
		res += indent+"</UML:ModelElement.taggedValue>\n";
		return res;
	}
	
	/** Compute XMI for a tagged value definition */
	protected String getXMI4TagDefinition(String indent, String id) {
		String res = "";
		res += indent+"<UML:TagDefinition xmi.id = '"+id+"' name = 'element.uuid' isSpecification = 'false' tagType = 'element.uuid'>\n";
		res += indent+tab+"<UML:TagDefinition.multiplicity>\n";
		res += indent+tab+tab+"<UML:Multiplicity xmi.id = '"+getNewId()+"'>\n";
		res += indent+tab+tab+tab+"<UML:Multiplicity.range>\n";
		res += indent+tab+tab+tab+tab+"<UML:MultiplicityRange xmi.id = '"+getNewId()+"' lower = '1' upper = '1'/>\n";
		res += indent+tab+tab+tab+"</UML:Multiplicity.range>\n";
		res += indent+tab+tab+"</UML:Multiplicity>\n";
		res += indent+tab+"</UML:TagDefinition.multiplicity>\n";
		res += indent+"</UML:TagDefinition>\n";
		return res;
	}
	
	/** Compute XMI for a Class */ 
	protected String getXMI4Class(String indent, String name, String id, String tagRef) {
		// Compute the xmi and return it 
		String res = "";
		res += indent+"<UML:Class xmi.id = '"+id+"' name = '"+name+"' visibility = 'public' isSpecification = 'false' isRoot = 'false' isLeaf = 'false' isAbstract = 'false' isActive = 'false'>\n";
		res += getXMI4TaggedValue(indent+tab, getNewId(), tagRef);
		res += getXMI4InheritedTypes(indent+tab, name);
		res += indent+"</UML:Class>\n";
		return res;
	}
	
	/** Compute all the super classes of a Class */ 
	protected List getSuperTypes(String typeName) {
		String types[] = typeName.split("_");
		List res = new Vector();
		for(int i=0; i<types.length; i++) res.add(types[i]);
		return res;
	}
	
	/** Compute XMI for super classes */
	protected String getXMI4InheritedTypes(String indent, String typeName) {
		String res = "";
		List types = getSuperTypes(typeName);
		if (types.size() >= 2) {
			res += indent+"<UML:GeneralizableElement.generalization>\n";
			for(int i=0; i<types.size(); i++) {
				String superType = (String)types.get(i);
				res += indent+tab+"<UML:Generalization xmi.idref = '"+getGeneralizationId(typeName+"->"+superType)+"'/>\n";
			}
			res += indent+"</UML:GeneralizableElement.generalization>\n";
		}
		return res;
	}
	
	/** Compute XMI for all inheritance relations */
	protected String getXMI4InheritanceRelations(String indent, String name, String id, String tagRef) {
		// Compute the xmi and return it 
		String res = "";
		// Compute the list of inherited types
		List superTypes = getSuperTypes(name);
		if (superTypes.size() >= 2) {
			// Scan all the inherited types and generate a inheritance relation
			for(int i=0; i<superTypes.size(); i++) {
				String superType = (String)superTypes.get(i);
				res += indent+"<UML:Generalization xmi.id = '"+getGeneralizationId(name+"->"+superType)+"' isSpecification = 'false'>\n";
				res += getXMI4TaggedValue(indent+tab, getNewId(), tagRef);
				res += indent+tab+"<UML:Generalization.child>\n";
				res += indent+tab+tab+"<UML:Class xmi.idref = '"+id+"'/>\n";
				res += indent+tab+"</UML:Generalization.child>\n";
				res += indent+tab+"<UML:Generalization.parent>\n";
				res += indent+tab+tab+"<UML:Class xmi.idref = '"+getTypeId(superType)+"'/>\n";
				res += indent+tab+"</UML:Generalization.parent>\n";
				res += indent+"</UML:Generalization>\n";
			}			
		}
		return res;
	}

	/** Provide unique ids */
	private int idNo = 0;
	protected Map type2id = new HashMap();
	protected Map generalization2id = new HashMap();
	protected String getNewId() {
		return "id-"+ ++idNo;
	}
	protected String getTypeId(String typeName) {
		String id = (String)type2id.get(typeName);
		if (id == null) {
			id = getNewId();
			type2id.put(typeName, id);
		} 
		return id;
	}
	protected String getGeneralizationId(String gName) {
		String id = (String)generalization2id.get(gName);
		if (id == null) {
			id = getNewId();
			generalization2id.put(gName, id);
		} 
		return id;
	}
	
	/** Make a string from a collection using the delimiter */
	protected String makeString(Collection objs, String del) {
		String res = "";
		Iterator i = objs.iterator();
		while (i.hasNext()) {
			Object obj = i.next();
			res += obj;
			if (i.hasNext()) res += del;
		}
		return res;		
	}

	// Members
	/** Debug/release flag */
	protected boolean debug = false;
	/** Logical expression parser */
	protected Parser parser = new ParserImpl();
	/** Concrete diagrams */
	protected Set cds;
	/** Concrete unitary diagrams */
	protected Set cuds;
	/** Log */
	protected ILog log;
	/** Display meaning for missing zones */
	protected static boolean MEANING_FOR_MISSING_ZONES = true;
}
