package rwd.rules;

import java.util.Collection;
import java.util.HashSet;
import java.util.List;

import rwd.abstractsyntax.spiderdiagram.Diagram;
import rwd.concretesyntax.*;
import rwd.concretesyntax.eulerdiagram.*;
import rwd.concretesyntax.spiderdiagram.ConcreteSpider;
import rwd.concretesyntax.spiderdiagram.ConcreteSpiderFoot;
import rwd.concretesyntax.spiderdiagram.ConcreteSpiderLeg;
import rwd.gui.editor.viewers.DiagramViewerInterface;
import rwd.rules.application.RuleApplicationSpec;

/**
 * A Constraint Diagram rule.
 * A constraint diagram rule operates on a List of arguments, representing the 
 * viewers, diagrams and diagram elements selected by the user, 
 * plus an optional series of extra arguments obtained by dialog with the user.
 * 
 * The class contains default methods for checking whether a rule applies to
 * the given arguments, and for applying a rule to the given arguments.
 * These methods should be overridden by subclasses.
 * 
 * This class contains protected methods for 
 * extracting the first viewer in the list,
 * the first concrete unitary diagram in the list, 
 * and the set of selected zones in the list.
 * 
 * It also contains protected methods for extracting a spider from a concrete
 * unitary diagram.
 * 
 * As CDRule currently derives from Rule (instead of being a sibling) it contains
 * dummy methods to fulfil the requirements of the Rule interface.
 * 
 * @author op, rnb
 * @version 10.6.05   10:30
 *
 */
public class CDRule 
	extends Rule
{
	
	// constructor
	// -----------
	
	public CDRule(String name) {
		super(name);
	}

	
	// commands
	// --------
	
	/**
	 * This is a default method to be overridden by subclasses.
	 * @return true iff this rule needs extra arguments from the user
	 */
	public boolean needsExtraArgs() {
		return false;
	}
	
	/**
	 * This is a default method to be overridden by subclasses.
	 * @param args list of arguments, to a sublist of which this rule may apply 
	 * @param extraArgs string representing extra arguments to which this rule may apply
	 * @return true iff this rule applies to the specified components and extra arguments
	 */
    public boolean canApplyTo(java.util.List args, String extraArgs) {
    	return false;
    }
    
    /**
	 * This is a default method to be overridden by subclasses.
     * @param args list of arguments, to a sublist of which this rule must apply
     * @param extraArgs string representing extra arguments to which this rule must apply
     * @return diagram representing the result of applying this rule
     */
    public ConcreteUnitaryDiagram applyRule(List args, String extraArgs) {
    	return getCUD(args);
    }
   
    
    // dummy methods to satisfy Rule interface
    // ---------------------------------------

    /**
     * This is a dummy method to fulfil the requirements of the Rule superclass.
     * It would be better to have a common superclass of which Rule and CDRule
     * were subclasses.
     * @param part diagram to which this rule is applicable
     * @return rule application spec for part, or null if none
     */
    public RuleApplicationSpec findApplicableSpec(Diagram part) {
		return null;
	}

    /**
     * Check applicability of rule for spider diagrams.
     * This is a dummy method to fulfil the requirements of the Rule superclass.
     * It would be better to have a common superclass of which Rule and CDRule
     * were subclasses.
     * @param ad diagram to which rule is to be applied
     * @return true iff this rule can be applied to ad
     */
	public boolean canApplyTo(Diagram ad) {
		return false;
	}

    /**
     * Applies a spider diagram rule.
     * This is a dummy method to fulfil the requirements of the Rule superclass.
     * It would be better to have a common superclass of which Rule and CDRule
     * were subclasses.
     * @param r application spec for rule
     * @return resultant diagram
     */
	public Diagram applyRule(RuleApplicationSpec r) {
		return null;
	}

    /**
     * This is a dummy method to fulfil the requirements of the Rule superclass.
     * It would be better to have a common superclass of which Rule and CDRule
     * were subclasses.
     * @return true iff this rule is invertible
     */
	public boolean isInvertible() {
		return false;
	}
	
	
	// protected utility methods for subclasses
    // ----------------------------------------
	
    /*
     * pre:  args = list of arguments for this rule, or null
     * post: returns first viewer in args, or null iff none
     */
	/*
	 * test:
	 *   null args
	 *   no viewer in args
	 *   two viewers in args
	 */
    protected DiagramViewerInterface getViewer(List args) {
    	if (args == null) return null;
    	DiagramViewerInterface viewer = null;
    	for(int i=0; i<args.size(); i++) {
    		Object obj = args.get(i);
    		if (obj instanceof DiagramViewerInterface) {
    			viewer = (DiagramViewerInterface)obj;
    			break;
    		}
    	}
   		return viewer;
    }

    
    /*
     * pre:  args = list of arguments for this rule, or null
     * post: returns first concrete unitary diagram from argument list, 
     *         or null iff none
     *       _viewer = first viewer in args, or null iff none
     */
    /*
     * test:
     *   null args
     *   no viewer in args
     *     no decoration box in args
     *     a decoration box in args
     *   a viewer in args
     *     no decoration box in args
     *     a decoration box in args
     */
    protected ConcreteUnitaryDiagram getCUD(List args) {
    	_viewer = getViewer(args);
     	if (args == null) return null;
    	DecorationBox box = null;
    	for(int i=0; i<args.size(); i++) {
    		Object obj = args.get(i);
    		if (obj instanceof DecorationBox) {
    			box = (DecorationBox)obj;
    			break;
    		}
    	}
   		if (_viewer != null && box != null)
   			return _viewer.getCUDCalled(box.getLabel());
   		else
   			return null;
    }

    /*
     * pre:  args = list of arguments for this rule, or null
     * post: returns set of selected zones from args 
     */
    /*
     * test:
     *   null args
     *   zero length args
     *   no concrete zone in args
     *   two concrete zones in args
     */
    protected Collection getConcreteZones(List args) {
    	Collection czones = new HashSet();
    	if (args != null){
    		for (int i=0; i<args.size(); i++) {
        		Object obj = args.get(i);
        		if (obj instanceof ConcreteZone)
        			czones.add(obj);
    		}
    	}
    	return czones;
    }


    /*
     * pre:  args = list of arguments for this rule, or null
     * post: returns set of selected spiders from args 
     */
    /*
     * test:
     *   null args
     *   zero length args
     *   args with no concrete diagram
     *   args with concrete diagram
     *     args with no spider bits
     *     args with bits of two spiders
     */
    protected Collection getConcreteSpiders(List args) {
    	Collection cspiders = new HashSet();
    	if (args != null){
    		ConcreteUnitaryDiagram cud = getCUD(args);
    		if (cud != null) { 
	    		for (int i=0; i<args.size(); i++) {
	    			Object o = args.get(i);
	    			if (o instanceof DiagramItem) {
	    				DiagramItem di = (DiagramItem) o;
	    				ConcreteSpider cs = getConcreteSpider(cud,di);
	    				if (cs != null)
	    					cspiders.add(cs);
	    			}
	    		}
    		}
    	}
    	return cspiders;
    }


    /**
     * @param cud concrete unitary diagram
     * @param item component of diagram cud
     * @return concrete spider in cud whose foot or leg is item, or null iff none
     */
    /*
     * test:
     *   item is foot of spider in cud
     *   item is leg of spider in cud
     *   item is not a foot or leg
     */
    protected ConcreteSpider getConcreteSpider(ConcreteUnitaryDiagram cud, DiagramItem item) {
    	if (item instanceof ConcreteSpiderFoot) {
    		return getConcreteSpider(cud, (ConcreteSpiderFoot)item);
    	} else if (item instanceof ConcreteSpiderLeg) {
    		ConcreteSpiderLeg leg = (ConcreteSpiderLeg)item;
    		ConcreteSpiderFoot foot1 = leg.getSource();
    		return getConcreteSpider(cud, foot1);
    	} else {
    		return null;
    	}
    }
    
    /**
     * @param cud concrete unitary diagram
     * @param foot foot of spider 
     * @return concrete spider in cud whose foot is foot, 
     *           or null iff none
     */
    /*
     * test:
     *   foot is foot of spider in cud
     */
    protected ConcreteSpider getConcreteSpider(ConcreteUnitaryDiagram cud, ConcreteSpiderFoot foot) {
    	return cud.getConcreteSpider(foot);
    }
    
    /*
     * pre:  args = list of arguments for this rule
     *       i = position in list of required argument
     *       _viewer non-null implies that _viewer is the first viewer in args
     * post: returns required argument, or null if argument does not exist,
     *         or viewer does not exist
     *       _viewer = first viewer in args, or null iff none
     */
    /*
     * test:
     *   _viewer is non-null
     *     i = args.size-1
     *     i = args.size
     *   _viewer is null
     *     there is a viewer in args
     *     there is no viewer in args
     */
    protected DiagramItem getArg(List args, int i) {
    	if (_viewer == null) _viewer = getViewer(args);
    	if (_viewer == null) return null;
    	if (args.size() <= i) return null;
    	DiagramItem item = (DiagramItem)args.get(i);
    	return item;
    }

    protected DiagramViewerInterface _viewer;
    
    /*
    // test program
    // ------------

    private static void pr(String s) {
      System.out.println(s);
    }

    private static void hd(String s) {
      System.out.println("\n" + s + "\n");
    }

    private static void pg() {
      try {
        int i;
        System.out.print("next page? (CR)");
        do i = System.in.read(); while (i != '\n');
        System.out.println();
      }
      catch (Exception e) {System.out.println("raised exception : " + e);}
    }

    public static void main(String args[]) throws Exception {
    	
    	CDRule cdr = new CDRule("test CDRule");
    	pr("cdr="+cdr);
    	pg();
     	/*
     	{
	    	hd("Tests of getViewer");
	    	CDRule cdr = new CDRule("test CDRule");
	    	pr("cdr="+cdr);
	    	pg();
	    	
	    	java.util.ArrayList a = new java.util.ArrayList();
	    	pr("a="+a);
	    	pr("cdr.getViewer(a)="+cdr.getViewer(a));
	    	pg();
	    	
	    	a.add(new Integer(42));
	    	pr("a="+a);
	    	pr("cdr.getViewer(a)="+cdr.getViewer(a));
	    	pg();
	    	
	    	DiagramViewerInterface dvc = 
	    	    new rwd.gui.editor.viewers.DiagramViewerComponent(100,200,false);
	    	a.add(dvc);
	    	pr("a="+a);
	    	pr("cdr.getViewer(a)="+cdr.getViewer(a));
	    	pg();
	    	
	    	DiagramViewerInterface dvc2 = 
	    	    new rwd.gui.editor.viewers.DiagramViewerComponent(300,400,true);
	    	a.add(dvc2);
	    	pr("a="+a);
	    	pr("cdr.getViewer(a)="+cdr.getViewer(a));
	    	pg();
     	}
        
    	{
	    	hd("Tests of getCud");
	    	CDRule cdr = new CDRule("test CDRule");
	    	pr("cdr="+cdr);
	    	pg();
	    	
	    	java.util.ArrayList a = new java.util.ArrayList();
	    	pr("a="+a);
	    	pr("cdr.getCUD(a)="+cdr.getCUD(a));
	    	pr("cdr._viewer="+cdr._viewer);
	    	pg();
	    	
	    	a.add(new Integer(42));
	    	pr("a="+a);
	    	pr("cdr.getCUD(a)="+cdr.getCUD(a));
	    	pr("cdr._viewer="+cdr._viewer);
	    	pg();
	    	
	    	DecorationBox db = new DecorationBox("db","",new java.awt.Polygon());
	    	ConcreteDiagram cd = new ConcreteDiagram();
	    	cd.addDecorationBox(db);
	    	DiagramViewerInterface dvc = 
	    	    new rwd.gui.editor.viewers.DiagramViewerComponent(cd,300,400,true);
	    	
	    	a.add(db);
	    	pr("a="+a);
	    	pr("cdr.getCUD(a)="+cdr.getCUD(a));
	    	pr("cdr._viewer="+cdr._viewer);
	    	pg();
	    	
	    	a.add(dvc);
	    	pr("a="+a);
	    	pr("cdr.getCUD(a)="+cdr.getCUD(a));
	    	pr("cdr._viewer="+cdr._viewer);
	    	pg();
	    	
	    	a.remove(1);  // remove the box
	    	pr("a="+a);
	    	pr("cdr.getCUD(a)="+cdr.getCUD(a));
	    	pr("cdr._viewer="+cdr._viewer);
	    	pg();
	    	
	    	a.clear();
	    	pr("a="+a);
	    	pr("cdr.getCUD(a)="+cdr.getCUD(a));
	    	pr("cdr._viewer="+cdr._viewer);
	    	pg();
	    	
     	}
       
    	{
    		hd("Tests of getZones");
    		pr("cdr.getZones(null)="+
    		    cdr.getZones(null));
    		pg();
    		
    		List targs = new java.util.ArrayList();
    		pr("targs="+targs);
    		pr("cdr.getZones(targs)="+
        		cdr.getZones(targs));
        	pg();
        		
    		
    		Object o1 = new Object();
    		Object o2 = new Object();
    		targs.add(o1);
    		targs.add(o2);
    		pr("targs="+targs);
    		pr("cdr.getZones(targs)="+
        		cdr.getZones(targs));
        	pg();
        		
    		targs = new java.util.ArrayList();
    		java.util.Vector v = new java.util.Vector();
    		ConcreteZone cz1 = new ConcreteZone(v,v);
    		pr("cz1="+cz1);
    		ConcreteZone cz2 = new ConcreteZone(v,v);
    		cz2.setBackColor(java.awt.Color.GRAY);
    		pr("cz2="+cz2);
    		targs.add(cz2);
    		targs.add(o1);
    		targs.add(o2);
    		targs.add(cz1);
    		pr("targs="+targs);
    		pr("cdr.getZones(targs)="+
        		cdr.getZones(targs));
        	pg();
    		
    	}
    	
    	{
    		hd("Tests of getConcreteSpiders");
    		pr("cdr.getConcreteSpiders(null)="+
    		    cdr.getConcreteSpiders(null));
    		pg();
    		
    		List targs = new java.util.ArrayList();
    		pr("targs="+targs);
    		pr("cdr.getConcreteSpiders(targs)="+
        		cdr.getConcreteSpiders(targs));
        	pg();
        	
        	Object o = new Object();
        	targs.add(o);
    		pr("targs="+targs);
    		pr("cdr.getConcreteSpiders(targs)="+
        		cdr.getConcreteSpiders(targs));
        	pg();
        	        	
    	}
        		

    }
    */
}
