package rwd.tableaux.diagram;

import java.util.*;

import rwd.concretesyntax.*;
import rwd.abstractsyntax.spiderdiagram.*;
import rwd.tableaux.semantic.Options;
import rwd.tableaux.syntax.ast.*;

import uk.ac.kent.cs.kmf.util.*;

/**
 * @author Octavian Patrascoiu
 *
 */
public class Expression2DiagramImpl 
	implements Expression2Diagram
{
	/** Constructors */
	public Expression2DiagramImpl(ILog log) {
		this.log = log;
	}
	
	/** Conversion methods */
	public Diagram convert(LogicalExp exp, List diags, Options opt) throws Exception {
		// Convert expression to diagram
		return logicExp2Diagram(exp, diags, opt);
	}
	
	//
	// Transform a logical expression into a diagram
	//
	/** Convert a logical expression into a diagram */
	private Diagram logicExp2Diagram(LogicalExp exp, List diags, Options opt) throws Exception {
		// P and Q
		Diagram result = null;
		if (exp instanceof AndExp) {
			// Make P and Q
			Diagram p = logicExp2Diagram(exp.getLeft(), diags, opt);
			Diagram q = logicExp2Diagram(exp.getRight(), diags, opt); 
			// Make an AND diagram
			result = makeAnd(p, q, opt);
		// P or Q	
		} else if (exp instanceof OrExp) {
			// Make P and Q
			Diagram p = logicExp2Diagram(exp.getLeft(), diags, opt);
			Diagram q = logicExp2Diagram(exp.getRight(), diags, opt); 
			// Make an OR diagram
			result = makeOr(p, q, opt);
		// P implies Q
		} else if (exp instanceof ImpliesExp) {
			// Make P and not P
			Diagram p = logicExp2Diagram(exp.getLeft(), diags, opt);
			Diagram notP = makeNot(p);
			// Make Q
			Diagram q = logicExp2Diagram(exp.getRight(), diags, opt); 
			// Make an OR diagram
			result = makeOr(notP, q, opt);
		// P equivalent Q 	
		} else if (exp instanceof EquivalentExp) {
			// Make P and not P
			Diagram p = logicExp2Diagram(exp.getLeft(), diags, opt);
			Diagram notP = makeNot(p);
			// Make Q and not Q
			Diagram q = logicExp2Diagram(exp.getLeft(), diags, opt);
			Diagram notQ = makeNot(q);
			// Make P => Q
			Diagram pImpliesQ = makeOr(notP, q, opt);
			// Make Q => P
			Diagram qImpliesP = makeOr(notQ, p, opt);
			// Make P <=> Q
			result = makeAnd(pImpliesQ, qImpliesP, opt);
		// not P
		} else if (exp instanceof NotExp) {
			// Make P
			Diagram p = logicExp2Diagram(exp.getLeft(), diags, opt);
			result = makeNot(p);				
		// atomic
		} else if (exp instanceof AtomicExp) {
			// Get diagram name
			String name = ((AtomicExp)exp).getName();
			// false
			if (name.equals("false")) {
				result = new FalseDiagram("");
			// true
			} else if (name.equals("true")) {
				result = new TrueDiagram("");
			// user
			} else {
				// Search the unitary concrete diagram
				ConcreteUnitaryDiagram cud = getUnitaryDiagram(diags, name);
				if (cud != null) {
					// Make the abstract unitary diagram
					UnitaryDiagram ud = cud.getAbstractUnitaryDiagram();
					result = ud;							
				} else {
					String message = "Unknown diagram '"+name+"'"; 
					throw new Exception(message);
				}
			}
		}
		return result;
	}
	/** Get the concrete unitary diagram associated with a name */ 
	private ConcreteUnitaryDiagram getUnitaryDiagram(List diags, String name) {
		for(int i=0; i<diags.size(); i++) {
			ConcreteDiagram cd = (ConcreteDiagram)diags.get(i);
			Set cuds = cd.getConcreteUnitaryDiagrams();
			Iterator j = cuds.iterator();
			while (j.hasNext()) {
				ConcreteUnitaryDiagram cud = (ConcreteUnitaryDiagram)j.next();
				if (cud.getName().equals(name)) return cud;
			}
		}
		return null;
	}
	/** Not, and, and or building methods */
	private Diagram makeNot(Diagram d) {
		Vector diags = new Vector();
		diags.add(d);
		Diagram result = new NotCompoundDiagram(diags, "");
		return result;
	}
	private Diagram makeAnd(Diagram d1, Diagram d2, Options opt) {
		List diags = new Vector();
		diags.add(d1);
		diags.add(d2);
		Diagram result = new AndCompoundDiagram(new Vector(), "");
		((CompoundDiagram)result).getChildren().addAll(diags);
		if (!opt.isBinaryOperators())
			result.flatten();
		return result;
	}
	private Diagram makeOr(Diagram d1, Diagram d2, Options opt) {
		List diags = new Vector();
		diags.add(d1);
		diags.add(d2);
		Diagram result = new OrCompoundDiagram(new Vector(), "");
		((CompoundDiagram)result).getChildren().addAll(diags);
		if (!opt.isBinaryOperators())
			result.flatten();
		return result;
	}
			
	/** Log file */
	protected ILog log;
}
