package rwd.tableaux.diagram;

import java.util.Vector;
import java.util.List;

import rwd.abstractsyntax.eulerdiagram.Zone;
import rwd.abstractsyntax.spiderdiagram.AndCompoundDiagram;
import rwd.abstractsyntax.spiderdiagram.CompoundDiagram;
import rwd.abstractsyntax.spiderdiagram.Diagram;
import rwd.abstractsyntax.spiderdiagram.FalseDiagram;
import rwd.abstractsyntax.spiderdiagram.NotCompoundDiagram;
import rwd.abstractsyntax.spiderdiagram.OrCompoundDiagram;
import rwd.abstractsyntax.spiderdiagram.Spider;
import rwd.abstractsyntax.spiderdiagram.TrueDiagram;
import rwd.abstractsyntax.spiderdiagram.UnitaryDiagram;
import rwd.tableaux.semantic.Options;
import uk.ac.kent.cs.kmf.util.ILog;

/**
 * @author Octavian Patrascoiu
 *
 */
public class NormaliserImpl implements Normaliser {
	/** Constructors */
	public NormaliserImpl(Options opt, ILog log) {
		this.opt = opt;
		this.log = log;
	}

	/** Convert a diagram to Disjunctive Normal Form */
	public Diagram toDNF(Diagram d) {
		if (opt.isDebug()) {
			log.reportMessage("Remove double negation ...");
		}
		d = removeDoubleNegation(d);
		if (opt.isDebug()) {
			log.reportMessage("Apply deMorgan ...");
		}
		d = applyDeMorgan(d);
		if (opt.isDebug()) {
			log.reportMessage("Distribute 'and' over 'or' ...");
		}
		d = distributeAndOverOr(d);
		return d;
	}

	/** Convert a diagram to Disjunctive Normal Form without Not*/
	public Diagram toDNFWithoutNegation(Diagram d) {
		d = removeNegation(d);
		d = distributeAndOverOr(d);
		return d;
	}

	/** Convert a diagram to Disjunctive Normal Form */
	public Diagram removeNegation(Diagram d) {
		if (opt.isDebug()) {
			log.reportMessage("Remove double negation ...");
		}
		d = removeDoubleNegation(d);
		if (opt.isDebug()) {
			log.reportMessage("Apply deMorgan ...");
		}
		d = applyDeMorgan(d);
		if (opt.isDebug()) {
			log.reportMessage("Negate unitary diagrams ...");
		}
		d = negateUnitaryDiagrams(d);
		return d;
	}

	/** Remove double negation: not not X -> X */
	public Diagram removeDoubleNegation(Diagram d) {
		Diagram newDiagram = scanDiagram4DubleNegation(d);
		return newDiagram;
	}
		
	/** Apply deMorgan rules */
	public Diagram applyDeMorgan(Diagram d) {
		apply = true;
		Diagram newDiagram = d;
		while (apply) {
			apply = false;
			newDiagram = scanDiagram4Morgan(newDiagram);
		}
		return newDiagram;
	}

	/** Replace not X with NOT_X */
	public Diagram negateUnitaryDiagrams(Diagram d) {
		Diagram newDiagram = scanDiagram2Negate(d);
		return newDiagram;
	}
	
	/** Distribute AND over OR */
	public Diagram distributeAndOverOr(Diagram d) {
		if (opt.isDebug()) {
			log.reportMessage("Distribute and over or ...");
		}
		apply = true;
		Diagram newDiagram = d;
		int i = 0;
		while (apply) {
			apply = false;
			i++;
			Diagram temp = scanDiagram4Distribution(newDiagram); 
			newDiagram = temp;
		}
		return newDiagram;
	}
		
	/** not not X -> X */
	public Diagram scanDiagram4DubleNegation(Diagram d) {
		// Null expression
		if (d == null) 
			return d;
		// atomic
		if (d instanceof UnitaryDiagram) {
			return d;
		// Not, And, Or, Inmplies, Equivalent
		} else if (d instanceof CompoundDiagram) {
			// not+ P
			if (d instanceof NotCompoundDiagram) {
				Diagram p = d;
				// Count the number of nots
				boolean hasNot = false;
				while (p instanceof NotCompoundDiagram) {
					hasNot = !hasNot;
					p = (Diagram)((CompoundDiagram)p).getChildren().get(0);
				}
				// Process p
				Diagram newDiagram = scanDiagram4DubleNegation(p);
				if (hasNot) {
					newDiagram = makeNotDiagram(newDiagram);				
				}
				return newDiagram;
			// And, Or
			} else {
				CompoundDiagram cd = (CompoundDiagram)d;
				List children = cd.getChildren();			
				// Make new children
				Vector newChildren = new Vector();
				for(int i = 0; i<children.size(); i++) {
					Diagram child = (Diagram)children.get(i);
					newChildren.add(scanDiagram4DubleNegation(child));
				}
				// Set children
				cd.setChildren(newChildren);
				if (!opt.isBinaryOperators())
					cd.flatten();
				return cd;
			}	
		// Else
		} else {
			return d;				
		}
	}

	/** Apply deMorgan rules */
	private Diagram scanDiagram4Morgan(Diagram d) {
		// Null expression
		if (d == null) 
			return d;
		// atomic
		if (d instanceof UnitaryDiagram) {
			return d;
		// Not, And, Or, Inmplies, Equivalent
		} else if (d instanceof CompoundDiagram) {
			CompoundDiagram cd = (CompoundDiagram)d;
			Vector children = cd.getChildren();
			// Not diagram
			if (d instanceof NotCompoundDiagram) {
				// Get child
				Diagram child = (Diagram)cd.getChildren().get(0);
				// not(P1 and P2 and ... and Pn) -> not P1 or not P2 or ... or not Pn  
				if (child instanceof AndCompoundDiagram) {
					// Set flag
					apply = true;
					List andChildren = ((CompoundDiagram)child).getChildren();			
					// Make new children
					Vector newChildren = new Vector();
					for(int i = 0; i<andChildren.size(); i++) {
						Diagram child1 = (Diagram)andChildren.get(i);
						newChildren.add(makeNotDiagram(child1));
					}
					// Make an OR diagram
					Diagram newDiagram = new OrCompoundDiagram(newChildren, d.getName());
					if (!opt.isBinaryOperators())
						newDiagram.flatten();
					return newDiagram;
				// not(P1 or P2 or ... or Pn) -> not P1 and not P2 and ... and not Pn  
				} else if (child instanceof OrCompoundDiagram) {
					// Set flag
					apply = true;
					List orChildren = ((CompoundDiagram)child).getChildren();			
					// Make new children
					Vector newChildren = new Vector();
					for(int i = 0; i<orChildren.size(); i++) {
						Diagram child1 = (Diagram)orChildren.get(i);
						newChildren.add(makeNotDiagram(child1));
					}
					// Make an AND diagram
					Diagram newDiagram = new AndCompoundDiagram(newChildren, d.getName());
					if (!opt.isBinaryOperators())
						newDiagram.flatten();
					return newDiagram;
				// Else
				} else {
					return d;
				}
			// AND, OR
			} else {
				// Make new children
				Vector newChildren = new Vector();
				for(int i = 0; i<children.size(); i++) {
					Diagram child1 = (Diagram)children.get(i);
					newChildren.add(scanDiagram4Morgan(child1));
				}
				// Update the diagram
				cd.setChildren(newChildren);
				if (!opt.isBinaryOperators())
					cd.flatten();
				return cd;
			}
		// Else
		} else {
			return d;				
		}
	}

	/** Replace not X with NOT_X */
	public Diagram scanDiagram2Negate(Diagram d) {
		// Null expression
		if (d == null) 
			return d;
		// atomic
		if (d instanceof UnitaryDiagram) {
			return d;
		// Not, And, Or, Inmplies, Equivalent
		} else if (d instanceof CompoundDiagram) {
			// not P where P is unitary diagram
			if (d instanceof NotCompoundDiagram) {
				UnitaryDiagram child = (UnitaryDiagram)((NotCompoundDiagram)d).getChildren().get(0);
				return makeNegationOf(child);
			// And, Or
			} else {
				CompoundDiagram cd = (CompoundDiagram)d;
				List children = cd.getChildren();			
				// Make new children
				Vector newChildren = new Vector();
				for(int i = 0; i<children.size(); i++) {
					Diagram child = (Diagram)children.get(i);
					newChildren.add(scanDiagram2Negate(child));
				}
				// Set children
				cd.setChildren(newChildren);
				if (!opt.isBinaryOperators())
					cd.flatten();
				return cd;
			}	
		// Else
		} else {
			return d;				
		}
	}
		
	/** Apply deMorgan rules */
	private Diagram scanDiagram4Distribution(Diagram d) {
		// Null expression
		if (d == null) 
			return d;
		// atomic
		if (d instanceof UnitaryDiagram) {
			return d;
		// Not, And, Or, Inmplies, Equivalent
		} else if (d instanceof CompoundDiagram) {
			CompoundDiagram cd = (CompoundDiagram)d;
			List children = cd.getChildren();
			// Not diagram
			if (d instanceof NotCompoundDiagram) {
				return d;
			// And 
			} else if (d instanceof AndCompoundDiagram) {
				// Get children
				AndCompoundDiagram and = (AndCompoundDiagram)d;
				List andChildren = and.getChildren();
				// Find an OR
				OrCompoundDiagram or = null;
				int i = 0;
				for(; i<andChildren.size(); i++) {
					Diagram child = (Diagram)andChildren.get(i);
					if (child instanceof OrCompoundDiagram) {
						or = (OrCompoundDiagram)child;
						break;
					}
				}
				// Found the pattern
				if (or != null) {
					// Set the flag which shows that a pattern was found
					apply = true;
					// Pi = Q1 or ... or Qm
					// P1 and ... and Pi-1 and (Q1 or Q2 or ... or Qm) and Pi+1 and ... and Pn
					// goes to
					// P1 and ... and Pi-1 and Q1 and Pi+1 and ... and Pn or
					// P1 and ... and Pi-1 and Q2 and Pi+1 and ... and Pn or
					// ....
					// P1 and ... and Pi-1 and Qm and Pi+1 and ... and Pn
					// Compute children for new OR
					Vector newOrChildren = new Vector();
					List orChildren = or.getChildren();
					for(int j=0; j<orChildren.size(); j++) {
						// Compute children for new AND
						Vector newAndChildren = new Vector();
						// Add P1, ... Pi-1
						for(int k=0; k<i; k++) {
							Diagram temp = (Diagram) andChildren.get(k);
							newAndChildren.add(temp.getClone());
						}
						// Add Qj
						newAndChildren.add(orChildren.get(j));
						// Add Pi+1, ..., Pn
						for(int k=i+1; k<andChildren.size(); k++) {
							Diagram temp = (Diagram) andChildren.get(k);
							newAndChildren.add(temp.getClone());
						}
						// Make a new AND diagram
						AndCompoundDiagram newAnd = new AndCompoundDiagram(new Vector(), "");
						newAnd.setChildren(newAndChildren);
						if (!opt.isBinaryOperators())
							newAnd.flatten();
						// Add to new children
						newOrChildren.add(newAnd);
					}
					// Make a new OR diagram
					OrCompoundDiagram newOr = new OrCompoundDiagram(new Vector(), "");
					newOr.setChildren(newOrChildren);
					if (!opt.isBinaryOperators())
						newOr.flatten();
					return newOr;
				// Pattern not found
				} else {
					// Make new children
					Vector newChildren = new Vector();
					for(i = 0; i<children.size(); i++) {
						Diagram child = (Diagram)children.get(i);
						newChildren.add(scanDiagram4Distribution(child));
					}
					// Create a new diagram
					AndCompoundDiagram newAnd = new AndCompoundDiagram(new Vector(), "");
					newAnd.setChildren(newChildren);
					if (!opt.isBinaryOperators())
						newAnd.flatten();
					return newAnd;
				}
			// OR
			} else {
				// Make new children
				Vector newChildren = new Vector();
				for(int i = 0; i<children.size(); i++) {
					Diagram child = (Diagram)children.get(i);
					newChildren.add(scanDiagram4Distribution(child));
				}
				// Create a new diagram
				OrCompoundDiagram newOr = new OrCompoundDiagram(new Vector(), "");
				newOr.setChildren(newChildren);
				if (!opt.isBinaryOperators())
					newOr.flatten();
				return newOr;
			}
		// Else
		} else {
			return d;				
		}
	}
	
	/** Make the negation of a unitary diagram */
	public Diagram makeNegationOf(UnitaryDiagram ud) {
		if (ud == null) {
			return null;
		} else if (ud instanceof FalseDiagram) {
			return new FalseDiagram("");
		} else if (ud instanceof TrueDiagram) {
			return new TrueDiagram("");
		} else {
			// Result is an OR or a unitary diagram
			List newChildren = new Vector();
			// For each zone
			List vennZones = UnitaryDiagram.allPossibleZones(ud.getCloneOfContourSet());
			List zones = ud.getCloneOfZoneSet();
			int noDiag = 1;
			// For each zone except outer zone
			for(int i=0; i<vennZones.size(); i++) {
				Zone z = (Zone)vennZones.get(i);
				if (z.getContainingContours().size() == 0) continue;
				// Make a single-footed spider touching z
				Vector habitat = new Vector();
				habitat.add(z);
				// Missing zone
				if (!zones.contains(z)) {
					// Create a diagram with one single-footed spider
					UnitaryDiagram newDiagram = new UnitaryDiagram((Vector)vennZones, ud.getName()+"-"+noDiag++);
					newDiagram.addSpider(habitat);
					// Add diagram
					newChildren.add(newDiagram);
				// z is in ud
				} else {
					// Compute isShaded flag and number of feet
					boolean isShaded = z.isShaded(ud); 
					int noFeet = computeNoFeet(z, ud);
					// Shaded zones
					if (isShaded) {
						// Zone without restriction
						if (noFeet == 0) {
							// Create a diagram with one spider touching unshaded zone z
							UnitaryDiagram newDiagram = new UnitaryDiagram((Vector)vennZones, ud.getName()+"-"+noDiag++);
							// Add only spider touching the zone z
							newDiagram.addSpider(habitat);
							// Add diagram
							newChildren.add(newDiagram);
						// Zone contains at most noFeet elements
						} else {
							// Add noFeet diagrams
							for(int j=0; j<noFeet; j++) {
								// Create a diagram with j spiders touching shaded zone z
								UnitaryDiagram newDiagram = new UnitaryDiagram((Vector)vennZones, ud.getName()+"-"+noDiag++);
								// Add j spiders touching zone z
								for(int k=0; k<j; k++)
									newDiagram.addSpider(habitat);
								// Shade the zone
								newDiagram.shadeZone(z);
								// Add diagram
								newChildren.add(newDiagram);			
							}
							// Create a diagram with n+1 spiders touching unshaded zone z
							UnitaryDiagram newDiagram = new UnitaryDiagram((Vector)vennZones, ud.getName()+"-"+noDiag++);
							// Add n+1 spiders touching zone z
							for(int k=0; k<noFeet+1; k++) {
								newDiagram.addSpider(habitat);
							}
							// Add diagram
							newChildren.add(newDiagram);						
						}
						// Not shaded zones
					} else {
						// Empty zone
						if (noFeet == 0) {
							// ignore it
						// Contains exactly noFeet elements					
						} else {
							// Add noFeet diagrams 
							for(int j=0; j<noFeet; j++) {
								// Create a diagram with j spiders touching shaded zone z
								UnitaryDiagram newDiagram = new UnitaryDiagram((Vector)vennZones, ud.getName()+"-"+noDiag++);
								// Add j spiders touching zone z
								for(int k=0; k<j; k++)
									newDiagram.addSpider(habitat);
								// Shade the zone
								newDiagram.shadeZone(z);
								// Add diagram
								newChildren.add(newDiagram);			
							}
						}						
					}
				}
			}
			
			// Make the result
			int size = newChildren.size(); 
			if (size == 0) {
				return null;
			} else if (size == 1) {
				return (Diagram) newChildren.get(0);
			} else {
				OrCompoundDiagram or = new OrCompoundDiagram(new Vector(), "");
				or.setChildren((Vector)newChildren);
				if (!opt.isBinaryOperators())
					or.flatten();
				return or;
			}
		}
	}
	
	/** Compute the number of spiders touching a zone */
	protected int computeNoFeet(Zone z, UnitaryDiagram ud) {
		int noFeet = 0;
		Vector spiders = ud.getSpiders();
		for (int i = 0; i < spiders.size(); i++) {
			Spider s = (Spider) spiders.get(i);
			if (s.getHabitat().contains(z))
				noFeet++;
		}
		return noFeet;
	}
	 
	/** Not, and, and or building methods */
	private Diagram makeNotDiagram(Diagram d) {
		Vector diags = new Vector();
		diags.add(d);
		Diagram result = new NotCompoundDiagram(diags, "");
		return result;
	}

	//
	// Local attributes
	//
	protected Options opt;
	protected ILog log;
	private boolean apply;
}
