package rwd.tableaux.semantic;

import java.util.*;

import rwd.abstractsyntax.eulerdiagram.Zone;
import rwd.abstractsyntax.spiderdiagram.*;
import rwd.tableaux.diagram.Checker;
import rwd.tableaux.diagram.CheckerImpl;
import rwd.tableaux.diagram.Equaliser;
import rwd.tableaux.diagram.EqualiserImpl;

import uk.ac.kent.cs.kmf.util.*;

/**
 * @author Octavian Patrascoiu
 *
 */
public class ENodeImpl
	implements ENode 
{
	/** Constructor */
	public ENodeImpl(ETableauImpl tableau, List diagrams) {
		this.tableau = tableau;
		this.diagrams = diagrams;
		this.options = tableau.opts;
		this.log = tableau.log;
		this.kind = OperationKind.UNDEFINED;
	}
	
	/** 
	 * Check if the node can be checked for contradictions:
	 * - all diagrams are literals
	 * - there is no not
	 * - all the diagrams are equalised
	 * - all the spiders have one foot
	 */
	public boolean isLeaf() {
		// Only one diagram
		if (diagrams.size() == 1) {
			Diagram d = (Diagram)diagrams.get(0);
			if (isLiteral(d))
				return true;
			else
				return false;
		// Several diagrams
		} else {
			if (!allDiagramsAreLiterals(diagrams))
				return false;
			if (containsNot())
				return false;
			if (!allDiagramsAreEqualised())
				return false;
			if (!allSpidersHaveOneFoot())
				return false;
			return true;
		}
	}

	/** Check if all the diagrams are literals: X and not X */
	public boolean allDiagramsAreLiterals() {
		return allDiagramsAreLiterals(diagrams);
	}
	public boolean allDiagramsAreLiterals(List diagrams) {
		for(int i=0; i<diagrams.size(); i++) {
			Diagram d = (Diagram)diagrams.get(i);
			if (!isLiteral(d)) 
				return false;
		}
		return true;		
	}

	/** Check if there is at least Not diagram (skip not true, not false) */
	public boolean containsNot() {
		return containsNot(diagrams);
	}
	protected boolean containsNot(List diagrams) {
		for(int i=0; i<diagrams.size(); i++) {
			Diagram d = (Diagram)diagrams.get(i);
			if (d instanceof NotCompoundDiagram) {
				Diagram child = (Diagram)((CompoundDiagram)d).getChildren().get(0);
				if (child instanceof TrueDiagram) {
				} else if (child instanceof TrueDiagram) {
				} else {
					return true;
				} 
			} else if (d instanceof CompoundDiagram) {
				List children = ((CompoundDiagram)d).getChildren();
				boolean flag = containsNot(children);
				if (flag)
					return true; 
			}
		}
		return false;				
	}

	/** Check if all the unitary diagrams are equalised */
	public boolean allDiagramsAreEqualised() {
		return allDiagramsAreEqualised(diagrams);
	}
	public boolean allDiagramsAreEqualised(List diagrams) {
		// Compute all the contours and the zones
		Equaliser eq = new EqualiserImpl(options, log);
		Set contours = new TreeSet();
		Set zones = new TreeSet();
		for(int i=0; i<diagrams.size(); i++) {
			Diagram d = (Diagram)diagrams.get(i);
			eq.computeContours(d, contours);
			eq.computeZones(d, zones);
		}
		// Check every unitary diagram
		return areEqualised(diagrams, contours, zones);
	}
	protected boolean areEqualised(List diagrams, Set contours, Set zones) {
		for(int i=0; i<diagrams.size(); i++) {
			Diagram d = (Diagram)diagrams.get(i);
			if (d instanceof TrueDiagram) {
			} else if (d instanceof FalseDiagram) {
			} else if (d instanceof UnitaryDiagram) {
				UnitaryDiagram ud = (UnitaryDiagram)d;
				List cs = ud.getCloneOfContourSet();
				List zs = ud.getCloneOfZoneSet();
				if (cs.containsAll(contours) && contours.containsAll(cs)) {
				} else {
					return false;
				}
				if (zs.containsAll(zones) && zones.containsAll(zs)) {
				} else {
					return false;
				}
			} else if (d instanceof CompoundDiagram) {
				List children = ((CompoundDiagram)d).getChildren();
				boolean flag = areEqualised(children, contours, zones);
				if (!flag)
					return false; 
			}
		}
		return true;						
	}

	/** Check if all the spiders have one foot */
	public boolean allSpidersHaveOneFoot() {
		return allSpidersHaveOneFoot(diagrams);
	}
	public boolean allSpidersHaveOneFoot(List diagrams) {
		for(int i=0; i<diagrams.size(); i++) {
			Diagram d = (Diagram)diagrams.get(i);
			if (d instanceof TrueDiagram) {
			} else if (d instanceof FalseDiagram) {
			} else if (d instanceof UnitaryDiagram) {
				UnitaryDiagram ud = (UnitaryDiagram)d;
				List spiders = ud.getSpiders();
				for(int j=0; j<spiders.size(); j++) {
					Spider sp = (Spider)spiders.get(j);
					if (sp.getHabitat().size() == 1) {
					} else {
						return false;
					}
				}
			} else if (d instanceof CompoundDiagram) {
				List children = ((CompoundDiagram)d).getChildren();
				boolean flag = allSpidersHaveOneFoot(children);
				if (!flag)
					return false; 
			}
		}
		return true;						
	}
	
	/** Check if all the unitary diagrams are contour-equalised */
	public boolean allContoursAreEqualised() {
		// Compute all the contours and the zones
		Equaliser eq = new EqualiserImpl(options, log);
		Set contours = new TreeSet();
		for(int i=0; i<diagrams.size(); i++) {
			Diagram d = (Diagram)diagrams.get(i);
			eq.computeContours(d, contours);
		}		
		return allDiagramsContainContours(new Vector(contours));
	}

	/** Check if all the unitary diagrams contain the contours */
	public boolean allDiagramsContainContours(List contours) {
		return allDiagramsContainContours(diagrams, contours);
	}
	public boolean allDiagramsContainContours(List diagrams, List contours) {
		for(int i=0; i<diagrams.size(); i++) {
			Diagram d = (Diagram)diagrams.get(i);
			if (d instanceof TrueDiagram) {
			} else if (d instanceof FalseDiagram) {
			} else if (d instanceof UnitaryDiagram) {
				UnitaryDiagram ud = (UnitaryDiagram)d;
				List cs = ud.getCloneOfContourSet();
				List zs = ud.getCloneOfZoneSet();
				if (cs.containsAll(contours) && contours.containsAll(cs)) {
				} else {
					return false;
				}
			} else if (d instanceof CompoundDiagram) {
				List children = ((CompoundDiagram)d).getChildren();
				boolean flag = allDiagramsContainContours(children, contours);
				if (!flag)
					return false; 
			}
		}
		return true;						
	}

	/** Check if all the unitary diagrams are contour-equalised */
	public boolean allZonesAreEqualised() {
		// Compute all the contours and the zones
		Equaliser eq = new EqualiserImpl(options, log);
		Set zones = new TreeSet();
		for(int i=0; i<diagrams.size(); i++) {
			Diagram d = (Diagram)diagrams.get(i);
			eq.computeZones(d, zones);
		}		
		return allDiagramsContainZones(new Vector(zones));
	}

	/** Check if all the unitary diagrams contain a zone */
	public boolean allDiagramsContainZone(Zone z) {
		List zones = new Vector();
		zones.add(z);
		return allDiagramsContainZones(zones);
	}
	public boolean allDiagramsContainZones(List zones) {
		return allDiagramsContainZones(zones);
	}
	public boolean allDiagramsContainZones(List diagrams, List zones) {
		for(int i=0; i<diagrams.size(); i++) {
			Diagram d = (Diagram)diagrams.get(i);
			if (d instanceof TrueDiagram) {
			} else if (d instanceof FalseDiagram) {
			} else if (d instanceof UnitaryDiagram) {
				UnitaryDiagram ud = (UnitaryDiagram)d;
				List zs = ud.getCloneOfZoneSet();
				if (zs.containsAll(zones) && zones.containsAll(zs)) {
				} else {
					return false;
				}
			} else if (d instanceof CompoundDiagram) {
				List children = ((CompoundDiagram)d).getChildren();
				boolean flag = allDiagramsContainZones(children, zones);
				if (!flag)
					return false; 
			}
		}
		return true;						
	}

	/** Search for a compound diagram (AND and OR) and create a list of nodes */
	public List expand() {
		// Compute result
		List res = new Vector();
		// Choose a binary expression
		int pos = choose();
		if (pos == -1) return res;
		Diagram d = (Diagram)diagrams.get(pos);
		// Set operation kind
		setOperationKind(OperationKind.UNDEFINED);
		// and(P1, P2, ..., Pn) goes to P1, P2, ..., Pn
		if (d instanceof AndCompoundDiagram) {
			// Set operation kind diagram
			setOperationKind(OperationKind.EXPAND_AND);
			// Compute the new diagrams			
			CompoundDiagram cd = (CompoundDiagram)d;
			List newDiagrams = new Vector();
			// Add all previous diagrams
			addDiagrams(newDiagrams, diagrams, 0, pos);
			// Add diagrams for the children
			List children = cd.getChildren();
			for(int i=0; i<children.size(); i++) {
				Diagram child = (Diagram)cd.getChildren().get(i);
				newDiagrams.add(child);
			}
			// Add all succesive expressions
			addDiagrams(newDiagrams, diagrams, pos+1, diagrams.size());
			// Make one node
			ENodeImpl n1 = new ENodeImpl(tableau, newDiagrams);
			// Add node to result
			res.add(n1);
		// or(P1, P2, ..., Pn) goes to P1 | P2 | ... |Pn
		} else if (d instanceof OrCompoundDiagram) {
			// Set operation kind diagram
			setOperationKind(OperationKind.EXPAND_OR);
			// For each children
			CompoundDiagram cd = (CompoundDiagram)d;
			List children = cd.getChildren();
			for(int i=0; i<children.size(); i++) {
				Diagram child = (Diagram)children.get(i);
				// Compute the new diagrams
				List newDiagrams = new Vector();
				// Add previous diagrams
				addDiagrams(newDiagrams, diagrams, 0, pos);
				// Add d's children
				newDiagrams.add(child);
				// Add all succesive expressions
				addDiagrams(newDiagrams, diagrams, pos+1, diagrams.size());
				// Make a tree node
				ENodeImpl childNode = new ENodeImpl(tableau, newDiagrams);
				// Add node to result
				res.add(childNode);
			}	
		// not	
		} else if (d instanceof NotCompoundDiagram) {
			// Set operation kind diagram
			setOperationKind(OperationKind.EXPAND_OR);
			// Get the child
			CompoundDiagram cd = (CompoundDiagram)d;
			Diagram temp = (Diagram)cd.getChildren().get(0);
			// not not P goes to P
			if (temp instanceof NotCompoundDiagram) {
				Diagram temp1 = (Diagram)((CompoundDiagram)temp).getChildren().get(0);
				// Compute the new diagrams
				List newDiagrams = new Vector();
				// Add all previous diagrams
				addDiagrams(newDiagrams, diagrams, 0, pos);
				// Add left diagram
				newDiagrams.add(temp1);
				// Add all succesive diagrams 
				addDiagrams(newDiagrams, diagrams, pos+1, diagrams.size());
				// Make one node
				ENodeImpl n1 = new ENodeImpl(tableau, newDiagrams);
				// Add node to result
				res.add(n1);
			// not and(P1, P2, ... Pn) goes to not P1 | not P2 | ... | Pn
			} else if (temp instanceof AndCompoundDiagram) {
				CompoundDiagram and = (CompoundDiagram)temp;
				// For each children
				List children = and.getChildren();
				for(int i=0; i<children.size(); i++) {
					// Compute the new diagrams
					List newDiagrams = new Vector();
					// Add previous diagrams
					addDiagrams(newDiagrams, diagrams, 0, pos);
					// Add not d's children
					Diagram child = (Diagram)children.get(i);
					newDiagrams.add(makeNot(child));
					// Add all succesive expressions
					addDiagrams(newDiagrams, diagrams, pos+1, diagrams.size());
					// Make a tree node
					ENodeImpl childNode = new ENodeImpl(tableau, newDiagrams);
					// Add node to result
					res.add(childNode);
				}	
			// not (P1 or P2 or ... Pn) goes to not P1, not P2, ..., not Pn
			} else if (temp instanceof OrCompoundDiagram) {
				CompoundDiagram or = (CompoundDiagram)temp;
				// Compute the new diagrams
				List newDiagrams = new Vector();
				// Add all previous diagrams
				addDiagrams(newDiagrams, diagrams, 0, pos);
				// Add not and's children
				List children = or.getChildren();
				for(int i=0; i<children.size(); i++) {
					Diagram child = (Diagram)children.get(i);
					newDiagrams.add(makeNot(child));
				}
				// Add all succesive expressions
				addDiagrams(newDiagrams, diagrams, pos+1, diagrams.size());
				// Make one node
				ENodeImpl n1 = new ENodeImpl(tableau, newDiagrams);
				// Add node to result
				res.add(n1);
			}
		}
		return res;
	}

	/** Compute contradiction */
	public Contradiction computeContradiction() {
		Contradiction con = new ContradictionImpl();
		// Check only leaves (unitary diagrams and not unitary diagrams
		if (!isLeaf()) 
			return con;
		// Create a checker
		Checker chk = new CheckerImpl(options, log);
		// Check for single diagrams: true, false, not true, not false, X, not X
		if (diagrams.size() == 1) {
			Diagram d = (Diagram)diagrams.get(0);
			// Compute contradiction for d
			con = chk.computeContradiction(d);			
		} else {
			// Check every pair of literals
			con.setContradiction(false);
			for(int i=0; i<diagrams.size(); i++) {
				Diagram d1 = (Diagram)diagrams.get(i);
				if (isLiteral(d1)) {
					for(int j=i+1; j<diagrams.size(); j++) {
						Diagram d2 = (Diagram)diagrams.get(j);
						if (isLiteral(d2)) {
							// Debug info 
							if (options.isDebug()) {
								log.reportMessage(" contradiction for "+toDebugString(d1)+" and "+toDebugString(d2));
							}

							// Compute contradiction between d1 and d2
							// both are unitary diagrams or not unitary diagrams
							Contradiction temp = chk.computeContradiction(d1, d2);			
							if (temp.isContradiction()) {
								con.setContradiction(true);
								con.getZones().addAll(temp.getZones());
							} 
						}					
					}
				}
			}
		}
		return con;
	}
	
	/** Choose a binary node as a splitting candidate */
	protected int choose() {
		for(int i=0; i<diagrams.size(); i++) {
			Diagram d = (Diagram)diagrams.get(i);
			if (isLiteral(d)) ;
			else return i;
		}
		return -1;
	}
	
	/** Make a not diagram */
	protected Diagram makeNot(Diagram d) {
		Vector diags = new Vector();
		diags.add(d);
		Diagram result = new NotCompoundDiagram(diags, "");
		return result;
	}
	
	/** Add diagrams */
	protected void addDiagrams(List target, List source, int l, int r) {
		for(int i=l; i<r; i++) {
			target.add(source.get(i));
		}		
	}

	/** Check if it's a literal */
	protected boolean isLiteral(Diagram d) {
		return isAtom(d) || isNotAtom(d);
	}

	/** Check if it's an atom: unitary diagram */
	protected boolean isAtom(Diagram d) {
		if (d instanceof UnitaryDiagram) return true;
		return false;
	}

	/** Check if it's not unitary diagram */
	protected boolean isNotAtom(Diagram d) {
		if (d instanceof NotCompoundDiagram) {
			Diagram d1 = (Diagram)((NotCompoundDiagram)d).getChildren().get(0);
			if (d1 instanceof UnitaryDiagram) return true;
		}
		return false;
	}

	/** Check if there are some obvious cases */
	public boolean isObviousCase() {
		return isObviousTrueCase() || isObviousFalseCase(); 
	}
	
	/** Check is there are some obvious true cases */
	public boolean isObviousTrueCase() {
		// Only one diagram
		if (diagrams.size() == 1) {
			Diagram d = (Diagram)diagrams.get(0);
			if (d instanceof TrueDiagram || d.getClass() == UnitaryDiagram.class) {
				return true;
			} else if (d instanceof NotCompoundDiagram) {
				Diagram child = (Diagram)((NotCompoundDiagram)d).getChildren().get(0);
				if (child instanceof FalseDiagram) {
					return true;
				} else {
					return false;
				}
			} else {
				return false;
			}
		// Several diagrams
		} else {
			// All are true and 
			for(int i=0; i<diagrams.size(); i++) {
				Diagram d = (Diagram)diagrams.get(i);
				if (d instanceof TrueDiagram) ;
				else return false;
			}
			return true;
		}
	}

	/** Check is there are some obvious false cases */
	public boolean isObviousFalseCase() {
		// Only one diagram
		if (diagrams.size() == 1) {
			Diagram d = (Diagram)diagrams.get(0);
			if (d instanceof FalseDiagram) {
				return true;
			} else if (d instanceof NotCompoundDiagram) {
				Diagram child = (Diagram)((NotCompoundDiagram)d).getChildren().get(0);
				if (child instanceof TrueDiagram) {
					return true;
				} else {
					return false;
				}
			} else {
				return false;
			}
		// Several diagrams
		} else {
			// Is a leaf
			if (!allDiagramsAreLiterals())
				return false;
			// At least a false
			for(int i=0; i<diagrams.size(); i++) {
				Diagram d = (Diagram)diagrams.get(i);
				if (d instanceof FalseDiagram) 
					return true;
			}
			// X and not X
			Set ds = new LinkedHashSet();
			Set nds = new LinkedHashSet();
			for(int i=0; i<diagrams.size(); i++) {
				Diagram d = (Diagram)diagrams.get(i);
				if (d instanceof UnitaryDiagram) {
					UnitaryDiagram ud = (UnitaryDiagram)d;
					ds.add(ud);
					if (contains(nds, ud))
						return true;
				} else if (d instanceof NotCompoundDiagram) {
					UnitaryDiagram child = (UnitaryDiagram)((NotCompoundDiagram)d).getChildren().get(0);
					nds.add(child);
					if (contains(ds, child))
						return true;
				}
			}
			return false;
		}
	}
	protected boolean contains(Set ds, UnitaryDiagram d) {
		for(Iterator i=ds.iterator(); i.hasNext(); ) {
			UnitaryDiagram temp = (UnitaryDiagram)i.next();
			if (temp.equals(d))
				return true;
		}
		return false;
	}
	
	/** Contained diagrams */
	public List getDiagrams() { return diagrams; }
	public void setDiagrams(List diagrams) { this.diagrams = diagrams; }
	public void addDiagram(Diagram diag) { diagrams.add(diag); }

	/** Operation kind */
	public OperationKind getOperationKind() { return kind; }
	public void setOperationKind(OperationKind kind) { this.kind = kind; }
	
	/** Debug string */
	protected String toDebugString(Diagram d) {
		if (d instanceof UnitaryDiagram && !(d instanceof TrueDiagram) && !(d instanceof FalseDiagram))
			return ((UnitaryDiagram)d).toDebugString();
		else
			return d.toString();
	}

	/** Override toString */
	public String toString() {
		String res = "{";
		for(int i=0; i<diagrams.size(); i++) {
			if (i != 0) res += ", ";
			res += diagrams.get(i).toString();
		}
		res += "}";
		return res;
	}
	
	// Local properties
	protected ETableauImpl tableau;
	protected Options options;
	protected ILog log;
	protected List diagrams = new Vector();
	protected OperationKind kind;
}
