package rwd.tableaux.semantic;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.swing.event.TreeModelEvent;
import javax.swing.event.TreeModelListener;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.*;

import java.util.*;
import java.util.List;

import rwd.abstractsyntax.eulerdiagram.*;
import rwd.abstractsyntax.spiderdiagram.*;
import rwd.concretesyntax.ConcreteDiagram;
import rwd.gui.editor.viewers.*;
import rwd.tableaux.diagram.*;
import rwd.tableaux.syntax.ast.*;
import rwd.tableaux.syntax.parser.*;
import rwd.views.abstracttoconcrete.forcemodel.SpiderPlacer;

import uk.ac.kent.cs.kmf.util.*;

/**
 * @author Octavian Patrascoiu
 *
 */
public class ETableauImpl
	extends JTree
	implements ETableau
{
	/** Constructors */
	public ETableauImpl(JDesktopPane desktop) {
		init(desktop, log);
	}
	public ETableauImpl(JDesktopPane desktop, ILog log) {
		init(desktop, log);
	}
	protected void init(JDesktopPane desktop, ILog log) {
		this.desktop = desktop;
		this.log = log;
		this.setEditable(false);
		this.addMouseListener(new TableauMouseListener(this, new RulePopupMenu(this)));
		// Add selection listener
		addTreeSelectionListener(new TreeSelectionListener() {
			public void valueChanged(TreeSelectionEvent e) {
				selectNodeAction(e);
			}
		});
		// Add changed listener
		getModel().addTreeModelListener(new TreeModelListener() {
			public void valueChanged(TreeModelEvent e) {
				expandTree();
			}
			public void treeNodesChanged(TreeModelEvent arg0) {			
				expandTree();
			}
			public void treeNodesInserted(TreeModelEvent arg0) {
				expandTree();
			}
			public void treeNodesRemoved(TreeModelEvent arg0) {
				expandTree();
			}
			public void treeStructureChanged(TreeModelEvent arg0) {
				expandTree();
			}
		});
	}
	
	//
	// Expandable tableau interface
	//
	/** Build a JTree representation of the tableau */
	public JTree build(String strExp, Options opt, List cDiags) {
		// Debug message
		log.reportMessage("Parsing ...");
		// Parse the input string and build an AST
		LogicalExp exp = parser.parse(strExp, log);
		// Init the tableau from AST
		return build(exp, opt, cDiags);
	}

	/** Build a JTree representation of the tableau */
	public JTree build(LogicalExp exp, Options opt, List diags) {
		try {
			// Convert the logic expression to a diagram
			log.reportMessage("Converting logical expression to diagrams ...");
			Expression2Diagram converter = new Expression2DiagramImpl(log);
			Diagram d = converter.convert(exp, diags, opt);

			// Build the tableau
			return build(d, opt, diags);
		} catch (Exception e) {
			e.printStackTrace();
			setDrawable(false);
			log.reportError("Cannot draw table "+e.getMessage());
			return null;
		}
	}
	/** Build a JTree representation of the tableau */
	public JTree build(Diagram d, Options opt, List diags) {
		if (d == null) {
			setDrawable(false);
			return null;
		} 
		
		// Set options
		this.opts = opt;
		this.diags = diags;
		
		// Debug info
		log.reportMessage("Building the tableau ...");
		if (opt.isDebug()) {
			log.reportMessage(diags.toString());
		}
				
		// Create the first node
		List diagrams = new Vector();
		diagrams.add(d);
		ENodeImpl en = new ENodeImpl(this, diagrams);
		DefaultMutableTreeNode root = makeTreeNode(en);

		// Create model
		myModel = new DefaultTreeModel((DefaultMutableTreeNode)root); 
		setModel(myModel);
		
		// Expand entire tree
		expandTree();
		ToolTipManager.sharedInstance().registerComponent(this);
		this.setCellRenderer(new ETableauRenderImpl(diags, root, opt, log));
		
		return this;
	}

	/**
	 * Make a JTree node and add graphic info
	 */
	protected DefaultMutableTreeNode makeTreeNode(ENode eNode) {
		DefaultMutableTreeNode root = new DefaultMutableTreeNode(eNode);
		// Add graphic information
		addGraphicInfo(root);
		return root;	
	}
				
	/** 
	 *  Add the graphic information for unitary diagrams 
	 *  List of unitary diagrams and negation of unitary diagrams
	 * 
	 **/
	protected void addGraphicInfo(DefaultMutableTreeNode node) {
		// Add graphic info only is a leaf: X and not X
		Object userObject = node.getUserObject();
		if (userObject instanceof ENode && userObject != null) {
			ENode eNode = (ENode)userObject;
			if (eNode.allDiagramsAreLiterals()) {
				DefaultMutableTreeNode info = new DefaultMutableTreeNode("diagrams", true);
				// Horizontal versus vertical display
				if (opts.isHorizontal()) {
					addHorizontalInfo(info, eNode);
				} else {
					addVerticalInfo(info, eNode);
				}
				// Add not empty info
				if (info.children() != null && info.children().hasMoreElements())
					addNode(node, info);
				// Add zones with contradiction
				if (eNode.isLeaf()) {
					Contradiction con = eNode.computeContradiction();
					Set contZones = con.getZones(); 
					if (contZones.size() != 0) {
						DefaultMutableTreeNode conRoot = new DefaultMutableTreeNode("contradictions");
						for(Iterator i=contZones.iterator(); i.hasNext();) {
							Zone z = (Zone)i.next();
							DefaultMutableTreeNode n = new DefaultMutableTreeNode(z);
							addNode(conRoot, n);
						}
						addNode(node, conRoot);
					}
				}
			}
		}
	}

	/** Make the horizontal graphic information */
	protected void addHorizontalInfo(DefaultMutableTreeNode parent, ENode eNode) {
		// Remove the previous graphic info if it exists
		removeHorizontalInfo(parent);
		// Compute the component that should be added
		if (eNode != null) {
			// Add zones with contradiction
			Set contZones = new TreeSet();
			if (eNode.isLeaf()) {
				Contradiction con = eNode.computeContradiction();
				contZones = con.getZones(); 
			}
			Component comp = makeDiagramsPanel(eNode.getDiagrams(), opts, contZones);
			if (comp == null) return;
			DefaultMutableTreeNode root = new DefaultMutableTreeNode(comp, true);
			addNode(parent, root);
		}
	}

	/** Make the vertical info */
	protected void addVerticalInfo(DefaultMutableTreeNode parent, ENode eNode) {
		// Remove the previous graphic info if it exists
		removeVerticalInfo(parent);
		// Compute the components that should be added
		if (eNode != null) {
			List diagrams = eNode.getDiagrams();
			diagrams = cleanDiagrams(diagrams);
			if (diagrams.size() == 0) return;
			// Create and add a node for each diagram
			for(int i=0; i<diagrams.size(); i++) {
				Diagram d = (Diagram)diagrams.get(i);
				DefaultMutableTreeNode n = new DefaultMutableTreeNode(d, true);
				addNode(parent, n);
			}
		}
	}

	// Remove the previous graphic info if it exists
	protected void removeHorizontalInfo(DefaultMutableTreeNode parent) {
		removeChildrenOf(parent);
	}

	// Remove the previous graphic info if it exists
	protected void removeVerticalInfo(DefaultMutableTreeNode parent) {
		removeChildrenOf(parent);
	}

	/** Create a panel to store the diagrams */
	protected Component makeDiagramsPanel(List diagrams, Options options, Set contZones) {
		// Clean the diagrams
		diagrams = cleanDiagrams(diagrams);
		if (diagrams.size() == 0) return null;
		// Create a panel and add components
		MatrixPanel generalPanel = new MatrixPanel();
		int line = 0;
		int column = 0;
		// Add components
		for(int i=0; i<diagrams.size(); i++) {
			Diagram d = (Diagram)diagrams.get(i);
			Component comp = makeComponent((UnitaryDiagram)d, options, contZones);
			// Add it
			generalPanel.add(comp, line, column);
			// Compute the next location to add
			if (column < options.getDiagramsPerLine()-1) {
				column++;				
			} else {
				column = 0;
				line++;
			}
		}
		JScrollPane scrollPane = new JScrollPane(generalPanel);
		return (Component)scrollPane;				
	}

	/** Remove true, false, and duplicates from a leaf */
	protected List cleanDiagrams(List diagrams) {
		// Clean the diagrams
		Set addedDiagrams = new TreeSet();
		List newDiags = new Vector();
		for(int i=0; i<diagrams.size(); i++) {
			// Create component
			Diagram d = (Diagram)diagrams.get(i);
			// Remove true, false, and duplicated unitary diagrams
			if (d.getClass() == UnitaryDiagram.class) {
				String key = d.getName();
				if (!addedDiagrams.contains(key)) {
					addedDiagrams.add(key);
					newDiags.add(d);
				}
			// Scan the compound diagrams
			} else if (d instanceof NotCompoundDiagram) {
				List children = ((NotCompoundDiagram)d).getChildren();
				Diagram child = (Diagram)children.get(0);
				if (child.getClass() == UnitaryDiagram.class) {
					String key = child.getName();
					if (!addedDiagrams.contains(key)) {
						addedDiagrams.add(key);
						newDiags.add(child);
					}
				}
			}
		}
		return newDiags;		
	}
	protected Component makeComponent(UnitaryDiagram ud, Options options, Set contZones) {
		// Create the viewer
		int jspSize = (int)(options.getDiagramSize());
		// Smooth or not
		if (options.isSmooth()) {
			DiagramViewerInterface viewer = new DiagramViewerComponent(jspSize, jspSize, ud, true, true); 
			SpiderPlacer sp = new SpiderPlacer(viewer);
			sp.embed();
			viewer.setContradictionZones(contZones);
			// Fit in
			viewer.fitToWidth(options.getDiagramSize());
			viewer.setBezier(true);
			ConcreteDiagram cd = viewer.getConcreteDiagram();
			cd.refreshZones(true);
			cd.notifyObservers();
			viewer.redrawImage();
			return (Component)viewer;
		} else {
			DiagramViewerComponent viewer = new DiagramViewerComponent(jspSize, jspSize, ud, true, false);
			viewer.setContradictionZones(contZones);
			SpiderPlacer sp = new SpiderPlacer(viewer);
			sp.embed();
			// Fit in
			viewer.fitToWidth(options.getDiagramSize());
			viewer.setBezier(false);
			viewer.redrawImage();
			return (Component)viewer;
		}
	}

	//
	// JTree functions
	//
	/** Expand the tree */
	protected void expandTree() {
		for(int i=0; i<getRowCount(); i++) {
			expandRow(i);
		}				
	}
	
	/** Add newParent as child or replace parent */
	protected void addOrReplace(DefaultMutableTreeNode parent, DefaultMutableTreeNode newParent) {
		boolean replace = false;
		if (replace) {
			// Update the tree and repaint it - change the model
			if (parent == getModel().getRoot()) {
				((DefaultTreeModel)getModel()).setRoot(newParent);
			} else {
				DefaultMutableTreeNode grand = (DefaultMutableTreeNode)parent.getParent();
				int index = grand.getIndex(parent);
				removeNode(parent);
				addNode(grand, newParent, index);
			}
		} else {
			// Remove all the children of but the graphic info
			Enumeration children = parent.children();
			while (children.hasMoreElements()) {
				DefaultMutableTreeNode child = (DefaultMutableTreeNode)children.nextElement();
				Object userObject = child.getUserObject();
				if (userObject instanceof ENode) 
					this.removeNode(child);
			}
			// Add new child
			addNode(parent, newParent);
		}
		expandTree();
	}

	/** Add several nodes */
	protected void addNode(DefaultMutableTreeNode parent, DefaultMutableTreeNode child, int i) {
		DefaultTreeModel model = (DefaultTreeModel)getModel(); 
		model.insertNodeInto(child, parent, i);
	}
	protected void addNode(DefaultMutableTreeNode parent, DefaultMutableTreeNode child) {
		addNode(parent, child, parent.getChildCount());
	}
	protected void addNodes(DefaultMutableTreeNode parent, List nodes) {
		for(int i=0; i<nodes.size(); i++) {
			DefaultMutableTreeNode n = (DefaultMutableTreeNode)nodes.get(i);
			addNode(parent, n);
			// There is no contradiction
			Object userObject = n.getUserObject();
			if (userObject instanceof ENode && userObject != null) {
				ENode eNode = (ENode)userObject;
			}
		}
		expandTree();
	}

	/** Remove a node */
	protected void removeChildrenOf(DefaultMutableTreeNode parent) {
		if (parent.children() == null || !parent.children().hasMoreElements()) return;
		// Make a clone
		List clone = new Vector();
		Enumeration children = parent.children();
		while (children.hasMoreElements()) {
			DefaultMutableTreeNode child = (DefaultMutableTreeNode)children.nextElement();
			clone.add(child);
		}
		for(int i=0; i<clone.size(); i++) {
			DefaultMutableTreeNode child = (DefaultMutableTreeNode)clone.get(i);
			removeNode(child);			
		}
	}

	/** Remove a node */
	protected void removeNode(DefaultMutableTreeNode child) {
		DefaultTreeModel model = (DefaultTreeModel)getModel(); 
		model.removeNodeFromParent(child);
	}

	/** Select node */
	protected void selectNodeAction(TreeSelectionEvent e) {
		TreePath path = getSelectionPath();
		if (path == null) return;        
		selectedNode = path.getLastPathComponent();
	}

	/** Is drawable flag */
	public boolean isDrawable() { return isDrawable; }
	public void setDrawable(boolean b) { this.isDrawable = b; }
	
	/** Log file */
	protected Options opts;
	protected List diags;
	protected DefaultTreeModel myModel;
	protected ENodeImpl root; 
	protected JDesktopPane desktop;
	protected ILog log;
	protected Parser parser = new ParserImpl();
	protected boolean isDrawable = true;
	protected Object selectedNode;
}

class TableauMouseListener 
	implements MouseListener 
{
	/** Constructor */
	public TableauMouseListener(ETableauImpl tableau, RulePopupMenu popup) {
		this.tableau = tableau;
		this.popup = popup;
	}
	
	// Catch all mouse events and redispatch them
	public void mouseMoved(MouseEvent e) {
		redispatchMouseEvent(e, false);
	}
	public void mouseDragged(MouseEvent e) {
		redispatchMouseEvent(e, false);
	}
	public void mouseClicked(MouseEvent e) {
		 redispatchMouseEvent(e, false);
	}
	public void mouseEntered(MouseEvent e) {
		redispatchMouseEvent(e, false);
	}
	public void mouseExited(MouseEvent e) {
		redispatchMouseEvent(e, false);
	}
	public void mousePressed(MouseEvent e) {
		redispatchMouseEvent(e, false);
	}
	public void mouseReleased(MouseEvent e) {
		redispatchMouseEvent(e, false);
	}

	private void redispatchMouseEvent(MouseEvent e, boolean repaint) {
		 // if it&apos;s a popup
		 if(e.isPopupTrigger()) {
			 // show the popup and return
			 popup.show(e.getComponent(), e.getX(), e.getY());
			 popup.validateItems();
		 } else {
//			 // get the mouse click point relative to the content pane
//			 Point containerPoint = SwingUtilities.convertPoint(
//			 	((JTree)tableau),
//			 	e.getPoint(),
//			 	contentPane);
//
//			 // find the component that under this point
//			 Component component = SwingUtilities.getDeepestComponentAt(
//			 	contentPane, 
//			 	containerPoint.x, 
//			 	containerPoint.y);
//
//			 // return if nothing was found
//			 if (component == null) {
//				 return;
//			 }
//
//			 // convert point relative to the target component
//			 Point componentPoint = SwingUtilities.convertPoint(
//			 	this, 
//			 	e.getPoint(), 
//			 	component);
//
//			 // redispatch the event
//			 component.dispatchEvent(new MouseEvent(
//			 	component,
//			 	e.getID(),
//			 	e.getWhen(),
//			 	e.getModifiers(),
//			 	componentPoint.x,
//			 	componentPoint.y,
//			 	e.getClickCount(),
//			 	e.isPopupTrigger()));
		 }
	 }

	// Local properties
	protected ETableauImpl tableau;
	protected RulePopupMenu popup;
}

// Context dependant menu
class RulePopupMenu 
	extends JPopupMenu
{
	/** Constructor */
	public RulePopupMenu(ETableauImpl tableau) {
		super();
		init(tableau);
	}
	
	/** Init menu items */
	protected void init(ETableauImpl tableau) {
		this.tableau = tableau;
		// Add menu items
		expand = makeItem("Expand compound diagram");
		add(expand);
		expand.addActionListener(new ExpandListener(tableau));
		
		expandAll = makeItem("Expand all compound diagrams");
		add(expandAll);
		expandAll.addActionListener(new ExpandAllListener(tableau));
		
		equaliseContours = makeItem("Equalise contours");
		add(equaliseContours);
		equaliseContours.addActionListener(new EqualiseContoursListener(tableau));
		
		equaliseZones = makeItem("Equalise zones");
		add(equaliseZones);
		equaliseZones.addActionListener(new EqualiseZonesListener(tableau));

		equaliseDiagrams = makeItem("Equalise diagrams");
		add(equaliseDiagrams);
		equaliseDiagrams.addActionListener(new EqualiseDiagramsListener(tableau));

		addContours = makeItem("Add contours");
		add(addContours);
		addContours.addActionListener(new AddContoursListener(tableau));
		
		addZone = makeItem("Add zone");
		add(addZone);
		addZone.addActionListener(new AddZoneListener(tableau));

		splitSpiders = makeItem("Split spiders");
		add(splitSpiders);
		splitSpiders.addActionListener(new SplitSpidersListener(tableau));

		splitSpider = makeItem("Split a spider");
		add(splitSpider);
		splitSpider.addActionListener(new SplitSpiderListener(tableau));

		negation = makeItem("Negate unitary diagrams");
		add(negation);
		negation.addActionListener(new NegationListener(tableau));
	}
	protected JMenuItem makeItem(String name) {
		JMenuItem item = new JMenuItem(name);
		item.setEnabled(false);
		return item;
	}
	protected void validateItems() {
		// Validate the menu items
		Object selectedNode = tableau.selectedNode;
		// Init menu fields
		// Get the user object and draw it		
		if (selectedNode instanceof DefaultMutableTreeNode && selectedNode != null) {
			Object userObject = ((DefaultMutableTreeNode)selectedNode).getUserObject();
			// User node
			if (userObject instanceof ENode) {
				ENode en = (ENode)userObject;
				// Debug version
				if (tableau.opts.isDebug()) {
					expand.setEnabled(true);
					expandAll.setEnabled(true);
					equaliseContours.setEnabled(true);
					equaliseZones.setEnabled(true);
					equaliseDiagrams.setEnabled(true);

					addContours.setEnabled(true);
					addZone.setEnabled(true);
						
					splitSpiders.setEnabled(true);
					splitSpider.setEnabled(true);
						
					negation.setEnabled(true);
					return;
				}
				// All diagrams are X or not X X: unitary
				if (en.allDiagramsAreLiterals()) {
					expand.setEnabled(false);
					expandAll.setEnabled(false);
					// Special case only true and false
					if (en.isObviousCase()) {
						equaliseContours.setEnabled(false);
						equaliseZones.setEnabled(false);
						equaliseDiagrams.setEnabled(false);

						addContours.setEnabled(false);
						addZone.setEnabled(false);
						
						splitSpiders.setEnabled(false);
						splitSpider.setEnabled(false);
						
						negation.setEnabled(false);
					} else {
						equaliseContours.setEnabled(!en.allDiagramsAreEqualised());
						equaliseZones.setEnabled(!en.allDiagramsAreEqualised());
						equaliseDiagrams.setEnabled(!en.allDiagramsAreEqualised());

						addContours.setEnabled(!en.allDiagramsAreEqualised());
						addZone.setEnabled(!en.allDiagramsAreEqualised());
						
						splitSpiders.setEnabled(!en.allSpidersHaveOneFoot());
						splitSpider.setEnabled(!en.allSpidersHaveOneFoot());
						
						negation.setEnabled(en.containsNot());
					}
				// At least an And, Or, or deMorgan
				} else if (!en.allDiagramsAreLiterals()) {
					expand.setEnabled(true);
					expandAll.setEnabled(true);

					equaliseContours.setEnabled(false);
					equaliseZones.setEnabled(false);
					equaliseDiagrams.setEnabled(false);

					addContours.setEnabled(false);
					addZone.setEnabled(false);
					
					splitSpiders.setEnabled(false);
					splitSpider.setEnabled(false);
					
					negation.setEnabled(false);
				}
			// Diagrams
			} else if (userObject instanceof Diagram) {
				expand.setEnabled(false);
				expandAll.setEnabled(false);

				equaliseContours.setEnabled(false);
				equaliseZones.setEnabled(false);
				equaliseDiagrams.setEnabled(false);				

				addContours.setEnabled(true);
				addZone.setEnabled(true);
				splitSpiders.setEnabled(true);
				splitSpider.setEnabled(true);
				negation.setEnabled(false);
			// Else 
			} else {
				expand.setEnabled(false);
				expandAll.setEnabled(false);

				equaliseContours.setEnabled(false);
				equaliseZones.setEnabled(false);
				equaliseDiagrams.setEnabled(false);				

				addContours.setEnabled(false);
				addZone.setEnabled(false);
				splitSpiders.setEnabled(false);
				splitSpider.setEnabled(false);
				negation.setEnabled(false);
			}
		}
	}
	
	// Local properties
	protected JMenuItem expand;
	protected JMenuItem expandAll;
	
	protected JMenuItem equaliseContours;
	protected JMenuItem equaliseZones;
	protected JMenuItem equaliseDiagrams;

	protected JMenuItem addContours;
	protected JMenuItem addZone;
	protected JMenuItem splitSpiders;
	protected JMenuItem splitSpider;
	protected JMenuItem negation;
	
	protected ETableauImpl tableau;
}

// Listeners for each operation
class ExpandListener
	implements ActionListener
{
	public ExpandListener(ETableauImpl tableau) {
		this.tableau = tableau;
	}
	
	public void actionPerformed(ActionEvent e) {
		Object selectedNode = tableau.selectedNode;
		if (selectedNode instanceof DefaultMutableTreeNode && selectedNode != null) {
			DefaultMutableTreeNode parent = (DefaultMutableTreeNode)selectedNode; 
			Object userObject = parent.getUserObject();
			if (userObject instanceof ENode && userObject != null) {
				ENode eNode = (ENode)userObject;
				if (!eNode.allDiagramsAreLiterals()) {
					// Remove all the children of
					tableau.removeChildrenOf(parent);
					// Split the node
					List eNodes = eNode.expand();
					List nodes = new Vector();
					for(int i=0; i<eNodes.size(); i++) {
						DefaultMutableTreeNode jNode = tableau.makeTreeNode((ENode)eNodes.get(i)); 
						tableau.addNode(parent, jNode);
					}
					// Expand tree
					tableau.expandTree();
				}
			}
		}
	}
	
	// Local properties
	protected ETableauImpl tableau;

}
// Expand all listener
class ExpandAllListener
	implements ActionListener
{

	public ExpandAllListener(ETableauImpl tableau) {
		this.tableau = tableau;
	}

	public void actionPerformed(ActionEvent e) {
		Object selectedNode = tableau.selectedNode;
		if (selectedNode instanceof DefaultMutableTreeNode && selectedNode != null) {
			DefaultMutableTreeNode parent = (DefaultMutableTreeNode)selectedNode; 
			Object userObject = parent.getUserObject();
			if (userObject instanceof ENode && userObject != null) {
				ENode eNode = (ENode)userObject;
				if (!eNode.allDiagramsAreLiterals()) {
					// Remove all the children of
					tableau.removeChildrenOf(parent);
					// Expand it
					DefaultMutableTreeNode root = expandAll(eNode);
					// Replace parent with root
					if (parent == tableau.getModel().getRoot()) {
						((DefaultTreeModel)tableau.getModel()).setRoot(root);
					} else {
						DefaultMutableTreeNode grand = (DefaultMutableTreeNode)parent.getParent();
						int index = grand.getIndex(parent);
						tableau.removeNode(parent);
						tableau.addNode(grand, root, index);
					}
					// Expand tree
					tableau.expandTree();
				}
			}
		}
	}

	/** Expands an ENode and all the subnodes: creates a list of tree nodes */
	protected DefaultMutableTreeNode expandAll(ENode eRoot) {	
		// Create a queue of nodes
		Queue nodes = new QueueImpl();
		Queue jNodes = new QueueImpl();
		// Push node into the queue
		nodes.push(eRoot);
		DefaultMutableTreeNode jRoot = new DefaultMutableTreeNode(eRoot);
		jNodes.push(jRoot);
		// While there are nodes
		while (!nodes.isEmpty()) {
			// Debug info
			if (tableau.opts.isDebug()) {
				tableau.log.reportMessage("Queue: "+nodes.toString());
			}
			
			// Get current node
			ENode crtNode = (ENode)nodes.pop();
			DefaultMutableTreeNode jCrtNode= (DefaultMutableTreeNode)jNodes.pop(); 
			
			// Debug info
			if (tableau.opts.isDebug()) {
				tableau.log.reportMessage("Current node: "+crtNode.toString());
			}
			
			// crtNode contains only literals:
			// unitary diagrams and not unitary diagram
			if (crtNode.allDiagramsAreLiterals()) {
			// crtNode contains at least one binary operator
			} else {
				// Split crtNode into several nodes
				List newNodes = crtNode.expand();
				// For each new node
				for(int i=0; i<newNodes.size(); i++) {
					ENode en = (ENode)newNodes.get(i);
					// Make a jNode
					DefaultMutableTreeNode jn =  tableau.makeTreeNode(en);
					// Push nodes
					nodes.push(en);
					jNodes.push(jn);
					// Add the child
					jCrtNode.add(jn);
				}				
			}
		}

		return jRoot;
	}

	// Local properties
	protected ETableauImpl tableau;
}
// Equalise contours listener
class EqualiseContoursListener
	implements ActionListener
{

	public EqualiseContoursListener(ETableauImpl tableau) {
		this.tableau = tableau;
	}

	public void actionPerformed(ActionEvent e) {
		Object selectedNode = tableau.selectedNode;
		if (selectedNode instanceof DefaultMutableTreeNode && selectedNode != null) {
			DefaultMutableTreeNode parent = (DefaultMutableTreeNode)selectedNode; 
			Object userObject = parent.getUserObject();
			if (userObject instanceof ENode && userObject != null) {
				ENode eNode = (ENode)userObject;
				if (!eNode.allContoursAreEqualised()) {
					// Set operation kind
					eNode.setOperationKind(OperationKind.EQUALISE_CONTOURS);
					// Build the new node
					List diagrams = eNode.getDiagrams();
					// Equalise the contours
					Equaliser eq = new EqualiserImpl(tableau.opts, tableau.log);
					List newDiagrams = eq.equaliseContours(diagrams);
					// Make a new ENodeImpl
					ENodeImpl newNode = new ENodeImpl(tableau, newDiagrams);
					DefaultMutableTreeNode newParent = tableau.makeTreeNode(newNode);
					// Add new parent into the model
					tableau.addOrReplace(parent, newParent);
				}
			}
		}
	}
	
	// Local properties
	protected ETableauImpl tableau;
}
// Equalise zones listener
class EqualiseZonesListener
	implements ActionListener
{

	public EqualiseZonesListener(ETableauImpl tableau) {
		this.tableau = tableau;
	}

	public void actionPerformed(ActionEvent e) {
		Object selectedNode = tableau.selectedNode;
		if (selectedNode instanceof DefaultMutableTreeNode && selectedNode != null) {
			DefaultMutableTreeNode parent = (DefaultMutableTreeNode)selectedNode; 
			Object userObject = parent.getUserObject();
			if (userObject instanceof ENode && userObject != null) {
				ENode eNode = (ENode)userObject;
				if (!eNode.allZonesAreEqualised()) {
					// Set operation kind
					eNode.setOperationKind(OperationKind.EQUALISE_ZONES);
					List diagrams = eNode.getDiagrams();
					// Equalise the contours
					Equaliser eq = new EqualiserImpl(tableau.opts, tableau.log);
					List newDiagrams = eq.equaliseZones(diagrams);
					// Make a new ENode
					ENodeImpl newNode = new ENodeImpl(tableau, newDiagrams);
					DefaultMutableTreeNode newParent = tableau.makeTreeNode(newNode);;
					// Add new parent into the model
					tableau.addOrReplace(parent, newParent);
				}
			}
		}
	}
	
	// Local properties
	protected ETableauImpl tableau;
}
// Equalise diagrams listener
class EqualiseDiagramsListener
	implements ActionListener
{

	public EqualiseDiagramsListener(ETableauImpl tableau) {
		this.tableau = tableau;
	}

	public void actionPerformed(ActionEvent e) {
		Object selectedNode = tableau.selectedNode;
		if (selectedNode instanceof DefaultMutableTreeNode && selectedNode != null) {
			DefaultMutableTreeNode parent = (DefaultMutableTreeNode)selectedNode; 
			Object userObject = parent.getUserObject();
			if (userObject instanceof ENode && userObject != null) {
				ENode eNode = (ENode)userObject;
				if (!eNode.allDiagramsAreEqualised()) {
					// Set operation kind
					eNode.setOperationKind(OperationKind.EQUALISE_DIAGRAMS);
					List diagrams = eNode.getDiagrams();
					// Equalise the contours
					Equaliser eq = new EqualiserImpl(tableau.opts, tableau.log);
					List newDiagrams = eq.equaliseDiagrams(diagrams);
					// Make a new ENode
					ENodeImpl newNode = new ENodeImpl(tableau, newDiagrams);
					DefaultMutableTreeNode newParent = tableau.makeTreeNode(newNode);
					// Update the tree and repaint it - change the model
					// Add new parent into the model
					tableau.addOrReplace(parent, newParent);
				}
			}
		}
	}
	
	// Local properties
	protected ETableauImpl tableau;
}

// Add contours listener
class AddContoursListener
	implements ActionListener
{
	public AddContoursListener(ETableauImpl tableau) {
		this.tableau = tableau;
	}

	public void actionPerformed(ActionEvent e) {
		Object selectedNode = tableau.selectedNode;
		if (selectedNode instanceof DefaultMutableTreeNode && selectedNode != null) {
			DefaultMutableTreeNode parent = (DefaultMutableTreeNode)selectedNode; 
			Object userObject = parent.getUserObject();
			// On ENodes - leaves
			if (userObject instanceof ENode && userObject != null) {
				ENode eNode = (ENode)userObject;
				// Ask for contours
				List contours = getContours();
				if (contours.size() != 0) {
					if (!eNode.allDiagramsContainContours(contours)) {
						// Set operation kind
						eNode.setOperationKind(OperationKind.ADD_CONTOURS);
						// Create new Diagrams
						List diagrams = eNode.getDiagrams();
						// For each diagram
						List newDiagrams = new Vector();
						Equaliser eq = new EqualiserImpl(tableau.opts, tableau.log);
						for(int i=0; i<diagrams.size(); i++) {
							// Add all contours
							Diagram d = (Diagram)diagrams.get(i);
							Diagram newDiagram = null;
							if (d instanceof UnitaryDiagram) {
								UnitaryDiagram ud = (UnitaryDiagram)d;
								newDiagram = eq.addContours(ud, contours);
							} else if (d instanceof NotCompoundDiagram) {
								UnitaryDiagram ud = (UnitaryDiagram)((NotCompoundDiagram)d).getChildren().get(0);
								ud = eq.addContours(ud, contours);
								newDiagram = new NotCompoundDiagram(new Vector(), "");
								((NotCompoundDiagram)newDiagram).getChildren().add(ud);
							}
							// Add diagram
							newDiagrams.add(newDiagram);
						}
						// Update the tree and repaint it - change the model
						ENode newENode = new ENodeImpl(tableau, newDiagrams);
						DefaultMutableTreeNode newParent = tableau.makeTreeNode(newENode);
						// Add new parent into the model
						tableau.addOrReplace(parent, newParent);
					}
				}
			}
		}
	}

	protected List getContours() {
		// Ask for contours
		String inputString = JOptionPane.showInputDialog(
			tableau,
			"Please enter contour list",
			"a b");
		List contours = new Vector();
		if (inputString != null) {
			// Compute contours
			String names[] = inputString.split(" ");
			// Add contours
			for(int i=0; i<names.length; i++) {
				Contour c = Contour.getContour(names[i]);
				if (!contours.contains(c)) 
					contours.add(c);
			}
		}
		return contours;
	}
	
	// Local properties
	protected ETableauImpl tableau;
}
//Add zone listener
class AddZoneListener
	implements ActionListener
{
	public AddZoneListener(ETableauImpl tableau) {
		this.tableau = tableau;
	}

	public void actionPerformed(ActionEvent e) {
		Object selectedNode = tableau.selectedNode;
		if (selectedNode instanceof DefaultMutableTreeNode && selectedNode != null) {
			DefaultMutableTreeNode parent = (DefaultMutableTreeNode)selectedNode; 
			Object userObject = parent.getUserObject();
			// On ENodes - leaves
			if (userObject instanceof ENode && userObject != null) {
				ENode eNode = (ENode)userObject;
				// Ask for zone
				Zone z = getZone();
				if (z != null) {
					if (!eNode.allDiagramsContainZone(z)) {
						// Set operation kind
						eNode.setOperationKind(OperationKind.ADD_ZONE);
						// Create new Diagrams
						List diagrams = eNode.getDiagrams();
						// For each diagram
						List newDiagrams = new Vector();
						Equaliser eq = new EqualiserImpl(tableau.opts, tableau.log);
						for(int i=0; i<diagrams.size(); i++) {
							// Add all contours
							Diagram d = (Diagram)diagrams.get(i);
							Diagram newDiagram = null;
							if (d instanceof UnitaryDiagram) {
								UnitaryDiagram ud = (UnitaryDiagram)d;
								newDiagram = eq.addZone(ud, z);
							} else if (d instanceof NotCompoundDiagram) {
								UnitaryDiagram ud = (UnitaryDiagram)((NotCompoundDiagram)d).getChildren().get(0);
								ud = eq.addZone(ud, z);
								newDiagram = new NotCompoundDiagram(new Vector(), "");
								((NotCompoundDiagram)newDiagram).getChildren().add(ud);
							}
							// Add diagram
							newDiagrams.add(newDiagram);
						}
						// Update the tree and repaint it - change the model
						ENode newENode = new ENodeImpl(tableau, newDiagrams);
						DefaultMutableTreeNode newParent = tableau.makeTreeNode(newENode);
						// Add new parent into the model
						tableau.addOrReplace(parent, newParent);
					}
				}
			}
		}
	}

	protected Zone getZone() {	
		String input = JOptionPane.showInputDialog(
			tableau,
			"Please enter contour list",
			"a | b");
		Zone z = null;
		if (input != null) {
			// Compute zone
			input = input.trim();
			// Compute inside and outside contours
			Set inside = new TreeSet();
			Set outside = new TreeSet();
			String name = "";
			boolean isInside = true;
			for(int i=0; i<input.length(); i++) {
				char ch = input.charAt(i);
				if (ch == ' ') {
					if (isInside) {
						if (name.length() != 0) inside.add(Contour.getContour(name));
					} else {
						if (name.length() != 0) outside.add(Contour.getContour(name));
					}
					name = "";
				} else if (ch == '|') {
					if (isInside) {
						if (name.length() != 0) inside.add(Contour.getContour(name));
					} else {
						if (name.length() != 0) outside.add(Contour.getContour(name));
					}
					isInside = false;
					name = "";
				} else {
					name += ch;
				}
			}
			if (isInside) {
				if (name.length() != 0) inside.add(Contour.getContour(name));
			} else {
				if (name.length() != 0) outside.add(Contour.getContour(name));
			}
			// Get zone
			z = Zone.getZone(new Vector(inside), new Vector(outside));
		}
		return z;
	}
	
	// Local properties
	protected ETableauImpl tableau;
}

// Split spiders listener
class SplitSpidersListener
	implements ActionListener
{
	public SplitSpidersListener(ETableauImpl tableau) {
		this.tableau = tableau;
	}

	public void actionPerformed(ActionEvent e) {
		Object selectedNode = tableau.selectedNode;
		if (selectedNode instanceof DefaultMutableTreeNode && selectedNode != null) {
			DefaultMutableTreeNode parent = (DefaultMutableTreeNode)selectedNode; 
			Object userObject = parent.getUserObject();
			// On ENodes - leaves
			if (userObject instanceof ENode && userObject != null) {
				ENode eNode = (ENode)userObject;
				if (!eNode.allSpidersHaveOneFoot()) {
					// Set operation kind
					eNode.setOperationKind(OperationKind.SPLIT_SPIDERS);
					// Create new Diagrams
					List diagrams = eNode.getDiagrams();
					// For each diagram
					List newDiagrams = new Vector();
					SpiderSplitter sp = new SpiderSplitterImpl(tableau.opts, tableau.log);
					for(int i=0; i<diagrams.size(); i++) {
						// Add all contours
						Diagram d = (Diagram)diagrams.get(i);
						Diagram newDiagram = null;
						if (d instanceof UnitaryDiagram) {
							UnitaryDiagram ud = (UnitaryDiagram)d;
							newDiagram = sp.splitSpiders(ud);
						} else if (d instanceof NotCompoundDiagram) {
							UnitaryDiagram ud = (UnitaryDiagram)((NotCompoundDiagram)d).getChildren().get(0);
							Diagram newChild = sp.splitSpiders(ud);
							newDiagram = new NotCompoundDiagram(new Vector(), "");
							((NotCompoundDiagram)newDiagram).getChildren().add(newChild);
						}
						// Add diagram
						newDiagrams.add(newDiagram);
					}
					// Update the tree and repaint it - change the model
					ENode newENode = new ENodeImpl(tableau, newDiagrams);
					DefaultMutableTreeNode newParent = tableau.makeTreeNode(newENode);
					// Add new parent into the model
					tableau.addOrReplace(parent, newParent);
				}
			}
		}
	}
	
	// Local properties
	protected ETableauImpl tableau;
}

// Split a spider listener
class SplitSpiderListener
	implements ActionListener
{
	public SplitSpiderListener(ETableauImpl tableau) {
		this.tableau = tableau;
	}

	public void actionPerformed(ActionEvent e) {
		Object selectedNode = tableau.selectedNode;
		if (selectedNode instanceof DefaultMutableTreeNode && selectedNode != null) {
			DefaultMutableTreeNode parent = (DefaultMutableTreeNode)selectedNode; 
			Object userObject = parent.getUserObject();
			// On ENodes - leaves
			if (userObject instanceof ENode && userObject != null) {
				ENode eNode = (ENode)userObject;
				if (!eNode.allSpidersHaveOneFoot()) {
					// For each diagram
					List diagrams = eNode.getDiagrams();
					List newDiagrams = chooseSpiderAndSplitIt(diagrams);
					if (newDiagrams != null && newDiagrams.size() != 0) {
						eNode.setOperationKind(OperationKind.SPLIT_SPIDERS);
						// Create new Diagrams
						// Update the tree and repaint it - change the model
						ENode newENode = new ENodeImpl(tableau, newDiagrams);
						DefaultMutableTreeNode newParent = tableau.makeTreeNode(newENode);
						// Add new parent into the model
						tableau.addOrReplace(parent, newParent);
					}
				}
			}
		}
	}
	
	/** Choose a diagram, a spider and split it */
	protected List chooseSpiderAndSplitIt(List diagrams) {
		// Create a dialog to choose diagram and spider
		SpiderDialogImpl dialog = new SpiderDialogImpl(diagrams, null, "Choose Spider", tableau.log);
		int returnVal = dialog.showOpenDialog(null);
		if (returnVal == JFileChooser.APPROVE_OPTION) {
			// Get diagram and spiders
			UnitaryDiagram ud = dialog.getDiagram();
			Vector habitat = dialog.getSpider();
			// Split only the selected diagram
			SpiderSplitter sp = new SpiderSplitterImpl(tableau.opts, tableau.log);
			List newDiagrams = new Vector();
			for(int i=0; i<diagrams.size(); i++) {
				Diagram d = (Diagram)diagrams.get(i);
				if (d == ud) {
					newDiagrams.add(sp.splitSpider(ud, habitat));
				} else {
					newDiagrams.add(d);
				}
			}
			return newDiagrams;
		}
		return null;
	}
	
	// Local properties
	protected ETableauImpl tableau;
}

// Negation listener
class NegationListener
	implements ActionListener
{
	public NegationListener(ETableauImpl tableau) {
		this.tableau = tableau;
	}

	public void actionPerformed(ActionEvent e) {
		Object selectedNode = tableau.selectedNode;
		if (selectedNode instanceof DefaultMutableTreeNode && selectedNode != null) {
			DefaultMutableTreeNode parent = (DefaultMutableTreeNode)selectedNode; 
			Object userObject = parent.getUserObject();
			// On ENodes - leaves
			if (userObject instanceof ENode && userObject != null) {
				ENode eNode = (ENode)userObject;
				if (eNode.allDiagramsAreLiterals() && eNode.containsNot()) {
					eNode.setOperationKind(OperationKind.NEGATE_DIAGRAM);
					// Create new Diagrams
					List diagrams = eNode.getDiagrams();
					// For each diagram
					List newDiagrams = new Vector();
					Normaliser norm = new NormaliserImpl(tableau.opts, tableau.log);
					for(int i=0; i<diagrams.size(); i++) {
						// Add all contours
						Diagram d = (Diagram)diagrams.get(i);
						Diagram newDiagram = d;
						if (d instanceof NotCompoundDiagram) {
							UnitaryDiagram ud = (UnitaryDiagram)((NotCompoundDiagram)d).getChildren().get(0);
							newDiagram = norm.makeNegationOf(ud);
						}
						// Add diagram
						newDiagrams.add(newDiagram);
					}
					// Update the tree and repaint it - change the model
					ENode newENode = new ENodeImpl(tableau, newDiagrams);
					DefaultMutableTreeNode newParent = tableau.makeTreeNode(newENode);
					// Add new parent into the model
					tableau.addOrReplace(parent, newParent);
				}
			}
		}
	}
	
	// Local properties
	protected ETableauImpl tableau;
}

class SpiderDialogImpl
	extends JDialog
{
	public SpiderDialogImpl(List diagrams, JFrame parent, String title, ILog log) {
		super(parent, title);
		this.diagrams = diagrams;
		this.spiders = new Vector();
		for(int i=0; i<diagrams.size(); i++) {
			Diagram d = (Diagram)diagrams.get(i);
			if (d instanceof UnitaryDiagram) {
				UnitaryDiagram ud = (UnitaryDiagram)d;
				spiders.addAll(ud.getSpiders());
			}
		}
		MatrixPanel mainPanel = new MatrixPanel(); 
		setContentPane(mainPanel);
		this.log = log;
		
		//
		//--- Create and fill PROJECT panel
		//
		dataPanel = new MatrixPanel();
		int line = 0;
		int column = 0;
		JLabel label;
		JScrollPane scrollPanel;
		
		//--- Diagram area ---
		dataPanel.addLabel("Diagram", line, column);
		diagramList = dataPanel.addList(new Vector(diagrams), line, column+1);
		line++;

		//--- Spider area ---
		dataPanel.addLabel("Spider", line, column);
		spiderList = dataPanel.addList(new Vector(spiders), line, column+1);
		line++;
    
		//
		//--- Create and fill CONFIRM panel ---
		//
		MatrixPanel confirmPanel = new MatrixPanel();
		//--- OK button ---
		JButton okButton = confirmPanel.addButton("OK", 0, 0);
		okButton.addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(java.awt.event.ActionEvent evt) {
				okAction(evt);
			}
		});
		//--- Cancel button ---
		JButton cancelButton = confirmPanel.addButton("Cancel", 0, 1);
		cancelButton.setPreferredSize(dataPanel.getButtonSize());
		cancelButton.addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(java.awt.event.ActionEvent evt) {
				cancelAction(evt);
			}
		});

		//--- Add DATA panel to MAIN panel ---
		((MatrixPanel)getContentPane()).add(dataPanel, 0, 0);

		//--- Add CONFIRM to MAIN panel ---
		((MatrixPanel)getContentPane()).add(confirmPanel, 1, 0, GridBagConstraints.CENTER);
    		
		// Put frame in the middle of the screen
		pack();
		Dimension screen = getToolkit().getScreenSize();
		int width  = this.getSize().width+20;
		int height = this.getSize().height+20;
		int left = (screen.width - width)/2;
		int top = (screen.height - height)/2;
		setBounds(left, top, width, height);
	}

	//--- OK action ---
	protected int returnVal = JFileChooser.CANCEL_OPTION;
	protected void okAction(java.awt.event.ActionEvent evt) {
		returnVal = JFileChooser.APPROVE_OPTION;
		Diagram d = getDiagram();
		Vector s = getSpider();
		if (d != null && s != null)  	
			dispose();
	}

	//--- Cancel action ---
	protected void cancelAction(java.awt.event.ActionEvent evt) {
		returnVal = JFileChooser.CANCEL_OPTION;   	
		dispose();
	}

	//--- Get XMI file name ----
	public int showOpenDialog(JFrame parent) {
		this.setModal(true);
		show();
		return returnVal;
	}
	
	//--- Get diagram ----
	public UnitaryDiagram getDiagram() {
		int index = diagramList.getSelectedIndex();
		if (index != -1)
			return (UnitaryDiagram)diagrams.get(index);
		else
			return null;
	}

	//--- Get spider ----
	public Vector getSpider() {
		int index = spiderList.getSelectedIndex();
		if (index != -1)
			return ((Spider)spiders.get(index)).getHabitat();
		else
			return null;
	}
	
	//
	// Local members
	//
	protected MatrixPanel dataPanel;
	protected JList diagramList;
	protected JList spiderList;

	protected ILog log;
	protected List diagrams;
	protected List spiders;
}
