package rwd.views.abstracttoconcrete.converting;

import rwd.concretesyntax.eulerdiagram.ConcreteContour;
import rwd.concretesyntax.eulerdiagram.DecorationBox;
import rwd.concretesyntax.ConcreteDiagram;
import rwd.concretesyntax.ConcreteUnitaryDiagram;
import rwd.abstractsyntax.eulerdiagram.Zone;
import rwd.abstractsyntax.eulerdiagram.Contour;
import rwd.abstractsyntax.spiderdiagram.UnitaryDiagram;
import schow.drawing.AbstractMonoED;
import schow.drawing.polyline.MonoPolyLineED;
import schow.drawing.polyline.JordanLayout;
import schow.drawing.Utils;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.BitSet;
import java.util.Vector;
import java.awt.Polygon;
import java.util.TreeSet;
import java.util.Enumeration;
import java.awt.geom.GeneralPath;
import java.awt.geom.AffineTransform;

/**
 * @author Stirling Chow
 *
 */

// This class takes an abstract UnitaryDiagram and applies the
// UVic radialiser algorithm to create a corresponding
// ConcreteUnitaryDiagram.

public class Radialiser  {
  private Hashtable contourMap;
  private ConcreteUnitaryDiagram cud;

    public Radialiser(UnitaryDiagram ud) {
      // Create a map that uniquely assigns a nature
      // number to each contour in the range 0..|contours|-1.
      contourMap = new Hashtable();

      // Convert the zones into a TreeSet and also
      // map each zone to a bitstring.
      BitSet[] zones = new BitSet[ud.getNumZones()];
      TreeSet Z = new TreeSet();

      Iterator zoneIt = ud.getZoneIterator();
      int i = 0;
      while (zoneIt.hasNext()) {
      Zone zone = (Zone) zoneIt.next();
      zones[i++] = mapZone(zone);
      Z.add(zone);
    }

      // Check whether the fullset is in the list of zones.
      boolean hasFullset = false;
      for (i = 0; i < ud.getNumZones() && !hasFullset; i++)
        if (zones[i].cardinality() == contourMap.size())
          hasFullset = true;

      // If there is no full set, add it to the list of bitstrings.
      if (!hasFullset) {
        BitSet[] newZones = new BitSet[zones.length+1];
        for (i = 0; i < zones.length; i++)
          newZones[i] = zones[i];
          newZones[zones.length] = new BitSet();
          for (int c = 0; c < contourMap.size(); c++)
            newZones[zones.length].set(c);
          zones = newZones;
        }

      // Generate a basic Euler diagram for the specified zones.
    MonoPolyLineED ED = new MonoPolyLineED(
                  contourMap.size(),
                  zones,
                  AbstractMonoED.EULER_EDWARDS,
                  MonoPolyLineED.LAYOUT_CONVEXDELTA);

    // If there is no full set, set it's population to 0.
    if (!hasFullset)
      ED.setPop(zones[zones.length-1], 0);

    // Create a concrete diagram
    cud = new ConcreteUnitaryDiagram(getBox(ud.getName(), ED.getLayout()));

    // Add a polygon for each contour to the concrete diagram.
    for (Enumeration e = contourMap.keys(); e.hasMoreElements() ;) {
      Contour contour = (Contour) e.nextElement();
      Polygon polygon = getPolygon(contour, ED.getLayout());
      cud.addConcreteContour(new ConcreteContour(contour, polygon));
    }
    }

    public ConcreteUnitaryDiagram getCUD() {
        return cud;
    }

    private BitSet mapZone(Zone z) {
      // Create an initially empty bitset to use as a label
      // for the zone.
      BitSet label = new BitSet();

      // For each contour containing the zone, map it to a unique
      // integer between 0 and #contours in ud, and set the bit
      // in the zone's label to indicate that it is enclosed by
      // that contour.
      Vector contours = z.getContainingContours();
        Iterator contourIt = contours.iterator();
      while (contourIt.hasNext()) {
        Contour contour = (Contour) contourIt.next();
        if (!contourMap.containsKey(contour))
          contourMap.put(contour, new Integer(contourMap.size()));
        Integer val = (Integer) contourMap.get(contour);
        label.set(val.intValue());
      }
      return label;
    }

    private Polygon getPolygon(Contour contour, JordanLayout layout) {
      // Get the corresponding curve's polygon from the layout
      Integer c = (Integer) contourMap.get(contour);
      GeneralPath path = layout.getC(c.intValue());

      // Create a scaling transformation that will place the layout
      // in a 2*ConcreteDiagram.SCALEx2*ConcreteDiagram.SCALE box.
      double scaleX = 2*ConcreteDiagram.SCALE/layout.getBounds().getWidth();
      double scaleY = 2*ConcreteDiagram.SCALE/layout.getBounds().getHeight();
      double scale = scaleX < scaleY ? scaleX : scaleY;

      AffineTransform tx = AffineTransform.getScaleInstance(scale, scale);

      // Scale the curve's polygon and convert it to an integer polygon.
      path.transform(tx);

      return Utils.pathToPolygon(path);
    }

    private DecorationBox getBox(String title, JordanLayout layout) {
        // Create a scaling transformation that will place the layout
      // in a 2*ConcreteDiagram.SCALEx2*ConcreteDiagram.SCALE box that
      // is padded by an extra 1/6.
      double scaleX = 2*ConcreteDiagram.SCALE/layout.getBounds().getWidth();
      double scaleY = 2*ConcreteDiagram.SCALE/layout.getBounds().getHeight();
      double scale = scaleX < scaleY ? scaleX : scaleY;

      AffineTransform tx = AffineTransform.getScaleInstance(scale*(1.0+1.0/6.0), scale*(1.0+1.0/6.0));

        // Get a scaled version of the layout's bounding box.
      GeneralPath path = new GeneralPath(layout.getBounds());
      path.transform(tx);

      // Convert the bounding box to an integer polygon and return
      // it as a DecorationBox.
    return new DecorationBox(title, "", Utils.pathToPolygon(path));
    }
}
