package rwd.views.smoother;

import java.awt.Polygon;
import java.awt.geom.PathIterator;
import java.awt.geom.Point2D;
import java.util.Vector;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import javax.swing.JLabel;

import rwd.concretesyntax.ConcreteUnitaryDiagram;
import rwd.concretesyntax.eulerdiagram.ConcreteContour;
import rwd.gui.editor.viewers.DiagramViewerInterface;
import rwd.metrics.Comparison;
import rwd.metrics.ContourPointsDifferenceComparison;
import rwd.metrics.ContourPositionComparison;
import rwd.metrics.Metric;

/**
 */
public class ComparisonMetricSmoother 
	extends MetricSmoother 
{    
	protected ConcreteUnitaryDiagram stillCud;
	protected ConcreteUnitaryDiagram newStillCud;
	
	public ComparisonMetricSmoother(
			ConcreteUnitaryDiagram cud, 
    		ConcreteUnitaryDiagram stillCud, 
    		DiagramViewerInterface viewer, 
    		int iterations) 
    {
		super(cud, viewer, iterations);

		this.stillCud = stillCud;

	    Vector metrics = new Vector();
		metrics.add(new ContourPositionComparison(cud, 1.0, stillCud));
		metrics.add(new ContourPointsDifferenceComparison(cud, 1.0, stillCud));

		List checkers = new Vector();
		checkers.add(new ZonePresenceChecker(cud));
		checkers.add(new SpiderChecker(cud));
		manipulator = new ComparisonFastHillClimber(checkers, cud, metrics, viewer, true, true, true);
	}

	/**
	 * This allows a manipulator to be passed to the smoother, if a dataDisplay 
	 * is required, it should be set directly in the manipulator
	 */
	public ComparisonMetricSmoother(
  			ConcreteUnitaryDiagram moveCUD, 
  			ConcreteUnitaryDiagram stillCud, 
  			DiagramViewerInterface viewer, 
  			Manipulator manipulator,
  			JLabel progressLabel, 
  			int iterations) 
  	{
		super(moveCUD, viewer, manipulator, progressLabel, iterations);
		this.stillCud = stillCud;
	}


	/** Use this to execute the smoother */
    public void run() {
        // create a clone diagram
        ConcreteUnitaryDiagram oldStillCud = newStillCud;
		newStillCud = (ConcreteUnitaryDiagram)stillCud.getClone();

		normalizeContourPolygons(newStillCud);

        // propagate the new cps through to the comparison metrics in the manipulator
		changeMetricsStillCud(newStillCud);

		newStillCud.refreshZones(true);
		newStillCud.notifyObservers();

		manipulator.makeChange(iterations, progressLabel);
		if(viewer != null) {
			manipulator.viewer.draw(manipulator.cud);
		}

        // make sure that the cps is copied back to the old one
		changeMetricsStillCud(oldStillCud);
	}


	/**
	 * 
	 * Iterates through the metrics of the manipulator and 
 	 * sets them to the given cps.
 	 * 
 	 * */
	public void changeMetricsStillCud(ConcreteUnitaryDiagram stillCud) {
		Iterator i = manipulator.metrics.iterator();
		while(i.hasNext()) {
			Metric metric = (Metric)i.next();
			if(metric instanceof Comparison) {
				Comparison comparisonMetric = (Comparison)metric;
				comparisonMetric.setStillCud(stillCud);
			}
		}
	}


	/**
	 * This normalizes the moving and still polys for number and
	 * rotation of points and also creates the new contour polys from
	 * the old ones for the newStillcpList.
	 */
	public void normalizeContourPolygons(ConcreteUnitaryDiagram newStillCud) {
        // For each still concrete contour
		Set stillContourPolys = newStillCud.getConcreteContours();
		Set moveContourPolys = cud.getConcreteContours();
        Iterator i = stillContourPolys.iterator();
        while(i.hasNext()){
			ConcreteContour oldStillcp = (ConcreteContour)i.next();
			// replace still cp with a copy, in case we have to change the number of
			// points in it
			Polygon newStillP = deepPolygonCopy(oldStillcp.getPolygon());            
			ConcreteContour newStillcp = new ConcreteContour(oldStillcp.getContour(), newStillP);
            // For each 'move' concrete contour
            Iterator j = moveContourPolys.iterator();
            while(j.hasNext()){
				ConcreteContour movecp = (ConcreteContour)j.next();
				Polygon moveP = movecp.getPolygon();

				if(newStillcp.getContour().equals(movecp.getContour())) {
					// set up the points list
					Vector stillPoints = getPointsList(newStillP);
					Vector movePoints = getPointsList(moveP);

					// equalize the number of points in moving poly
					int stillSize = stillPoints.size();
					int moveSize = movePoints.size();
					if(stillSize > moveSize) {
						addPoints(movePoints,stillSize-moveSize);
					}
					if(moveSize > stillSize) {
						addPoints(stillPoints,moveSize-stillSize);
					}
					// check direction of points and reverse moving poly if necessary
					if(isClockwise(stillPoints) != isClockwise(movePoints)) {
						Collections.reverse(movePoints);
					}
					// find the shift value for rotating the points
					int[] nearestIndexes = findNearestPoints(movePoints,stillPoints);

					// rotate the points of the polys
					rotateList(movePoints,nearestIndexes[0]);
					rotateList(stillPoints,nearestIndexes[1]);

					// switch both points list back to poly
				  	makePolyFromPoints(newStillP,stillPoints);
				   	makePolyFromPoints(moveP,movePoints);
				}
				oldStillcp.setPolygon(newStillcp.getPolygon());
			}
//			newStillCud.addConcreteContour(newStillcp);
		}
	}


/**
 * Rotates the given list so that the element at index is
 * the first element. Leaves the list alone if index is over
 * the list bounds.
 */
	public static void rotateList(Vector list, int index) {
		if(index > list.size()-1) {
			return;
		}
		for (int i = 0; i <index ; i++) {
			Object o = list.remove(0);
			list.add(o);
		}

	}


/**
 */
	public static int[] findNearestPoints(Vector firstPoints, Vector secondPoints) {
		

		int bestFirstIndex = 0;
		int bestSecondIndex = 0;
		
		double smallestDistance = Double.MAX_VALUE;
// first we find the shift value to place the polygon centres on top of each other
		Polygon firstPolygon = new Polygon();
		Polygon secondPolygon = new Polygon();

		makePolyFromPoints(firstPolygon, firstPoints);
		makePolyFromPoints(secondPolygon, secondPoints);

		Point2D.Double firstCentre = ContourPositionComparison.polygonCentre(firstPolygon);
		Point2D.Double secondCentre = ContourPositionComparison.polygonCentre(secondPolygon);

		double shiftX = firstCentre.x-secondCentre.x;
		double shiftY = firstCentre.y-secondCentre.y;

		double[] firstPoint = {0.0,0.0,0.0,0.0,0.0,0.0};
		double[] secondPoint = {0.0,0.0,0.0,0.0,0.0,0.0};

		for (int i = 0; i <= firstPoints.size()-1; i++) {
			firstPoint = (double[])firstPoints.get(i);
			double firstX = firstPoint[0] - shiftX;
			double firstY = firstPoint[1] - shiftY;

			for (int j = 0; j <= secondPoints.size()-1; j++) {
				secondPoint = (double[])secondPoints.get(j);
				double secondX = secondPoint[0];
				double secondY = secondPoint[1];

				double dx = secondX - firstX;
				double dy = secondY - firstY;
				double distance = Math.sqrt(dx * dx + dy * dy);
				if(distance < smallestDistance) {
					smallestDistance = distance;
					bestFirstIndex = i;
					bestSecondIndex = j;
				}
			}
		}

		int[] ret = {bestFirstIndex,bestSecondIndex};

		return ret;
	}


/**
 * For debugging
 */
	public static void outputPolygon(Polygon p) {
		System.out.println("Number of Points "+p.npoints);
		double[] pathPoint = {0.0,0.0,0.0,0.0,0.0,0.0};
		PathIterator pi = p.getPathIterator(null);
		while(!pi.isDone()) {
			int segmentType = pi.currentSegment(pathPoint);
			if(segmentType == PathIterator.SEG_MOVETO) {
				System.out.println("MOVETO x "+pathPoint[0]+" y "+pathPoint[1]);
			}
			if(segmentType == PathIterator.SEG_LINETO) {
				System.out.println("LINETO x "+pathPoint[0]+" y "+pathPoint[1]);
			}
			if(segmentType == PathIterator.SEG_CLOSE) {
				System.out.println("CLOSE x "+pathPoint[0]+" y "+pathPoint[1]);
			}
			pi.next();
		}
	}


/**
 * For debugging
 */
	public static void outputPoints(Vector points) {
		System.out.println("Number of Points "+points.size());
		double[] point = {0.0,0.0,0.0,0.0,0.0,0.0};
		Iterator i = points.iterator();
		while(i.hasNext()) {
			point = (double[])i.next();
			System.out.println(point[0]+" "+point[1]);
		}

	}

/**
 * Adds the required number of points to the points list.
 */
	public void addPoints(Vector pointList,int numberToAdd) {

		int remaining = numberToAdd;
		while(remaining > 0) {

			int jump = 1;
			if(numberToAdd < pointList.size()) {
				jump = pointList.size()/remaining;
			}

			int index = 0;
			while(index < pointList.size() && remaining > 0) {
				double[] firstPoint = {0.0,0.0,0.0,0.0,0.0,0.0};
				if(index == 0) {
					firstPoint = (double[])pointList.get(pointList.size()-1);
				} else {
					firstPoint = (double[])pointList.get(index-1);
				}
				double[] secondPoint = (double[])pointList.get(index);
				double[] newPoint = getPointBetween(firstPoint,secondPoint);
				pointList.add(index,newPoint);				
				index += jump+1;
				remaining--;
			}
		}
	}

	public double[] getPointBetween(double[] firstPoint, double[] secondPoint) {
		double[] ret = {0.0,0.0,0.0,0.0,0.0,0.0};

		int x1 = (int)firstPoint[0];
		int y1 = (int)firstPoint[1];
		int x2 = (int)secondPoint[0];
		int y2 = (int)secondPoint[1];

		ret[0] = x1+(x2-x1)/2;
		ret[1] = y1+(y2-y1)/2;

		return ret;
	}


/**
 * Deletes the current points from the poly, and gives it the new
 * points in the point list. This is a list of arrays.
 * Note that because the polygon is reused, the size of the
 * arrays in the polygon are only made larger, never smaller
 * so that iterations across the array may cause difficulties.
 * This should not be a problem when using path iterators.
 */
	public static void makePolyFromPoints(Polygon p,Vector pointList){

		p.reset();

		double[] point = {0.0,0.0,0.0,0.0,0.0,0.0};
		Iterator i = pointList.iterator();
		while(i.hasNext()) {
			point = (double[])i.next();
			p.addPoint((int)(point[0]),(int)(point[1]));
		}
	}

/**
 * Copys a simple polygon, dupicating all the points.
 */
	public static Polygon deepPolygonCopy(Polygon p) {

		Polygon ret = new Polygon();

		double[] pathPoint = {0.0,0.0,0.0,0.0,0.0,0.0};
		PathIterator pi = p.getPathIterator(null);
		while(!pi.isDone()) {
			int segmentType = pi.currentSegment(pathPoint);
			if(segmentType == PathIterator.SEG_LINETO || segmentType == PathIterator.SEG_MOVETO) {
				ret.addPoint((int)(pathPoint[0]),(int)(pathPoint[1]));
			}
			pi.next();
		}
		return ret;
	}
    
/**
 * This gives a list of points of the contour polygon,
 * without the start and end duplicated. As this is the default
 * for the current implementation, nothing clever is required at
 * the moment. This might have to change.
 */
	public static Vector getPointsList(Polygon p) {

		Vector ret = new Vector();

		double[] pathPoint;
		PathIterator pi = p.getPathIterator(null);
		while(!pi.isDone()) {
			pathPoint = new double[6];
			int segmentType = pi.currentSegment(pathPoint);
			if(segmentType == PathIterator.SEG_LINETO || segmentType == PathIterator.SEG_MOVETO) {
				ret.add(pathPoint);
			}

			pi.next();
		}

		return ret;
	}

/**
 * Paul can you implement this please, finds out if the list of points
 * representing a polygon go round clockwise or not. Each point is an array
 * where index 0 is the x coordinate and index 1 is the y coordinate.
 */
    public static boolean isClockwise(Vector stillPoints) {

        double angle = 0;
        for (int i = 0; i < stillPoints.size(); i++) {
            double[] p1 = (double[]) stillPoints.get(i);
            double[] p2 = (double[]) stillPoints.get((i + 1) % stillPoints.size());
            double[] p3 = (double[]) stillPoints.get((i + 2) % stillPoints.size());
            double x1 = p1[0];
            double y1 = p1[1];
            double x2 = p2[0];
            double y2 = p2[1];
            double x3 = p3[0];
            double y3 = p3[1];
            double theta1 = Math.atan2(y2 - y1, x2 - x1);
            double theta2 = Math.atan2(y3 - y2, x3 - x2); 
            double turn = theta1 - theta2;
            if (turn > Math.PI) {
                turn -= 2 * Math.PI;                
            }
            else if (turn < -Math.PI){
                turn += 2 * Math.PI;
            }
            angle += turn;
        }
    
        return angle > 0;
    }


	public ConcreteUnitaryDiagram getNewStillCud() {
		return newStillCud;
	}

}

