package rwd.views.smoother;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.Vector;
import java.util.List;

import javax.swing.JLabel;

import rwd.concretesyntax.ConcreteUnitaryDiagram;
import rwd.gui.editor.viewers.DiagramViewerInterface;
import rwd.metrics.EulerMetric;
import rwd.metrics.Metric;

public abstract class Manipulator {
	// has a set of Metrics
	// has a set of structure-checkers
	// enacts some change in some order
	private List checkers;
    protected ConcreteUnitaryDiagram cud;
	public Vector metrics;
	protected double oldQuality = 0.0;
	public DiagramViewerInterface viewer;
	/** maximum distance in x or y direction that a point can be moved */
	public double maxMoveContourDistance = 200.0;
	public double maxMoveFeetDistance = maxMoveContourDistance / 2;
	/** set this to reduce the maximum move linearly to 0 over the number of steps */
	public boolean cooling = true;
	/** set this to move whole contours as well as points */
	public boolean contourMove = true;
	/** set this to output data as the manipulator progresses */
	public boolean outputData = false;
	public boolean moveContours = true;
	public boolean moveFeet = true;
	public boolean animated = true;
	/** set to true to return from the thread */
	protected boolean stopFlag = false;
	protected AbstractParameterWindow dataDisplay;
	protected NumberFormat formatter = new DecimalFormat("0.#####E0");

	public Manipulator(List scs, ConcreteUnitaryDiagram cud, Vector metrics, DiagramViewerInterface viewer, boolean moveContours, boolean moveFeet, boolean animated) {
		checkers = scs;
        this.cud = cud;
		this.metrics = metrics;
		//System.out.println("metrics in Manipulator are " + metrics);
		this.viewer = viewer;
		this.moveContours = moveContours;
		this.moveFeet = moveFeet;
		this.animated = animated;
		if (checkers != null) {
			for(int i=0; i<checkers.size(); i++) {
				StructureChecker checker = (StructureChecker)checkers.get(i);
				checker.startAt(cud);
			}
		}
	}

	protected boolean structureStillOK() { // uses checker
		//System.out.println("call ok on checker");
		if (checkers != null) {
			for(int i=0; i<checkers.size(); i++) {
				StructureChecker checker = (StructureChecker)checkers.get(i);
				boolean ok = checker.checkStillOK(cud);
				if (!ok)
					return false;
			}
		}
		return true;
	}

	/**
	 * Get the quality of all metrics
	 */
	protected Quality currentQuality() {
        //System.out.println("quality with "+metrics);
		return (getMetricsQuality(metrics));
	}

	/**
	 * Get only the quality of metrics that deal with Points
	 */
	protected Quality currentQualityPoints() {
		Vector metricList = getPointMetrics();
		return (getMetricsQuality(metricList));
	}

	/**
	 * Get only the quality of metrics that deal with polygons
	 */
	protected Quality currentQualityPolygons() {
		Vector metricList = getPolygonMetrics();
		return (getMetricsQuality(metricList));
	}

	/**
	 * Test for presence of metrics that deal with polygons
	 */
	protected boolean hasPointMetrics() {
		Vector metricList = getPointMetrics();
		if (metricList.size() == 0) {
			return false;
		} else {
			return true;
		}
	}

	/**
	 * Test for presence of metrics that deal with polygons
	 */
	protected boolean hasPolygonMetrics() {
		Vector metricList = getPolygonMetrics();
		if (metricList.size() == 0) {
			return false;
		} else {
			return true;
		}
	}

	protected Vector getPointMetrics() {
		Vector metricList = new Vector();
		for (int index = 0; index < metrics.size(); index++) {
			Metric m = (Metric) (metrics.get(index));
			if (m.usedByPointMove == true) {
				metricList.add(m);
			}
		}
		return (metricList);
	}

	protected Vector getPolygonMetrics() {
		Vector metricList = new Vector();
		for (int index = 0; index < metrics.size(); index++) {
			Metric m = (Metric) (metrics.get(index));
			if (m.usedByPolygonMove == true) {
				metricList.add(m);
			}
		}
		return (metricList);
	}

	protected Quality getMetricsQuality(Vector metricList) {
		if (outputData) {
			dataDisplay.say("");
		}
		double eulerScore = 0;
		double spiderScore = 0;
		double score = 0;
		for (int index = 0; index < metricList.size(); index++) {
			Metric m = (Metric) (metricList.get(index));
			//System.out.println("using metric " + m.getClass()+" weight "+m.getWeighting());
			double measure = m.measure();
			if (m instanceof EulerMetric)
				eulerScore += measure;
			else
				spiderScore += measure;
			score += measure;
			if (outputData) {
				// this exports String quality description
				// without knowledge of whether its an improvement or not
				String value = formatter.format(measure);
				if (value.equals("?"))
					value = "0";
				dataDisplay.say("" + m.getName() + ":  " + value);
			}
		}
		if (outputData) {
			dataDisplay.say("SCORE: " + formatter.format(eulerScore) +" | "+ formatter.format(spiderScore));
		}
		return new Quality(eulerScore, spiderScore);
	}

	public boolean getMoveContours() {
		return moveContours;
	}

	public void setMoveContours(boolean b) {
		this.moveContours = b;
	}

	public boolean getMoveFeet() {
		return moveFeet;
	}

	public void setMoveFeet(boolean b) {
		this.moveFeet = b;
	}

	public boolean getAnimated() {
		return animated;
	}

	public void setAnimated(boolean b) {
		this.animated = b;
	}

	public boolean getStopFlag() {
		return stopFlag;
	}

	public void setStopFlag(boolean stopFlag) {
		this.stopFlag = stopFlag;
	}

	// updateLabel should be changed using setText son each iteration
	abstract void makeChange(int numberOfSteps, JLabel updateLabel);

	protected void setFlags(AbstractParameterWindow window) {
		setMoveContours(window.moveContoursBox.isSelected());
		setMoveFeet(window.moveFeetBox.isSelected());
		setAnimated(window.animateBox.isSelected());
	}
}
